# --8<--8<--8<--8<--
#
# Copyright (C) 2006 Smithsonian Astrophysical Observatory
#
# This file is part of yaxx
#
# yaxx is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# yaxx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the 
#       Free Software Foundation, Inc. 
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--

package Yaxx;
use strict;
use warnings;

##****************************************************************************
# HERE are a handful of accessor methods which ensure that the values are
# properly initialized.  These may throw a fatal error if that is not
# possible.
##****************************************************************************

##****************************************************************************
sub X {
##****************************************************************************
    my $self = shift;
    if (@_) {
	$self->{X} = $_[0];
	return $self->{X};	# Can't fall through because of pairwise XY
    }
    
    return ($self->XY)[0];
}

##****************************************************************************
sub Y {
##****************************************************************************
    my $self = shift;
    if (@_) {
	$self->{Y} = $_[0];
	return $self->{Y};	# Can't fall through because of pairwise XY
    }
    
    return ($self->XY)[1];
}


##****************************************************************************
sub ra {
##****************************************************************************
    my $self = shift;
    if (@_) {
	$self->{ra} = $_[0];
	return $self->{ra};	# Can't fall through
    }
    
    return ($self->ra_dec)[0];
}

##****************************************************************************
sub dec {
##****************************************************************************
    my $self = shift;
    if (@_) {
	$self->{dec} = $_[0];
	return $self->{dec};	# Can't fall through
    }
    
    return ($self->ra_dec)[1];
}

##****************************************************************************
sub counts {
##****************************************************************************
    my $self = shift;
    my $det = $self->{detector} || '';
    if (@_) {
	$self->{counts}{$det} = $_[0];
    }
    
    return $self->{counts}{$det};
}

##****************************************************************************
sub exposure {
##****************************************************************************
    my $self = shift;
    my $det = $self->{detector} || '';
    if (@_) {
	$self->{exposure}{$det} = $_[0];
    }
    
    return $self->{exposure}{$det};
}

##****************************************************************************
sub object {
##****************************************************************************
    my $self = shift;
    $self->{object} = $_[0] if (@_);

    if (not defined $self->{object}) {
	my $hdr = fits_read_hdr($self->file('obsid_evt'), 'events');
	$self->{object} = $hdr->{OBJECT};
	message(1, "WARNING - no object name defined") unless defined $self->{object};
    }

    return $self->{object}
}

##****************************************************************************
sub gal_nh {
#
# Get galactic N_H for source RA,Dec and set $self->gal_nh
##****************************************************************************
    my $self = shift;
    return $self->{gal_nh} if (defined $self->{gal_nh});  # Already set

    # Try to get gal_nh from existing mdl.in file (faster)
    my $src_dir = $self->file('src_dir');
    if (my @mdl_files = glob "$src_dir/*.in") {
	foreach (@mdl_files) {
	    my $lines << io($_);
	    if ($lines =~ /\b gal\.nh \s* = \s* (\S+)/x) {
		$self->{gal_nh} =  $1;
		return $self->{gal_nh};
	    }
	}
    }

    my $tmp1 = POSIX::tmpnam;
    my $tmp2 = POSIX::tmpnam;

    my ($ra_hms, $dec_hms) = dec2hms($self->ra, $self->dec);
    $ra_hms =~ s/:/ /g;
    $dec_hms =~ s/:/ /g;
    
    my $cmd = "prop_colden_exe d nrao f j2000 :$tmp1:$tmp2 > /dev/null";

    open IN, "> $tmp1";
    print IN "$ra_hms $dec_hms\n";
    close IN;
    
    system $cmd;

    open OUT, "$tmp2" or die "ERROR - Colden did not create expected output file\n";
    while (<OUT>) {
	my @vals = split;
	next unless (@vals >= 9 and $vals[6] =~ /^[-0-9\.]+$/ and $vals[7] =~ /^[-0-9\.]+$/);
	if ($vals[8] eq '-') {
	    message(1, "ERROR - Colden did not give valid NH\n");
	    return;
	}
	$self->{gal_nh} =  $vals[8] / 100.0; # Convert from units of 10^20 (colden) to 10^22 (sherpa)
    }
    close OUT;
    unlink $tmp1, $tmp2;

    return $self->{gal_nh};
}

##****************************************************************************
sub min_energy_kev {
# Read-only accessor to get min energy in keV
##****************************************************************************
    my $self = shift;
    return $self->min_energy / 1000.0;
}

##****************************************************************************
sub max_energy_kev {
# Read-only accessor to get min energy in keV
##****************************************************************************
    my $self = shift;
    return $self->max_energy / 1000.0;
}

1;
