%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% $Id: sherpa_plotfns.sl,v 1.1.1.1 2004/02/20 16:01:30 aldcroft Exp $
%
% File:
%   sherpa_plotfns.sl
%
% Thread:
%   http://cxc.harvard.edu/ciao/threads/sherpa_slangplot/
%   http://ledas-cxc.star.le.ac.uk/ciao/threads/sherpa_slangplot/
%
% Requires:
%   CIAO 3.0
%
% Aim:
%   See how the new plotting capabilities in Sherpa (introduced in
%   CIAO 3.0) help us when customising Sherpa plots.
%
%   In CIAO 2.3 we needed to use the lpfd/lpfr functions for
%   the customisation. In CIAO 3.0 we do not need this since the
%   customisations can be made to work for "lplot 2 fit resid".
%   The lpfr() and lpfd() functions are left in in case people
%   have got used to using them.
%
% How it works:
%   There are two steps:
%     1) set up the fields in the relevant Sherpa configuration
%        variables (aka state object)
%     2) set up functions that will be called before and after every
%        "residual" style plot (resid, delchi, ratio) which makes the 
%        tweaks we can not do via the configuration variables
%
% Notes:
%  . Since the prefunc can not find out everything about the
%    requested plot it can not "avoid" situations like
%      sherpa> lplot 2 fit 1 resid 2
%    or - a perhaps more common plot -
%      sherpa> lplot 3 fit 1 resid 2 arf 3
%
% History:
%   2002/04/25 - added lpfd()
%   2002/06/05 - take advantage of sherpa_utils functions if they exist
%   2003/07/24 - Major re-work for CIAO 3.0: do not really need lpfr/lpfd
%                anymore but left in for backwards compatability
%   2003/08/11 - fix lpfd() to plot delchi and not resid
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% execute a ChIPS command, exiting on error
private define ceval( command ) {
  variable retval = chips_eval( command );
  if ( 0 != retval )
    error( "\nError: unable to execute the following ChIPS command\n" +
           command + "\n" );
} % ceval()

%
% Usage:
%   sherpa.resplot.prefunc  = &_sherpa_resid_prefunc;
%   sherpa.resplot.postfunc = &_sherpa_resid_postfunc;
%
% Aim:
%   Apply customisations to plots that match
%     "lplot 2 fit resid"
%   or - more generally -
%     "lplot 2 plottype n resid/delchi/ratio n"
%
%   In these plots we FORCE the residual plot to have a
%   linear y scale, no matter what the setting of
%   sherpa.resplot.y_log. This is because "set_log" also
%   changes the sherpa.resplot.y_log field and we do not
%   want to make the user have to reset it each time
%   This is done in the postfunc hook since we need the
%   plot to have been drawn at this point.
%
%   We also ensure that the first two panes have a common x range
%   (the minimum ratehr than maximum so that "lp 2 arf resid"
%    does not have a lot of blank area in the residual plot)
%   in the postfunc hook
%
% Note:
%   We try to make the hook smart enough not to do anything
%   when it should not - eg during
%     "lplot resid"
%   However, we can not make it foolproof, as noted in the
%   "Notes" section at the head of the file
%

static define _sherpa_resid_prefunc(data, label, dataset, type) {

  % assume we only want the re-sizing happening IF
  % the residuals are being drawn in the second pane
  %  - this does NOT correctly handle situations like
  %    "lplot 3 fit resid arf"
  %
  if ( 2 != chips_get_pane() ) return;

  % set up the title
  % - we add the reduced chi-square value IF it is defined
  %
  variable model  = get_source_expr(dataset);
  variable fitres = get_goodness(dataset);  % get_fit() does not work correctly in CIAO 3.0
  variable title;
  if ( NULL != fitres.rstat )
    title = sprintf( "title '%s \\chi^2_r= %f'", model, fitres.rstat );
  else
    title = sprintf( "title '%s'", model );
  ceval( title );

  % make the second plot smaller (and remove the gap between the plots)
  ceval( "d 1 location 0.15 0.9 0.4 0.9" );
  ceval( "d 2 location 0.15 0.9 0.1 0.4" );
  
  % remove the x axis values from the first plot
  () = chips_set_pane(1);
  ceval( "tickvals x off" );

  % as we are a pre-func and the previous command has moved us 
  % back to the first pane we need to go back to the second
  % pane (or else the residuals would be drawn in the first pane)
  %
  () = chips_set_pane(2);

} % _sherpa_resid_prefunc

% cleans up the setting of the y_log field back to its user-selected
% value AND ensures that the first two panels have the same
% axis range. Note that if one is in logarithmic spacing
% and the other linear we leave as is.
%
static define _sherpa_resid_postfunc() {

  if ( 2 != chips_get_pane() ) return;

  % ignore the sherpa.resplot.y_log setting
  () = chips_set_yscale(1);

  % ensure that the x-axis ranges of the first two panes
  % are the same (even if they do not both have the same scale)
  % In CIAO 3.0 chips_get_xscale() does not "ignore" the scale
  % setting for the axis so we have to convert the values for
  % logged axes
  %
  () = chips_set_pane(1);
  variable xs1 = chips_get_xscale;
  variable xlo1, xhi1;
  ( xlo1, xhi1 ) = chips_get_xrange();
  if ( 0 == xs1 ) {
    xlo1 = 10^xlo1;
    xhi1 = 10^xhi1;
  }

  () = chips_set_pane(2);
  variable xs2 = chips_get_xscale;
  variable xlo2, xhi2;
  ( xlo2, xhi2 ) = chips_get_xrange();
  if ( 0 == xs2 ) {
    xlo2 = 10^xlo2;
    xhi2 = 10^xhi2;
  }

  % could find out if we need to do this but easier just to do it
  % - originally used the largest range but that can result in
  %   plots with a lot of space, so restrict to the smaller range
  %   (which will likely be the residual part)
  %
  variable xlo = max([xlo1,xlo2]);
  variable xhi = min([xhi1,xhi2]);
  () = chips_set_pane(1);
  () = chips_set_xrange(xlo,xhi);
  () = chips_set_pane(2);
  () = chips_set_xrange(xlo,xhi);

  % Ensure the two panes have the same X-axis scale - we take it
  % from the main plot in case there is a difference.
  % It is not a perfect solution - the scaling is not as good as
  % when the x_log fields match, but it just makes sure the plots
  % are usable
  %
  () = chips_set_xscale(xs1);

} % _sherpa_resid_postfunc

%%%%%%%%%%%%%%%%%%%%
%%
%% apply the customisations
%%
%% it is just a matter of setting fields in the relevant
%% configuration variable (aka state object)
%%

%% set up the hooks for residual plots
sherpa.resplot.prefunc  = &_sherpa_resid_prefunc;
sherpa.resplot.postfunc = &_sherpa_resid_postfunc;

% customise the residual plots
sherpa.resplot.x_errorbars = 0;
sherpa.resplot.symbolstyle = "bigpoint";
sherpa.resplot.symbolsize  = 1;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% public routines
%
% These are not needed any more but left in for backwards compatability
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
% Usage:
%   lpfr
%   lpfr()
%   lprf(n)
%
% Aim:
%   Plots fit and residual
%
public define lpfr () {

  % are we called correctly?
  variable n = 1;
  switch ( _NARGS )
  { case 0: }
  { case 1: n = (); }
  { () = __pop_args(_NARGS); usage( _function_name + "( [dataset number] )\n" ); }

  % if it fails the default error message is good enough
  () = sherpa_eval( "lplot 2 fit " + string(n) + " resid " + string(n) );

} % lpfr()

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
% Usage:
%   lpfd
%   lpfd()
%   lpfd(n)
%
% Aim:
%   Plots fit and delchi
%
public define lpfd () {

  % are we called correctly?
  variable n = 1;
  switch ( _NARGS )
  { case 0: }
  { case 1: n = (); }
  { () = __pop_args(_NARGS); usage( _function_name + "( [dataset number] )\n" ); }

  % if it fails the default error message is good enough
  () = sherpa_eval( "lplot 2 fit " + string(n) + " delchi " + string(n) );

} % lpfd()

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
