#!/usr/bin/perl

# --8<--8<--8<--8<--
#
# Copyright (C) 2006 Smithsonian Astrophysical Observatory
#
# This file is part of yaxx
#
# yaxx is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# yaxx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the 
#       Free Software Foundation, Inc. 
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--


#####################################################################
#
# Yaxx: Yet Another Xray Xtractor
#
# Copyright (C) 2006 by the Smithsonian Astrophysical Observatory
# 
# This code is released under the GNU General Public License.  You may
# find a copy at L<http://www.fsf.org/copyleft/gpl.html>.
#
# Creator: Tom Aldcroft
#
#####################################################################

use warnings;
use strict;

# Load various packages
use POSIX qw(strftime strtod);
use Data::Dumper;
use Cwd;
use English;
use File::Copy;
use File::Basename;
use File::Path;
use File::Glob qw(:glob);
use File::Find;
use File::Spec;
use File::Temp;
use IO::All;
use Config::General;
use Clone;
use Getopt::Long;
use Carp;
use HTML::Table;
use List::MoreUtils qw(uniq);
use Text::ParseWords;
use Regexp::Common;

# Standard yaxx modules
use Yaxx::ConfigFile;
use Yaxx::ObjectFile;

# 'Local' packages (not on CPAN)
use Ska::Convert;
use CFITSIO::Simple;

# Set up global file and directory variables
my $program_dir = dirname( io($PROGRAM_NAME)->rel2abs );
my ($program_name) = fileparse($PROGRAM_NAME, qr{\.pl|\.bin});
my $config_file = "${program_name}.cfg";
my $analysis_dir = cwd;		# Directory where yaxx command called
my @summary;			# Data for source fitting summary
my @sources;			# YaxxSource objects for every source in sample

# Check for ciao (existence and version) and setenv PGPLOT_DIR
# check_for_ciao();

# Read in yaxx options.  First Yaxx cmd line options, then defaults, then user-spec'd config file
# In addition, if the -test flag is given, include config from 'yaxx.cfg.test' if available
my %par = get_cmd_line_options();	# Yaxx command line options
my ($cfg_files, %cfg) = get_config();

# Set Directory where yaxx results go.  
# Actually chdir there and use cwd to get resolved path in case of links
my $output_dir = eval { my $dir = io($cfg{output_dir})->chdir; cwd };
croak "ERROR - no output directory $cfg{output_dir}\n" if ($@ or not defined $output_dir);

# Load the appropriate Yaxx modules based on the data type (e.g. Chandra or XMM)
croak "ERROR - must define the 'mission' parameter in config file yaxx.cfg\n"
  unless defined $cfg{mission};
eval "use Yaxx::$cfg{mission}";
croak "ERROR - mission parameter '$cfg{mission}' not supported: $@\n" if ($@);

init_log_and_message();		# initialize log file and message subroutine

# Print the list of config files (now that the messages are set up)
message(1,"Reading config file $_\n") foreach (map { $_->file } @{$cfg_files});
set_environment();		# Set various environment / program options

# Copy thread resource files and change thread attrib to be the thread name
# (possibly stripping path info)
$cfg{thread} = copy_thread_resource_files($cfg_files);

make_param_dir();		# Create temp location for parameter files (local to this process)

# Read list of source objects for fitting
my @src_list = Yaxx::get_object_data($par{objlist} || $cfg{objlist});	

# Make the global array of sources.  This includes all sources in sample
# (even if not processed this time) so that next_source and prev_source
# are correct.  Note that 'ignored' sources *are* skipped at this stage.

my %source_attr_def =  ( output_dir    => $output_dir,
			 analysis_dir  => $analysis_dir,
			 config_file   => $config_file,
			 program_dir   => $program_dir,
			 preclean      => $par{preclean},
			 clean         => $par{clean},
		       );

my $last_source;
foreach my $src_data (@src_list) {
    my $obsid = $src_data->{obsid};

    chdir $output_dir;

    for my $fakeid (0 .. $cfg{fake_instances}) {
	# and use these values to initialize Yaxx objects.
	my $source = Yaxx->new(\%cfg,
			       { %source_attr_def,
				 fakeid => $fakeid,
			       },
			       $src_data,
			      );

	# Skip source if there is an IGNORE file present for this obsid
	next if (-e $source->file('ignore_obs')
		 or -e $source->file('ignore_src'));

	# Set process flag based on whether source passes input filter criteria (obsid,srcid)
	$source->process( filter_src($src_data, \%par) );

	# Set the next_source and prev_source attributes
	push @sources, $source;
	if (defined $last_source) {
	    $last_source->next_source($source);
	    $source->prev_source($last_source);
	}
	$last_source = $source;
    }
}

# Make a dummy source for preprocessing
my $source = Yaxx->new(\%cfg,
		       \%source_attr_def,
		      );

PREPROCESS_STEP: while (my $process_step = $source->next_process_step('pre')) {
    foreach my $loop_val (Yaxx::build_task_loops($process_step)) {
	$process_step->{loop_val} = $loop_val;
	my $ok = $source->run_task( $process_step );

	# Stopped_at could be OK, for instance if source is locked by another process
	# or quick_check decides that it has been processed successfully
	$source->stopped_at($process_step->{name}) unless ($ok or defined $source->stopped_at);

	# However, unless the success_on_not_OK param is set, then a stop => failure
	$source->failed_at($process_step->{name})  unless ($ok or defined $source->failed_at
							   or $process_step->{success_on_not_OK});
    }
}

# Now do the actual processing of each source

SOURCE: foreach my $source (@sources) {
    chdir $output_dir;

    # Skip source if process flag not set (could be in src_list data file)
    next SOURCE unless $source->process;

    next SOURCE if (defined $par{src} and $par{src} != $source->srcid);

    # If doing a summary, just extract relevant information and go to the next Obsid
    #   THIS SHOULD go in Yaxx just as for validate (in fact real summary work is
    #   already there)
    if ($par{summary}) {
	push @summary, get_summary($source);
	next SOURCE;
    }

    # Now the real work for each source to be processed

    chdir $output_dir;
    next if (-e $source->file('ignore_src')); # Skip if IGNORE file is present in src dir

    message(1, "\n");
    message(1, '*' x 70 . "\n");
    message(1, sprintf("***  %-60s  ***\n", $source->unique_identifier));
    message(1, '*' x 70 . "\n");
    message(1, '', { time=>1 });

    next if $par{dryrun};

    # Go through the "pipeline" of processing steps

  PROCESS_STEP: while (my $process_step = $source->next_process_step) {
	foreach my $loop_val (Yaxx::build_task_loops($process_step)) {
	    $process_step->{loop_val} = $loop_val;
	    my $ok = $source->run_task( $process_step );

	    # Stopped_at could be OK, for instance if source is locked by another process
	    # or quick_check decides that it has been processed successfully
	    $source->stopped_at($process_step->{name}) unless ($ok or defined $source->stopped_at);

	    # However, unless the success_on_not_OK param is set, then a stop => failure
	    $source->failed_at($process_step->{name})  unless ($ok or defined $source->failed_at
							       or $process_step->{success_on_not_OK});
	}
    }

    message(1,(defined $source->failed_at ? "*** FAILURE" : "*** SUCCESS") .
	    " for " . $source->unique_identifier . "\n",
	    { time=>1 });
    $source->close_log();
}

# Make a dummy source for postprocessing
$source = Yaxx->new(\%cfg,
		    \%source_attr_def,
		   );

POSTPROCESS_STEP: while (my $process_step = $source->next_process_step('post')) {
    foreach my $loop_val (Yaxx::build_task_loops($process_step)) {
	$process_step->{loop_val} = $loop_val;
	my $ok = $source->run_task( $process_step );

	# Stopped_at could be OK, for instance if source is locked by another process
	# or quick_check decides that it has been processed successfully
	$source->stopped_at($process_step->{name}) unless ($ok or defined $source->stopped_at);

	# However, unless the success_on_not_OK param is set, then a stop => failure
	$source->failed_at($process_step->{name})  unless ($ok or defined $source->failed_at
							   or $process_step->{success_on_not_OK});
    }
}

if ($par{summary}) {
    make_summary_table(\@summary);
} else {
    make_html_report_index(\@sources);
}
    
##****************************************************************************
sub clean_up {
#  This gets called either by END or signal trap
##****************************************************************************
    my $sig = shift;

    # Force dmcoords process in Yaxx to die.  This way of doing it is
    # ugly.  Need to move the DMcoord_open routine into yaxx.pl and provide
    # dmcoord access to $source in the main loop by passing DMcoord object ref
    # to Yaxx object.  Then yaxx.pl can properly start and stop the
    # dmcoord instances.  Do that next release!
    {
	no warnings;
	undef $Yaxx::Chandra::DMcoord;
    }

    if ($sig) {
	print STDERR "Stopped because of signal $sig\n" ;
	exit(1);
    }
    return;
}
    
##****************************************************************************
sub get_summary {
# 
# Call Yaxx functions to generate summary of Sherpa fit data for
# each specified source 
##****************************************************************************
    my $source = shift;
    my @summary;

    my $summ = $source->summary($par{summary});
    if ($summ) {
	message(1, "Generated summary for " . $source->unique_identifier . "\n");
	push @summary, $summ;
    }

    return @summary;
}

##****************************************************************************
sub get_cmd_line_options {
##****************************************************************************
    my %par;

    my $status = GetOptions(\%par,
			    'config=s',
			    'obsid=s',
			    'src=i',
			    'preclean=s',
			    'summary=s',
			    'objlist=s',
			    'test!',
			   );

    die "ERROR - Invalid command line option\n" unless ($status);

    return %par;
}

##****************************************************************************
sub check_for_ciao {
##****************************************************************************
    local $_;
    foreach (qw (ASCDS_INSTALL ASCDS_BIN ASCDS_LIB CALDB)) {
	unless (defined $ENV{$_}) {
	    message(0, "\nERROR - CIAO variable '$_' not defined.  Are you in the CIAO environment?\n");
	    die "\n";
	}
    }
}

##****************************************************************************
sub set_environment {
##****************************************************************************
    $| = 1;			# Set stdout flushing
    $SIG{INT} = \&clean_up;		# Make sur to clean up if ctrl-c is pressed
    $SIG{QUIT} = \&clean_up;		# Make sur to clean up if die'ing

   # Set up env. vars based on tool environment definitions (e.g. CIAO, SAS, ftools etc)
    foreach my $env_name (listify( $cfg{use_environment} )) {
	message(1, "Initializing $env_name environment\n");
	my $env = $cfg{environment}{$env_name};
	unless ($env) {
	    message(0, "ERROR - 'environment' system config parameter for '$env_name' not found\n");
	    die "\n";
	}
	foreach my $setenv (listify( $env->{setenv} )) {
            Shell::GetEnv->new('csh', "setenv $setenv", {Startup=>0})->import_envs();
	}
	foreach my $script (listify( $env->{script} )) {
	    eval { Shell::GetEnv->new($env->{shell}, $script, {Startup=>0})->import_envs() };
	    if ($@) {
		message(0, "ERROR - failed to initialize $env_name environment using '$script'\n");
		die "\n";
	    }
	    message(5, "Successfully inited $env_name using $script\n");
	}
    }

    $ENV{SLANG_SCRIPT_PATH} = join ":", "$program_dir/slang", $ENV{SLANG_SCRIPT_PATH};
}      

##****************************************************************************
sub get_config {
# Get the system defaults, project, thread, and test config files
##****************************************************************************
    my $config = {};
    my @cfg_files;

    # System config file
    push @cfg_files, Yaxx::ConfigFile->new(path     => $program_dir,
					   name     => $config_file,
					   type     => 'system',
					   priority => 10,
					  );

    # Project config file.  Use command line '-config' value if supplied
    push @cfg_files, Yaxx::ConfigFile->new(path     => $analysis_dir,
					   name     => $par{config} || $config_file,
					   type     => 'project',
					   priority => 30,
					  );

    # Assemble config to this point in order to discover the thread name
    $config = $_->merge($config) for @cfg_files;

    # Specify info about thread and test config files. 
    my $thread = $config->{thread} or die "Need a thread value in the system or project config file\n";

    # Thread config file
    push @cfg_files, Yaxx::ConfigFile->new(path     => [ "$analysis_dir",
							 "$program_dir/resources" ],
					   name     => "$thread/$config_file",
					   type     => 'thread',
					   priority => 20,
					  );
    
    # Config file for development testing
    push @cfg_files, Yaxx::ConfigFile->new(path     => $analysis_dir,
					   name     => $config_file,
					   required => 0,
					   name     => "${config_file}.test",
					   type     => 'test',
					   priority => 40,
					  )
      if $par{test};

    # Reset config and assemble final config (accounting for priorities)
    @cfg_files = sort {$a->priority <=> $b->priority} @cfg_files;
    $config = {};		
    $config = $_->merge($config) for (@cfg_files);

    return \@cfg_files, %{ $config };
}

##****************************************************************************
sub init_log_and_message {
##****************************************************************************
    my $dir = io($output_dir)->chdir;

    # Make a unique log file name, and create log directory if needed
    if ($cfg{log_file} and $cfg{log_file} =~ /\/$/) {
	eval { io($cfg{log_file})->mkpath };
	die "Could not make log file dir $cfg{log_file}: $@" if $@;
	$cfg{log_file} .= strftime("%Y-%m-%d_%H:%M", localtime);
    }

    # Create a message object to handle all program output
    my $msg = Ska::Message->new(stdout  => 1,
				file    => $cfg{log_file},
				auto_cr => 0,
			       );
    # Initialize the Yaxx::message routine
    Yaxx::message($msg, $cfg{verbose_stdout}, $cfg{verbose_master_log});
}


##****************************************************************************
sub make_html_report_index {
##****************************************************************************
    my $sources = shift;
    return unless @{$sources};	# Make sure there is something to do
    return unless (defined $cfg{file_definition}{report_index}
		   and $cfg{file_definition}{report_index} =~ /\S/);

    message(2, "\nCreating HTML report index\n");

    # Not great, but use methods from first source for help
    my $source0 = $sources->[0]; 
    
    # Make path to report_index file, make sure we can create it, and chdir to path
    my $report_index_file = io($source0->file('report_index.html'));
    eval { $report_index_file->touch };
    die "ERROR - could not make path to $report_index_file: $@" if $@;
    my $dir = eval { io($report_index_file->absolute->filepath)->chdir };
    die "ERROR - could not chdir to " . $report_index_file->filepath . ": $@" if $@;

    my @cols = listify($source0->html_report_index->{column});

    # Generated table header, using either formatting_rule label or capitalized var name
    my @table_data = ( [ map { $_->{header} } @cols ] );
  SOURCE:
    for my $source (@{$sources}) {

	# Get a copy of the stored (via Dumper) object in case current source
	# object was not processed this time
	unless ($source->process) {
	    my $new_source;
	    unless ($new_source = $source->restore_object) {
		message(1, "WARNING - could not restore source for $source->unique_identifier\n");
		next SOURCE;
	    }
	    $source = $new_source;
	}

	my @row_data;
	for my $col (@cols) {
	    # Put value into index table.  They might not be defined if processing failed,
	    # so suppress the 'use of uninitialized var' warnings
	    no warnings;
	    my $col_value = $source->interpolate_string($col->{value});
	    push @row_data, q{<a href="} . $source->file('report.html') . qq{"> $col_value </a>};
	}
	push @table_data, \@row_data;
    }

    my $html_table = HTML::Table->new(-class => 'style1',
				      -data => \@table_data);
    $html_table->setRowHead(1);

    my $report_index_html < io($source0->file('report_index_template.html'));
    eval {
	$report_index_html = $source0->interpolate_string($report_index_html,
							  INDEX_TABLE => $html_table->getTable,
							 );
    };
    message(2, "ERROR - $@\n") if $@;

    # Write the new index HTML content to the index file
    $report_index_html > io( $source0->file('report_index.html') );
}

##****************************************************************************
sub make_summary_table {
##****************************************************************************
    my $summ = shift;
    my $s;
    my $c;
    my @col;		# Ordered list of columns from summary->{cols}
    my %data;
    my $NAN = (-1)**0.5;	# Generate a NaN for undefined elements in summary table

    # Find all the columns.  {data} hash may have additional upper and lower limit cols
    @col = uniq map { @{$_->{cols}} } @{$summ};

    # collect the data into the structures for writing to FITS file
    foreach $c (@col) {
	foreach $s (@{$summ}) {
	    push @{$data{$c}}, defined ($s->{data}->{$c}) ? $s->{data}->{$c} : $NAN;
	}
    }

    if (fits_write_bintbl("$analysis_dir/summ_$par{summary}.fits", %data, {col_names => \@col})) {
	message(3, "Wrote summary for model $par{summary} to summ_$par{summary}.fits\n");
    } else {
	message(2, "ERROR - failed to write summary for model $par{summary} to summ_$par{summary}.fits\n");
    }
}

##****************************************************************************
sub filter_src {
##****************************************************************************
    my $src_data = shift;
    my $par = shift;
    local $_;

    # If input parameter set to a non-trivial value, take it as a comma-separated
    # list of allowed values.  Make corresponding sure $src_data vals match
    for (qw(obsid srcid)) {
	if ($par->{$_}) {
	    my %filter = map { $_ => 1 } split(/\s*,\s*/, $par->{$_});
	    return 0 unless $filter{ $src_data->{$_} };
	}
    }
    return 1;
}

##****************************************************************************
sub make_param_dir {
##****************************************************************************
    my $clean = shift;
    my $dir = File::Temp::tempdir(CLEANUP => 1);

    die "PFILES env variable is empty" unless $ENV{PFILES};
    my @pfiles = split ';', $ENV{PFILES};
    $ENV{PFILES} = join ';', ($dir, $pfiles[-1]);
    print "Temporary dir is $dir\n";
}

########################################################################################
sub listify {
########################################################################################
    if (@_ and defined $_[0]) {
	return ref($_[0]) eq 'ARRAY' ? @{$_[0]} : ($_[0]);
    } else {
	return ();
    }
}

########################################################################################
sub message {
########################################################################################
    Yaxx::message(@_);
}

END {
    clean_up();
}

########################################################################################
sub copy_thread_resource_files {
# Make sure that all thread files are available in the output directory
########################################################################################
    my $cfg_files = shift;
    my $thread_file;

    # Find the thread name.  Expect success since thread is required in config files
    foreach my $cfg_file (@{$cfg_files}) {
	$thread_file = $cfg_file->file if $cfg_file->type eq 'thread';
    }

    my @in_resources_dir  = io($thread_file)->splitdir;
    pop @in_resources_dir;	# Drop the yaxx.cfg
    my $thread_name = pop @in_resources_dir;

    my $out_resources_dir = io->catdir($cfg{output_dir}, $cfg{file_definition}{resources_dir});
    $out_resources_dir->mkpath;
    my $out_resources_dir_path = $out_resources_dir->absolute->pathname;

    # Go to resources_yaxx_dir, get files recursively and copy structure to $resources_dir
    my $in_resources_parent_dir = io->catdir(@in_resources_dir)->chdir;
    for my $yaxx_file (io($thread_name), io($thread_name)->deep->all ) {
	my $local_file = io->join($out_resources_dir_path,
				  $yaxx_file->relative->pathname); #  . '/' .
	next if grep(/^\.svn$/, $yaxx_file->splitdir);
	if ($yaxx_file->is_dir) {
	    io($local_file->name)->mkpath;
	} else {
	    # Copy unless local file exists and local file is newer than yaxx file
	    $yaxx_file > $local_file unless ($local_file->exists
					     and $local_file->mtime > $yaxx_file->mtime );
	}
    }

    return $thread_name;
}

