%
% File:
%   acisabs.sl
%
% Version:
%   FORTRAN XSPEC model: 1.1
%   S-Lang conversion:   2
%
% History:
%   v1.1-1          Initial version
%   v1.1-2 11/21/02 Added cacheing of results in simple case
%                   Added bintype parameter (0 = keV, 1 = Angstrom)
%                   Henke data now included in file (no need for HENKEDIR env var)
%                   Cacheing is now controlled by a model parameter (cache)
%                   Removed the norm parameter since it does nothing
%                   Some speed improvements
%                   Added acisabs namespace; can be loaded into chips
%                   Added apply_acisabs() function to modify an ARF by the model
%                   [see the apply_acisabs shell script]
%
% Thread:
%   http://cxc.harvard.edu/ciao/threads/sherpa_acisabs/
%
% Questions:
%   See the CXC Helpdesk: http://cxc.harvard.edu/helpdesk/
%
% Acknowledgements:
%   See comments to acisabs function (below) for complete
%   comments and acknowledgements and
%   http://www.astro.psu.edu/users/chartas/xcontdir/xcont.html
%   for the original FORTRAN code/data
%
% Notes:
%  Gratings analysis
%   v1.1-2 added the bintype parameter, with a default setting of 0
%   which means that the analysis is assumed to be done in energy
%   space. If you are analysing grating data (more correctly, if you
%   are using 'analysis wave') then you need to set this parameter
%   to 1.
%
%   Prior to v1.1-2, everything was done assuming that 'analysis energy'
%   was being used, which led to erroneous results for the grating case.
%
%  Cacheing of results
%   v1.1-2 introduced the capability of the model to cache itself
%   (controlled by the cache parameter)
%   this speeds up fits in Sherpa when all the parameters are fixed
%   (the general mode for non grating data)
%   The way it is done is a hack [it will not be needed in CIAO 3.0]
%   and so may not work in all cases.
%   An alternative implementation would be to cache the results
%   from all parameter settings, using an associative array, but
%   that can be memory intensive. 
%
%  Henke files
%   Prior to v1.1-2, the Henke data were read from files
%   (called X.nff for X = o, h, n, c) that were stored in the
%   directory pointed to by the $HENKEDIR env variable.
%   These datasets are now included in this file, so HENKEDIR is
%   no longer needed
%

% define our own namespace
%
implements("acisabs");

% version string
public variable _acisabs_version_string = "1.1-2";

% We only register the models if the command is available
% (ie if we are in Sherpa). To allow CIAO 3 testing we also check
% for the CIAO 3 version of sherpa_register_model (register_model)
%

% have we been eval'ed before?
private variable _runonce;

if (__is_initialized(&_runonce) == 0) {

  % don't do this initialisation next time the file is loaded in this session
  _runonce = 1;

  % is [sherpa_]register_model() available
  %
  private variable _is_reg_model = 0;
  if ( is_defined("sherpa_register_model") == 2 ) _is_reg_model = 1;
  else if ( is_defined("register_model") == 2 )   _is_reg_model = 2;

  % is fits_read_key in the _ciao or Global namespace
  % (CIAO 2.3/3.0 difference)
  %
  private variable _is_read_key_global = 0;
  if ( is_defined("fits_read_key") == 2 ) _is_read_key_global = 1;

  % at some point varmm will stop cacheing the data
  %
  private variable _does_varmm_cache = 1;

  % do we need to bother about checking the output names used by apply_acisabs()?
  private variable _apply_acisabs_outnames = NULL;
  if ( _does_varmm_cache ) _apply_acisabs_outnames = Assoc_Type [Integer_Type,0];

}

%
% internal routine to read in a FITS keyword
% - fails on error
% - uses the correct namespace for fits_read_key()
%   [it changes CIAO 2.3 -> CIAO 3.0]
%
variable _keyval; % needs to be a package-wide variable for eval to see it
private define _get_fits_key( file, key ) {

  _keyval = NULL;
  if ( _is_read_key_global ) {
    eval( "acisabs->_keyval = fits_read_key(\"" + file + "\",\"" + key + "\");" );
  } else {
    eval( "acisabs->_keyval = _ciao->fits_read_key(\"" + file + "\",\"" + key + "\");" );
  }

  % I think fits_read_key() "dies" so this check is unnescessary
  if ( _keyval == NULL ) {
    message( "\n*** ERROR: unable to find " + key + " in header of " + file + "\n\n" );
    return NULL;
  }

  return _keyval;

} % _get_fits_key()

%
% Usage:
%   tdays = tstart2days( in );
%
% Aim:
%   convert TSTART to number of days since launch
%
%   in is either a number (assumed to be TSTART) or
%   the name of a file which contains TSTART in its header.
%
%   Returns the number of days since launch or NULL on
%   error
%
public define tstart2days () { % file or tstart
  variable usage_str = 
    "Usage: Tdays = tstart2days(in);\n\n" +
    "where in is either a number (assumed to be the value of TSTART)\n" +
    "or a string (assumed to be the name of a file containing the TSTART\n" +
    "keyword in its header.\n";

  if ( _NARGS != 1 ) {
    message( usage_str );
    return NULL;
  }
  variable in = ();
  variable tstart;

  switch( typeof(in) )
  {
    % assume it's the name of a file
    case String_Type:
      tstart = _get_fits_key( in, "TSTART" );
  }
  {
    % should really bother with UInteger Long/ULong types
    case Integer_Type:
      tstart = 1.0 * in;
  }
  {
    case Double_Type:
      tstart = in;
  }
  {
    % unrecognized
    message( usage_str );
    return NULL;
  }

  % convert to days since launch (TSTART=49091521.0)
  return ( tstart - 49091521.0 ) / (24.0 * 3600.0);

} % tstart2days


% prior to 1.1-3 the algorithm performed linear interpolation on
% the logarithm of the data points if they are all > 0.0
% (otherwise it would fall back to doing linear interpolation of the
%  actual data): ie
%
%    if ((x[jl] > 0. and x[ju] > 0.) and (y[jl] > 0. and y[ju] > 0.)) {
%       grad = (log10(y[ju]) - log10(y[jl]))/(log10(x[ju]) - log10(x[jl]));
%       df = grad * (log10(z[iZ]) - log10(x[jl]));
%       result[iZ] = 10.0^(log10(y[jl]) + df);
%    } else {
%       result[iZ] = y[jl]+(y[ju]-y[jl])*((z[iZ]-x[jl])/(x[ju]-x[jl]));
%    }
%
% The loop is ~ 10% faster if it doesn't include the logarithmic
% scheme [note the speed increase in a fit will be much smaller than this] 
%
% Note that the original XSPEC model uses this method of interpolation.
%
define interpol ( x, y, z_in ) {

   variable n = length(x);
   variable z;
   if (typeof(z_in) != Array_Type) {
      z = [z_in];
   } else {
      z = @z_in;
   }
   
   % the -1 is because _for(a,b,c) is a faster way of saying
   %  for ( i=a; i<=b; i+=c )
   % except that c van be -ve and the loop variable is put onto the stack, not i
   % 
   variable nz = length(z) - 1;

   variable xstart = x[0];
   variable xend   = x[n-1];
   variable inc    = (xend > xstart);
   
   variable result = 0.0*@z;

   variable iZ, jl, ju, jm, grad, df;
   _for ( 0, nz, 1 ) {
      iZ = ();
      if (( inc == 1 and ((z[iZ] < xstart) or (z[iZ] > xend))) or
          ( inc == 0 and ((z[iZ] > xstart) or (z[iZ] < xend)))) {
         vmessage("interpol: Asking for %e, min is %e, max is %e",z[iZ],
                  xstart,xend);
         message("interpol: Returning 0 for this value.");
         result[iZ] = 0;
      } else {

         jl = 0;
         ju = n;
         while (ju - jl > 1) {
            jm = (ju + jl) / 2;
            if ((z[iZ] > x[jm]) == inc) {
               jl = jm;
            } else {
               ju = jm;
            }
         }

         % ------ Z is sandwiched between JL and JU ------

	 % use linear interpolation (could interpolate the logarithms of the
         % data but it doesn't appear to be worth the speed decrease)
         %
         result[iZ] = y[jl]+(y[ju]-y[jl])*((z[iZ]-x[jl])/(x[ju]-x[jl]));
      }
   }

   if (typeof(z_in) != Array_Type) result = result[0];

   return result;

} % interpol()

%---------------------------------------

% Henke data (previously read in from X.nff for
% X = o, h, n, c
%

static variable awO=15.9994;
static variable awH=1.00797;
static variable awN=14.0067;
static variable awC=12.01115;

static variable enO, muO;
static variable enH, muH;
static variable enN, muN;
static variable enC, muC;

% species = O
enO = Float_Type[502];
muO = Float_Type[502];

% The columns here are the values from o.nff
enO[0] =  1.000000e+01; muO[0] =  7.032800e-01; 
enO[1] =  1.016170e+01; muO[1] =  7.072260e-01; 
enO[2] =  1.032610e+01; muO[2] =  7.073770e-01; 
enO[3] =  1.049310e+01; muO[3] =  7.075280e-01; 
enO[4] =  1.066280e+01; muO[4] =  7.076780e-01; 
enO[5] =  1.083530e+01; muO[5] =  7.078300e-01; 
enO[6] =  1.101060e+01; muO[6] =  7.079800e-01; 
enO[7] =  1.118860e+01; muO[7] =  7.081310e-01; 
enO[8] =  1.136960e+01; muO[8] =  7.082820e-01; 
enO[9] =  1.155350e+01; muO[9] =  7.084330e-01; 
enO[10] =  1.174040e+01; muO[10] =  7.085840e-01; 
enO[11] =  1.193030e+01; muO[11] =  7.101690e-01; 
enO[12] =  1.212320e+01; muO[12] =  7.124210e-01; 
enO[13] =  1.231930e+01; muO[13] =  7.146800e-01; 
enO[14] =  1.251860e+01; muO[14] =  7.169470e-01; 
enO[15] =  1.272100e+01; muO[15] =  7.192210e-01; 
enO[16] =  1.292680e+01; muO[16] =  7.215020e-01; 
enO[17] =  1.313590e+01; muO[17] =  7.237900e-01; 
enO[18] =  1.334830e+01; muO[18] =  7.260850e-01; 
enO[19] =  1.356420e+01; muO[19] =  7.336400e-01; 
enO[20] =  1.378360e+01; muO[20] =  7.460660e-01; 
enO[21] =  1.400660e+01; muO[21] =  7.587020e-01; 
enO[22] =  1.423310e+01; muO[22] =  7.715530e-01; 
enO[23] =  1.446330e+01; muO[23] =  7.846220e-01; 
enO[24] =  1.469730e+01; muO[24] =  8.014060e-01; 
enO[25] =  1.493500e+01; muO[25] =  8.329150e-01; 
enO[26] =  1.517650e+01; muO[26] =  8.656630e-01; 
enO[27] =  1.542200e+01; muO[27] =  8.996990e-01; 
enO[28] =  1.567140e+01; muO[28] =  9.734960e-01; 
enO[29] =  1.592490e+01; muO[29] =  1.076910e+00; 
enO[30] =  1.618250e+01; muO[30] =  1.191310e+00; 
enO[31] =  1.644420e+01; muO[31] =  1.317870e+00; 
enO[32] =  1.671020e+01; muO[32] =  1.457870e+00; 
enO[33] =  1.698050e+01; muO[33] =  1.612740e+00; 
enO[34] =  1.725510e+01; muO[34] =  1.784070e+00; 
enO[35] =  1.753420e+01; muO[35] =  1.973590e+00; 
enO[36] =  1.781780e+01; muO[36] =  2.190840e+00; 
enO[37] =  1.810600e+01; muO[37] =  2.445890e+00; 
enO[38] =  1.839890e+01; muO[38] =  2.730650e+00; 
enO[39] =  1.869640e+01; muO[39] =  3.048560e+00; 
enO[40] =  1.899880e+01; muO[40] =  3.315410e+00; 
enO[41] =  1.930610e+01; muO[41] =  3.479520e+00; 
enO[42] =  1.961840e+01; muO[42] =  3.651750e+00; 
enO[43] =  1.993570e+01; muO[43] =  3.778790e+00; 
enO[44] =  2.025820e+01; muO[44] =  3.836500e+00; 
enO[45] =  2.058580e+01; muO[45] =  3.895090e+00; 
enO[46] =  2.091880e+01; muO[46] =  3.954570e+00; 
enO[47] =  2.125710e+01; muO[47] =  4.014970e+00; 
enO[48] =  2.160090e+01; muO[48] =  4.070780e+00; 
enO[49] =  2.195030e+01; muO[49] =  4.113860e+00; 
enO[50] =  2.230530e+01; muO[50] =  4.157390e+00; 
enO[51] =  2.266610e+01; muO[51] =  4.201380e+00; 
enO[52] =  2.303270e+01; muO[52] =  4.245850e+00; 
enO[53] =  2.340530e+01; muO[53] =  4.290770e+00; 
enO[54] =  2.378380e+01; muO[54] =  4.336180e+00; 
enO[55] =  2.416850e+01; muO[55] =  4.371080e+00; 
enO[56] =  2.455940e+01; muO[56] =  4.391980e+00; 
enO[57] =  2.495660e+01; muO[57] =  4.412970e+00; 
enO[58] =  2.536030e+01; muO[58] =  4.434070e+00; 
enO[59] =  2.577050e+01; muO[59] =  4.455270e+00; 
enO[60] =  2.618730e+01; muO[60] =  4.476570e+00; 
enO[61] =  2.661090e+01; muO[61] =  4.497970e+00; 
enO[62] =  2.704130e+01; muO[62] =  4.519480e+00; 
enO[63] =  2.747860e+01; muO[63] =  4.541080e+00; 
enO[64] =  2.792310e+01; muO[64] =  4.562800e+00; 
enO[65] =  2.837470e+01; muO[65] =  4.584610e+00; 
enO[66] =  2.883370e+01; muO[66] =  4.606530e+00; 
enO[67] =  2.930000e+01; muO[67] =  4.628550e+00; 
enO[68] =  2.977390e+01; muO[68] =  4.650680e+00; 
enO[69] =  3.025550e+01; muO[69] =  4.672920e+00; 
enO[70] =  3.074490e+01; muO[70] =  4.695260e+00; 
enO[71] =  3.124210e+01; muO[71] =  4.717700e+00; 
enO[72] =  3.174750e+01; muO[72] =  4.740260e+00; 
enO[73] =  3.226090e+01; muO[73] =  4.762920e+00; 
enO[74] =  3.278270e+01; muO[74] =  4.785690e+00; 
enO[75] =  3.331300e+01; muO[75] =  4.808570e+00; 
enO[76] =  3.385180e+01; muO[76] =  4.831560e+00; 
enO[77] =  3.439930e+01; muO[77] =  4.854660e+00; 
enO[78] =  3.495570e+01; muO[78] =  4.877870e+00; 
enO[79] =  3.552110e+01; muO[79] =  4.901190e+00; 
enO[80] =  3.609560e+01; muO[80] =  4.924620e+00; 
enO[81] =  3.667940e+01; muO[81] =  4.908530e+00; 
enO[82] =  3.727270e+01; muO[82] =  4.881770e+00; 
enO[83] =  3.787550e+01; muO[83] =  4.855150e+00; 
enO[84] =  3.848820e+01; muO[84] =  4.828680e+00; 
enO[85] =  3.911070e+01; muO[85] =  4.802350e+00; 
enO[86] =  3.974320e+01; muO[86] =  4.776170e+00; 
enO[87] =  4.038610e+01; muO[87] =  4.750120e+00; 
enO[88] =  4.103930e+01; muO[88] =  4.724230e+00; 
enO[89] =  4.170310e+01; muO[89] =  4.698460e+00; 
enO[90] =  4.237760e+01; muO[90] =  4.672850e+00; 
enO[91] =  4.306300e+01; muO[91] =  4.647370e+00; 
enO[92] =  4.375950e+01; muO[92] =  4.622030e+00; 
enO[93] =  4.446730e+01; muO[93] =  4.596830e+00; 
enO[94] =  4.518650e+01; muO[94] =  4.571760e+00; 
enO[95] =  4.591740e+01; muO[95] =  4.546830e+00; 
enO[96] =  4.666000e+01; muO[96] =  4.522040e+00; 
enO[97] =  4.741470e+01; muO[97] =  4.497380e+00; 
enO[98] =  4.818160e+01; muO[98] =  4.472860e+00; 
enO[99] =  4.896090e+01; muO[99] =  4.439240e+00; 
enO[100] =  4.975280e+01; muO[100] =  4.385300e+00; 
enO[101] =  5.055760e+01; muO[101] =  4.332010e+00; 
enO[102] =  5.137530e+01; muO[102] =  4.279370e+00; 
enO[103] =  5.220620e+01; muO[103] =  4.227370e+00; 
enO[104] =  5.305060e+01; muO[104] =  4.176010e+00; 
enO[105] =  5.390870e+01; muO[105] =  4.125260e+00; 
enO[106] =  5.478060e+01; muO[106] =  4.075140e+00; 
enO[107] =  5.566670e+01; muO[107] =  4.025620e+00; 
enO[108] =  5.656700e+01; muO[108] =  3.976700e+00; 
enO[109] =  5.748200e+01; muO[109] =  3.928380e+00; 
enO[110] =  5.841170e+01; muO[110] =  3.880650e+00; 
enO[111] =  5.935640e+01; muO[111] =  3.833490e+00; 
enO[112] =  6.031650e+01; muO[112] =  3.786910e+00; 
enO[113] =  6.129210e+01; muO[113] =  3.740900e+00; 
enO[114] =  6.228340e+01; muO[114] =  3.695440e+00; 
enO[115] =  6.329080e+01; muO[115] =  3.650540e+00; 
enO[116] =  6.431450e+01; muO[116] =  3.606180e+00; 
enO[117] =  6.535470e+01; muO[117] =  3.562360e+00; 
enO[118] =  6.641180e+01; muO[118] =  3.519070e+00; 
enO[119] =  6.748590e+01; muO[119] =  3.476310e+00; 
enO[120] =  6.857750e+01; muO[120] =  3.434070e+00; 
enO[121] =  6.968670e+01; muO[121] =  3.392340e+00; 
enO[122] =  7.081380e+01; muO[122] =  3.352060e+00; 
enO[123] =  7.195910e+01; muO[123] =  3.312480e+00; 
enO[124] =  7.312300e+01; muO[124] =  3.273360e+00; 
enO[125] =  7.430570e+01; muO[125] =  3.234700e+00; 
enO[126] =  7.550760e+01; muO[126] =  3.196510e+00; 
enO[127] =  7.672890e+01; muO[127] =  3.158770e+00; 
enO[128] =  7.796990e+01; muO[128] =  3.121460e+00; 
enO[129] =  7.923100e+01; muO[129] =  3.084600e+00; 
enO[130] =  8.051250e+01; muO[130] =  3.048180e+00; 
enO[131] =  8.181470e+01; muO[131] =  3.012180e+00; 
enO[132] =  8.313800e+01; muO[132] =  2.976610e+00; 
enO[133] =  8.448270e+01; muO[133] =  2.941460e+00; 
enO[134] =  8.584910e+01; muO[134] =  2.906730e+00; 
enO[135] =  8.723770e+01; muO[135] =  2.872400e+00; 
enO[136] =  8.864870e+01; muO[136] =  2.838480e+00; 
enO[137] =  9.008250e+01; muO[137] =  2.804960e+00; 
enO[138] =  9.153950e+01; muO[138] =  2.771840e+00; 
enO[139] =  9.302010e+01; muO[139] =  2.739110e+00; 
enO[140] =  9.452460e+01; muO[140] =  2.706760e+00; 
enO[141] =  9.605350e+01; muO[141] =  2.674800e+00; 
enO[142] =  9.760710e+01; muO[142] =  2.643210e+00; 
enO[143] =  9.918580e+01; muO[143] =  2.606980e+00; 
enO[144] =  1.007900e+02; muO[144] =  2.558210e+00; 
enO[145] =  1.024200e+02; muO[145] =  2.510350e+00; 
enO[146] =  1.040770e+02; muO[146] =  2.463390e+00; 
enO[147] =  1.057600e+02; muO[147] =  2.417300e+00; 
enO[148] =  1.074710e+02; muO[148] =  2.372080e+00; 
enO[149] =  1.092090e+02; muO[149] =  2.327700e+00; 
enO[150] =  1.109750e+02; muO[150] =  2.284160e+00; 
enO[151] =  1.127700e+02; muO[151] =  2.241430e+00; 
enO[152] =  1.145940e+02; muO[152] =  2.199490e+00; 
enO[153] =  1.164480e+02; muO[153] =  2.158340e+00; 
enO[154] =  1.183310e+02; muO[154] =  2.117970e+00; 
enO[155] =  1.202450e+02; muO[155] =  2.078340e+00; 
enO[156] =  1.221900e+02; muO[156] =  2.039460e+00; 
enO[157] =  1.241660e+02; muO[157] =  2.001310e+00; 
enO[158] =  1.261750e+02; muO[158] =  1.963870e+00; 
enO[159] =  1.282150e+02; muO[159] =  1.927130e+00; 
enO[160] =  1.302890e+02; muO[160] =  1.891080e+00; 
enO[161] =  1.323970e+02; muO[161] =  1.855700e+00; 
enO[162] =  1.345380e+02; muO[162] =  1.821020e+00; 
enO[163] =  1.367140e+02; muO[163] =  1.789190e+00; 
enO[164] =  1.389250e+02; muO[164] =  1.757910e+00; 
enO[165] =  1.411720e+02; muO[165] =  1.727170e+00; 
enO[166] =  1.434560e+02; muO[166] =  1.696980e+00; 
enO[167] =  1.457760e+02; muO[167] =  1.667310e+00; 
enO[168] =  1.481340e+02; muO[168] =  1.638160e+00; 
enO[169] =  1.505300e+02; muO[169] =  1.609520e+00; 
enO[170] =  1.529640e+02; muO[170] =  1.581380e+00; 
enO[171] =  1.554390e+02; muO[171] =  1.553740e+00; 
enO[172] =  1.579530e+02; muO[172] =  1.526570e+00; 
enO[173] =  1.605070e+02; muO[173] =  1.499890e+00; 
enO[174] =  1.631030e+02; muO[174] =  1.473660e+00; 
enO[175] =  1.657420e+02; muO[175] =  1.447900e+00; 
enO[176] =  1.684220e+02; muO[176] =  1.422590e+00; 
enO[177] =  1.711460e+02; muO[177] =  1.397720e+00; 
enO[178] =  1.739150e+02; muO[178] =  1.373280e+00; 
enO[179] =  1.767270e+02; muO[179] =  1.349270e+00; 
enO[180] =  1.795860e+02; muO[180] =  1.325680e+00; 
enO[181] =  1.824910e+02; muO[181] =  1.297750e+00; 
enO[182] =  1.854420e+02; muO[182] =  1.259850e+00; 
enO[183] =  1.884420e+02; muO[183] =  1.223060e+00; 
enO[184] =  1.914890e+02; muO[184] =  1.187340e+00; 
enO[185] =  1.945870e+02; muO[185] =  1.152660e+00; 
enO[186] =  1.977340e+02; muO[186] =  1.119000e+00; 
enO[187] =  2.009320e+02; muO[187] =  1.086320e+00; 
enO[188] =  2.041820e+02; muO[188] =  1.054590e+00; 
enO[189] =  2.074850e+02; muO[189] =  1.023790e+00; 
enO[190] =  2.108400e+02; muO[190] =  9.938940e-01; 
enO[191] =  2.142510e+02; muO[191] =  9.648690e-01; 
enO[192] =  2.177160e+02; muO[192] =  9.366900e-01; 
enO[193] =  2.212370e+02; muO[193] =  9.093350e-01; 
enO[194] =  2.248160e+02; muO[194] =  8.827770e-01; 
enO[195] =  2.284520e+02; muO[195] =  8.569970e-01; 
enO[196] =  2.321470e+02; muO[196] =  8.319680e-01; 
enO[197] =  2.359020e+02; muO[197] =  8.076720e-01; 
enO[198] =  2.397170e+02; muO[198] =  7.851410e-01; 
enO[199] =  2.435950e+02; muO[199] =  7.670500e-01; 
enO[200] =  2.475350e+02; muO[200] =  7.493770e-01; 
enO[201] =  2.515380e+02; muO[201] =  7.321110e-01; 
enO[202] =  2.556070e+02; muO[202] =  7.152460e-01; 
enO[203] =  2.597410e+02; muO[203] =  6.987660e-01; 
enO[204] =  2.639420e+02; muO[204] =  6.826660e-01; 
enO[205] =  2.682110e+02; muO[205] =  6.669360e-01; 
enO[206] =  2.725490e+02; muO[206] =  6.515680e-01; 
enO[207] =  2.769570e+02; muO[207] =  6.363900e-01; 
enO[208] =  2.814370e+02; muO[208] =  6.213600e-01; 
enO[209] =  2.859890e+02; muO[209] =  6.065670e-01; 
enO[210] =  2.906150e+02; muO[210] =  5.921060e-01; 
enO[211] =  2.953150e+02; muO[211] =  5.779870e-01; 
enO[212] =  3.000920e+02; muO[212] =  5.642060e-01; 
enO[213] =  3.049450e+02; muO[213] =  5.507560e-01; 
enO[214] =  3.098780e+02; muO[214] =  5.376250e-01; 
enO[215] =  3.148900e+02; muO[215] =  5.248090e-01; 
enO[216] =  3.199830e+02; muO[216] =  5.122960e-01; 
enO[217] =  3.251580e+02; muO[217] =  5.000780e-01; 
enO[218] =  3.304180e+02; muO[218] =  4.881560e-01; 
enO[219] =  3.357620e+02; muO[219] =  4.765180e-01; 
enO[220] =  3.411920e+02; muO[220] =  4.651600e-01; 
enO[221] =  3.467110e+02; muO[221] =  4.540680e-01; 
enO[222] =  3.523190e+02; muO[222] =  4.432400e-01; 
enO[223] =  3.580170e+02; muO[223] =  4.326770e-01; 
enO[224] =  3.638080e+02; muO[224] =  4.223590e-01; 
enO[225] =  3.696920e+02; muO[225] =  4.122870e-01; 
enO[226] =  3.756720e+02; muO[226] =  4.024550e-01; 
enO[227] =  3.817480e+02; muO[227] =  3.928590e-01; 
enO[228] =  3.879220e+02; muO[228] =  3.833990e-01; 
enO[229] =  3.941970e+02; muO[229] =  3.739650e-01; 
enO[230] =  4.005730e+02; muO[230] =  3.647340e-01; 
enO[231] =  4.070520e+02; muO[231] =  3.557180e-01; 
enO[232] =  4.136350e+02; muO[232] =  3.469110e-01; 
enO[233] =  4.203260e+02; muO[233] =  3.383320e-01; 
enO[234] =  4.271240e+02; muO[234] =  3.299660e-01; 
enO[235] =  4.340320e+02; muO[235] =  3.218050e-01; 
enO[236] =  4.410520e+02; muO[236] =  3.138420e-01; 
enO[237] =  4.481860e+02; muO[237] =  3.060780e-01; 
enO[238] =  4.554350e+02; muO[238] =  2.985250e-01; 
enO[239] =  4.628020e+02; muO[239] =  2.911710e-01; 
enO[240] =  4.702870e+02; muO[240] =  2.840040e-01; 
enO[241] =  4.778940e+02; muO[241] =  2.770050e-01; 
enO[242] =  4.856230e+02; muO[242] =  2.701840e-01; 
enO[243] =  4.934780e+02; muO[243] =  2.635350e-01; 
enO[244] =  5.014590e+02; muO[244] =  2.570560e-01; 
enO[245] =  5.095700e+02; muO[245] =  2.507220e-01; 
enO[246] =  5.178120e+02; muO[246] =  2.445420e-01; 
enO[247] =  5.261870e+02; muO[247] =  2.385280e-01; 
enO[248] =  5.346980e+02; muO[248] =  2.326620e-01; 
enO[249] =  5.430000e+02; muO[249] =  2.271650e-01; 
enO[250] =  5.432000e+02; muO[250] =  4.410170e+00; 
enO[251] =  5.433460e+02; muO[251] =  4.408430e+00; 
enO[252] =  5.521340e+02; muO[252] =  4.306050e+00; 
enO[253] =  5.610650e+02; muO[253] =  4.206030e+00; 
enO[254] =  5.701390e+02; muO[254] =  4.108740e+00; 
enO[255] =  5.793610e+02; muO[255] =  4.016310e+00; 
enO[256] =  5.887320e+02; muO[256] =  3.926680e+00; 
enO[257] =  5.982540e+02; muO[257] =  3.839480e+00; 
enO[258] =  6.079300e+02; muO[258] =  3.754170e+00; 
enO[259] =  6.177630e+02; muO[259] =  3.670710e+00; 
enO[260] =  6.277550e+02; muO[260] =  3.588780e+00; 
enO[261] =  6.379080e+02; muO[261] =  3.505190e+00; 
enO[262] =  6.482260e+02; muO[262] =  3.418510e+00; 
enO[263] =  6.587110e+02; muO[263] =  3.330720e+00; 
enO[264] =  6.693650e+02; muO[264] =  3.245180e+00; 
enO[265] =  6.801910e+02; muO[265] =  3.170610e+00; 
enO[266] =  6.911930e+02; muO[266] =  3.104400e+00; 
enO[267] =  7.023720e+02; muO[267] =  3.040970e+00; 
enO[268] =  7.137330e+02; muO[268] =  2.972530e+00; 
enO[269] =  7.252770e+02; muO[269] =  2.901380e+00; 
enO[270] =  7.370080e+02; muO[270] =  2.832100e+00; 
enO[271] =  7.489280e+02; muO[271] =  2.764410e+00; 
enO[272] =  7.610420e+02; muO[272] =  2.698240e+00; 
enO[273] =  7.733510e+02; muO[273] =  2.632990e+00; 
enO[274] =  7.858590e+02; muO[274] =  2.567070e+00; 
enO[275] =  7.985700e+02; muO[275] =  2.500970e+00; 
enO[276] =  8.114860e+02; muO[276] =  2.436560e+00; 
enO[277] =  8.246110e+02; muO[277] =  2.373810e+00; 
enO[278] =  8.379490e+02; muO[278] =  2.312670e+00; 
enO[279] =  8.515020e+02; muO[279] =  2.254900e+00; 
enO[280] =  8.652740e+02; muO[280] =  2.200480e+00; 
enO[281] =  8.792690e+02; muO[281] =  2.149040e+00; 
enO[282] =  8.934910e+02; muO[282] =  2.098800e+00; 
enO[283] =  9.079430e+02; muO[283] =  2.049740e+00; 
enO[284] =  9.226280e+02; muO[284] =  2.001830e+00; 
enO[285] =  9.375510e+02; muO[285] =  1.951250e+00; 
enO[286] =  9.527150e+02; muO[286] =  1.901020e+00; 
enO[287] =  9.681240e+02; muO[287] =  1.851290e+00; 
enO[288] =  9.837830e+02; muO[288] =  1.802840e+00; 
enO[289] =  9.996950e+02; muO[289] =  1.755750e+00; 
enO[290] =  1.015860e+03; muO[290] =  1.710010e+00; 
enO[291] =  1.032290e+03; muO[291] =  1.666250e+00; 
enO[292] =  1.048990e+03; muO[292] =  1.623040e+00; 
enO[293] =  1.065960e+03; muO[293] =  1.580350e+00; 
enO[294] =  1.083200e+03; muO[294] =  1.538770e+00; 
enO[295] =  1.100720e+03; muO[295] =  1.498280e+00; 
enO[296] =  1.118520e+03; muO[296] =  1.458850e+00; 
enO[297] =  1.136610e+03; muO[297] =  1.420480e+00; 
enO[298] =  1.155000e+03; muO[298] =  1.383110e+00; 
enO[299] =  1.173680e+03; muO[299] =  1.346710e+00; 
enO[300] =  1.192660e+03; muO[300] =  1.311270e+00; 
enO[301] =  1.211950e+03; muO[301] =  1.276760e+00; 
enO[302] =  1.231550e+03; muO[302] =  1.243160e+00; 
enO[303] =  1.251470e+03; muO[303] =  1.210220e+00; 
enO[304] =  1.271720e+03; muO[304] =  1.177890e+00; 
enO[305] =  1.292290e+03; muO[305] =  1.146090e+00; 
enO[306] =  1.313190e+03; muO[306] =  1.115160e+00; 
enO[307] =  1.334430e+03; muO[307] =  1.085070e+00; 
enO[308] =  1.356010e+03; muO[308] =  1.055790e+00; 
enO[309] =  1.377940e+03; muO[309] =  1.027320e+00; 
enO[310] =  1.400230e+03; muO[310] =  9.996040e-01; 
enO[311] =  1.422880e+03; muO[311] =  9.726110e-01; 
enO[312] =  1.445890e+03; muO[312] =  9.463800e-01; 
enO[313] =  1.469280e+03; muO[313] =  9.207660e-01; 
enO[314] =  1.493040e+03; muO[314] =  8.957010e-01; 
enO[315] =  1.517190e+03; muO[315] =  8.708060e-01; 
enO[316] =  1.541730e+03; muO[316] =  8.465490e-01; 
enO[317] =  1.566670e+03; muO[317] =  8.229970e-01; 
enO[318] =  1.592010e+03; muO[318] =  8.000620e-01; 
enO[319] =  1.617760e+03; muO[319] =  7.777670e-01; 
enO[320] =  1.643920e+03; muO[320] =  7.561110e-01; 
enO[321] =  1.670510e+03; muO[321] =  7.350370e-01; 
enO[322] =  1.697530e+03; muO[322] =  7.145480e-01; 
enO[323] =  1.724990e+03; muO[323] =  6.946270e-01; 
enO[324] =  1.752890e+03; muO[324] =  6.750300e-01; 
enO[325] =  1.781240e+03; muO[325] =  6.559080e-01; 
enO[326] =  1.810050e+03; muO[326] =  6.372700e-01; 
enO[327] =  1.839320e+03; muO[327] =  6.191580e-01; 
enO[328] =  1.869070e+03; muO[328] =  6.015600e-01; 
enO[329] =  1.899300e+03; muO[329] =  5.844620e-01; 
enO[330] =  1.930020e+03; muO[330] =  5.678490e-01; 
enO[331] =  1.961240e+03; muO[331] =  5.517080e-01; 
enO[332] =  1.992960e+03; muO[332] =  5.360270e-01; 
enO[333] =  2.025200e+03; muO[333] =  5.209800e-01; 
enO[334] =  2.057950e+03; muO[334] =  5.060590e-01; 
enO[335] =  2.091240e+03; muO[335] =  4.915240e-01; 
enO[336] =  2.125060e+03; muO[336] =  4.773840e-01; 
enO[337] =  2.159430e+03; muO[337] =  4.636340e-01; 
enO[338] =  2.194360e+03; muO[338] =  4.502510e-01; 
enO[339] =  2.229850e+03; muO[339] =  4.372320e-01; 
enO[340] =  2.265920e+03; muO[340] =  4.245680e-01; 
enO[341] =  2.302570e+03; muO[341] =  4.122450e-01; 
enO[342] =  2.339810e+03; muO[342] =  4.002610e-01; 
enO[343] =  2.377660e+03; muO[343] =  3.886050e-01; 
enO[344] =  2.416110e+03; muO[344] =  3.772690e-01; 
enO[345] =  2.455190e+03; muO[345] =  3.662390e-01; 
enO[346] =  2.494900e+03; muO[346] =  3.555230e-01; 
enO[347] =  2.535260e+03; muO[347] =  3.450850e-01; 
enO[348] =  2.576260e+03; muO[348] =  3.349430e-01; 
enO[349] =  2.617930e+03; muO[349] =  3.250790e-01; 
enO[350] =  2.660270e+03; muO[350] =  3.154890e-01; 
enO[351] =  2.703300e+03; muO[351] =  3.061590e-01; 
enO[352] =  2.747030e+03; muO[352] =  2.970840e-01; 
enO[353] =  2.791460e+03; muO[353] =  2.882660e-01; 
enO[354] =  2.836610e+03; muO[354] =  2.796930e-01; 
enO[355] =  2.882490e+03; muO[355] =  2.713520e-01; 
enO[356] =  2.929110e+03; muO[356] =  2.632380e-01; 
enO[357] =  2.976480e+03; muO[357] =  2.553490e-01; 
enO[358] =  3.024630e+03; muO[358] =  2.476870e-01; 
enO[359] =  3.073550e+03; muO[359] =  2.402320e-01; 
enO[360] =  3.123260e+03; muO[360] =  2.329810e-01; 
enO[361] =  3.173780e+03; muO[361] =  2.259310e-01; 
enO[362] =  3.225110e+03; muO[362] =  2.190770e-01; 
enO[363] =  3.277270e+03; muO[363] =  2.124170e-01; 
enO[364] =  3.330280e+03; muO[364] =  2.059370e-01; 
enO[365] =  3.384150e+03; muO[365] =  1.996440e-01; 
enO[366] =  3.438880e+03; muO[366] =  1.935250e-01; 
enO[367] =  3.494500e+03; muO[367] =  1.875740e-01; 
enO[368] =  3.551020e+03; muO[368] =  1.817870e-01; 
enO[369] =  3.608460e+03; muO[369] =  1.761600e-01; 
enO[370] =  3.666820e+03; muO[370] =  1.706810e-01; 
enO[371] =  3.726130e+03; muO[371] =  1.653630e-01; 
enO[372] =  3.786400e+03; muO[372] =  1.601960e-01; 
enO[373] =  3.847640e+03; muO[373] =  1.551670e-01; 
enO[374] =  3.909870e+03; muO[374] =  1.502820e-01; 
enO[375] =  3.973110e+03; muO[375] =  1.455310e-01; 
enO[376] =  4.037380e+03; muO[376] =  1.408990e-01; 
enO[377] =  4.102680e+03; muO[377] =  1.364210e-01; 
enO[378] =  4.169030e+03; muO[378] =  1.320790e-01; 
enO[379] =  4.236460e+03; muO[379] =  1.278710e-01; 
enO[380] =  4.304980e+03; muO[380] =  1.237890e-01; 
enO[381] =  4.374620e+03; muO[381] =  1.198330e-01; 
enO[382] =  4.445370e+03; muO[382] =  1.159980e-01; 
enO[383] =  4.517270e+03; muO[383] =  1.122800e-01; 
enO[384] =  4.590330e+03; muO[384] =  1.086780e-01; 
enO[385] =  4.664580e+03; muO[385] =  1.051850e-01; 
enO[386] =  4.740030e+03; muO[386] =  1.018000e-01; 
enO[387] =  4.816690e+03; muO[387] =  9.851930e-02; 
enO[388] =  4.894600e+03; muO[388] =  9.533980e-02; 
enO[389] =  4.973770e+03; muO[389] =  9.225920e-02; 
enO[390] =  5.054210e+03; muO[390] =  8.927320e-02; 
enO[391] =  5.135960e+03; muO[391] =  8.638010e-02; 
enO[392] =  5.219030e+03; muO[392] =  8.357790e-02; 
enO[393] =  5.303440e+03; muO[393] =  8.086200e-02; 
enO[394] =  5.389220e+03; muO[394] =  7.823090e-02; 
enO[395] =  5.476390e+03; muO[395] =  7.568130e-02; 
enO[396] =  5.564970e+03; muO[396] =  7.321170e-02; 
enO[397] =  5.654980e+03; muO[397] =  7.082050e-02; 
enO[398] =  5.746440e+03; muO[398] =  6.850400e-02; 
enO[399] =  5.839390e+03; muO[399] =  6.625960e-02; 
enO[400] =  5.933830e+03; muO[400] =  6.408570e-02; 
enO[401] =  6.029810e+03; muO[401] =  6.197920e-02; 
enO[402] =  6.127330e+03; muO[402] =  5.994160e-02; 
enO[403] =  6.226440e+03; muO[403] =  5.796590e-02; 
enO[404] =  6.327150e+03; muO[404] =  5.605450e-02; 
enO[405] =  6.429480e+03; muO[405] =  5.420320e-02; 
enO[406] =  6.533480e+03; muO[406] =  5.240970e-02; 
enO[407] =  6.639150e+03; muO[407] =  5.067250e-02; 
enO[408] =  6.746540e+03; muO[408] =  4.899160e-02; 
enO[409] =  6.855650e+03; muO[409] =  4.736500e-02; 
enO[410] =  6.966540e+03; muO[410] =  4.578940e-02; 
enO[411] =  7.079220e+03; muO[411] =  4.426440e-02; 
enO[412] =  7.193720e+03; muO[412] =  4.278750e-02; 
enO[413] =  7.310070e+03; muO[413] =  4.135820e-02; 
enO[414] =  7.428310e+03; muO[414] =  3.997590e-02; 
enO[415] =  7.548450e+03; muO[415] =  3.863590e-02; 
enO[416] =  7.670540e+03; muO[416] =  3.734020e-02; 
enO[417] =  7.794610e+03; muO[417] =  3.608510e-02; 
enO[418] =  7.920680e+03; muO[418] =  3.487100e-02; 
enO[419] =  8.048790e+03; muO[419] =  3.369800e-02; 
enO[420] =  8.178980e+03; muO[420] =  3.256120e-02; 
enO[421] =  8.311260e+03; muO[421] =  3.146120e-02; 
enO[422] =  8.445690e+03; muO[422] =  3.039670e-02; 
enO[423] =  8.582290e+03; muO[423] =  2.936700e-02; 
enO[424] =  8.721110e+03; muO[424] =  2.837070e-02; 
enO[425] =  8.862160e+03; muO[425] =  2.740700e-02; 
enO[426] =  9.005500e+03; muO[426] =  2.647460e-02; 
enO[427] =  9.151160e+03; muO[427] =  2.557280e-02; 
enO[428] =  9.299170e+03; muO[428] =  2.470050e-02; 
enO[429] =  9.449580e+03; muO[429] =  2.385660e-02; 
enO[430] =  9.602420e+03; muO[430] =  2.304060e-02; 
enO[431] =  9.757730e+03; muO[431] =  2.225130e-02; 
enO[432] =  9.915550e+03; muO[432] =  2.148830e-02; 
enO[433] =  1.007590e+04; muO[433] =  2.075010e-02; 
enO[434] =  1.023890e+04; muO[434] =  2.003630e-02; 
enO[435] =  1.040450e+04; muO[435] =  1.934610e-02; 
enO[436] =  1.057280e+04; muO[436] =  1.867880e-02; 
enO[437] =  1.074380e+04; muO[437] =  1.803360e-02; 
enO[438] =  1.091760e+04; muO[438] =  1.740990e-02; 
enO[439] =  1.109420e+04; muO[439] =  1.680690e-02; 
enO[440] =  1.127360e+04; muO[440] =  1.622400e-02; 
enO[441] =  1.145590e+04; muO[441] =  1.566070e-02; 
enO[442] =  1.164120e+04; muO[442] =  1.511610e-02; 
enO[443] =  1.182950e+04; muO[443] =  1.458980e-02; 
enO[444] =  1.202080e+04; muO[444] =  1.408120e-02; 
enO[445] =  1.221530e+04; muO[445] =  1.358970e-02; 
enO[446] =  1.241280e+04; muO[446] =  1.311470e-02; 
enO[447] =  1.261360e+04; muO[447] =  1.265580e-02; 
enO[448] =  1.281760e+04; muO[448] =  1.221240e-02; 
enO[449] =  1.302500e+04; muO[449] =  1.178410e-02; 
enO[450] =  1.323560e+04; muO[450] =  1.137020e-02; 
enO[451] =  1.344970e+04; muO[451] =  1.097050e-02; 
enO[452] =  1.366720e+04; muO[452] =  1.058440e-02; 
enO[453] =  1.388830e+04; muO[453] =  1.021150e-02; 
enO[454] =  1.411290e+04; muO[454] =  9.851320e-03; 
enO[455] =  1.434120e+04; muO[455] =  9.503520e-03; 
enO[456] =  1.457310e+04; muO[456] =  9.167660e-03; 
enO[457] =  1.480890e+04; muO[457] =  8.843390e-03; 
enO[458] =  1.504840e+04; muO[458] =  8.530300e-03; 
enO[459] =  1.529180e+04; muO[459] =  8.228050e-03; 
enO[460] =  1.553910e+04; muO[460] =  7.936280e-03; 
enO[461] =  1.579040e+04; muO[461] =  7.654650e-03; 
enO[462] =  1.604580e+04; muO[462] =  7.382830e-03; 
enO[463] =  1.630540e+04; muO[463] =  7.120490e-03; 
enO[464] =  1.656910e+04; muO[464] =  6.867350e-03; 
enO[465] =  1.683710e+04; muO[465] =  6.623080e-03; 
enO[466] =  1.710940e+04; muO[466] =  6.387420e-03; 
enO[467] =  1.738610e+04; muO[467] =  6.160060e-03; 
enO[468] =  1.766740e+04; muO[468] =  5.940730e-03; 
enO[469] =  1.795310e+04; muO[469] =  5.729190e-03; 
enO[470] =  1.824350e+04; muO[470] =  5.525160e-03; 
enO[471] =  1.853860e+04; muO[471] =  5.328410e-03; 
enO[472] =  1.883840e+04; muO[472] =  5.138690e-03; 
enO[473] =  1.914310e+04; muO[473] =  4.955770e-03; 
enO[474] =  1.945270e+04; muO[474] =  4.779430e-03; 
enO[475] =  1.976740e+04; muO[475] =  4.609450e-03; 
enO[476] =  2.008710e+04; muO[476] =  4.446520e-03; 
enO[477] =  2.041200e+04; muO[477] =  4.288580e-03; 
enO[478] =  2.074210e+04; muO[478] =  4.136320e-03; 
enO[479] =  2.107760e+04; muO[479] =  3.989540e-03; 
enO[480] =  2.141850e+04; muO[480] =  3.848040e-03; 
enO[481] =  2.176500e+04; muO[481] =  3.711620e-03; 
enO[482] =  2.211700e+04; muO[482] =  3.580090e-03; 
enO[483] =  2.247470e+04; muO[483] =  3.453280e-03; 
enO[484] =  2.283820e+04; muO[484] =  3.331000e-03; 
enO[485] =  2.320760e+04; muO[485] =  3.213100e-03; 
enO[486] =  2.358300e+04; muO[486] =  3.099410e-03; 
enO[487] =  2.396440e+04; muO[487] =  2.989780e-03; 
enO[488] =  2.435200e+04; muO[488] =  2.884070e-03; 
enO[489] =  2.474590e+04; muO[489] =  2.782120e-03; 
enO[490] =  2.514620e+04; muO[490] =  2.683800e-03; 
enO[491] =  2.555290e+04; muO[491] =  2.588990e-03; 
enO[492] =  2.596620e+04; muO[492] =  2.497550e-03; 
enO[493] =  2.638610e+04; muO[493] =  2.409350e-03; 
enO[494] =  2.681290e+04; muO[494] =  2.324290e-03; 
enO[495] =  2.724660e+04; muO[495] =  2.242250e-03; 
enO[496] =  2.768730e+04; muO[496] =  2.163110e-03; 
enO[497] =  2.813510e+04; muO[497] =  2.086780e-03; 
enO[498] =  2.859020e+04; muO[498] =  2.013160e-03; 
enO[499] =  2.905260e+04; muO[499] =  1.942140e-03; 
enO[500] =  2.952250e+04; muO[500] =  1.873630e-03; 
enO[501] =  3.000000e+04; muO[501] =  1.807540e-03; 

% and now we convert them to en/mu columns
% (we want the energy grid to be in keV for acisabs
%  to avoid some multiplication)
muO *= 4.208e7 / ( enO * awO );
enO /= 1.0e3;

% species = H
enH = Float_Type[501];
muH = Float_Type[501];

% The columns here are the values from h.nff
enH[0] =  1.000000e+01; muH[0] =  2.395400e-17; 
enH[1] =  1.016170e+01; muH[1] =  2.434140e-17; 
enH[2] =  1.032610e+01; muH[2] =  2.473510e-17; 
enH[3] =  1.049310e+01; muH[3] =  2.513520e-17; 
enH[4] =  1.066280e+01; muH[4] =  2.554170e-17; 
enH[5] =  1.083530e+01; muH[5] =  2.595490e-17; 
enH[6] =  1.101060e+01; muH[6] =  2.637460e-17; 
enH[7] =  1.118860e+01; muH[7] =  2.680120e-17; 
enH[8] =  1.136960e+01; muH[8] =  2.723470e-17; 
enH[9] =  1.155350e+01; muH[9] =  2.767520e-17; 
enH[10] =  1.174040e+01; muH[10] =  2.812290e-17; 
enH[11] =  1.193030e+01; muH[11] =  2.857770e-17; 
enH[12] =  1.212320e+01; muH[12] =  2.903990e-17; 
enH[13] =  1.231930e+01; muH[13] =  2.950960e-17; 
enH[14] =  1.251860e+01; muH[14] =  2.998690e-17; 
enH[15] =  1.272100e+01; muH[15] =  3.047200e-17; 
enH[16] =  1.292680e+01; muH[16] =  3.096480e-17; 
enH[17] =  1.313590e+01; muH[17] =  3.146560e-17; 
enH[18] =  1.334830e+01; muH[18] =  3.197460e-17; 
enH[19] =  1.356420e+01; muH[19] =  3.249170e-17; 
enH[20] =  1.378360e+01; muH[20] =  3.301730e-17; 
enH[21] =  1.400660e+01; muH[21] =  3.355130e-17; 
enH[22] =  1.423310e+01; muH[22] =  3.409400e-17; 
enH[23] =  1.446330e+01; muH[23] =  3.464540e-17; 
enH[24] =  1.469730e+01; muH[24] =  3.520580e-17; 
enH[25] =  1.493500e+01; muH[25] =  3.577520e-17; 
enH[26] =  1.517650e+01; muH[26] =  3.635380e-17; 
enH[27] =  1.542200e+01; muH[27] =  3.694180e-17; 
enH[28] =  1.567140e+01; muH[28] =  3.753930e-17; 
enH[29] =  1.590000e+01; muH[29] =  3.808680e-17; 
enH[30] =  1.610000e+01; muH[30] =  6.119150e-15; 
enH[31] =  1.618250e+01; muH[31] =  3.924740e-13; 
enH[32] =  1.644420e+01; muH[32] =  1.851040e-07; 
enH[33] =  1.671020e+01; muH[33] =  8.730130e-02; 
enH[34] =  1.698050e+01; muH[34] =  9.602290e-01; 
enH[35] =  1.725510e+01; muH[35] =  9.313640e-01; 
enH[36] =  1.753420e+01; muH[36] =  9.033670e-01; 
enH[37] =  1.781780e+01; muH[37] =  8.762110e-01; 
enH[38] =  1.810600e+01; muH[38] =  8.498730e-01; 
enH[39] =  1.839890e+01; muH[39] =  8.243250e-01; 
enH[40] =  1.869640e+01; muH[40] =  7.995460e-01; 
enH[41] =  1.899880e+01; muH[41] =  7.755110e-01; 
enH[42] =  1.930610e+01; muH[42] =  7.521990e-01; 
enH[43] =  1.961840e+01; muH[43] =  7.291310e-01; 
enH[44] =  1.993570e+01; muH[44] =  7.056420e-01; 
enH[45] =  2.025820e+01; muH[45] =  6.829080e-01; 
enH[46] =  2.058580e+01; muH[46] =  6.609070e-01; 
enH[47] =  2.091880e+01; muH[47] =  6.396150e-01; 
enH[48] =  2.125710e+01; muH[48] =  6.190090e-01; 
enH[49] =  2.160090e+01; muH[49] =  5.990670e-01; 
enH[50] =  2.195030e+01; muH[50] =  5.797670e-01; 
enH[51] =  2.230530e+01; muH[51] =  5.610890e-01; 
enH[52] =  2.266610e+01; muH[52] =  5.430130e-01; 
enH[53] =  2.303270e+01; muH[53] =  5.255190e-01; 
enH[54] =  2.340530e+01; muH[54] =  5.085880e-01; 
enH[55] =  2.378380e+01; muH[55] =  4.922030e-01; 
enH[56] =  2.416850e+01; muH[56] =  4.763470e-01; 
enH[57] =  2.455940e+01; muH[57] =  4.607950e-01; 
enH[58] =  2.495660e+01; muH[58] =  4.455950e-01; 
enH[59] =  2.536030e+01; muH[59] =  4.308970e-01; 
enH[60] =  2.577050e+01; muH[60] =  4.166830e-01; 
enH[61] =  2.618730e+01; muH[61] =  4.029380e-01; 
enH[62] =  2.661090e+01; muH[62] =  3.896470e-01; 
enH[63] =  2.704130e+01; muH[63] =  3.767940e-01; 
enH[64] =  2.747860e+01; muH[64] =  3.643650e-01; 
enH[65] =  2.792310e+01; muH[65] =  3.523460e-01; 
enH[66] =  2.837470e+01; muH[66] =  3.407230e-01; 
enH[67] =  2.883370e+01; muH[67] =  3.294840e-01; 
enH[68] =  2.930000e+01; muH[68] =  3.186160e-01; 
enH[69] =  2.977390e+01; muH[69] =  3.081060e-01; 
enH[70] =  3.025550e+01; muH[70] =  2.985150e-01; 
enH[71] =  3.074490e+01; muH[71] =  2.897180e-01; 
enH[72] =  3.124210e+01; muH[72] =  2.810820e-01; 
enH[73] =  3.174750e+01; muH[73] =  2.726000e-01; 
enH[74] =  3.226090e+01; muH[74] =  2.643770e-01; 
enH[75] =  3.278270e+01; muH[75] =  2.564030e-01; 
enH[76] =  3.331300e+01; muH[76] =  2.486670e-01; 
enH[77] =  3.385180e+01; muH[77] =  2.411630e-01; 
enH[78] =  3.439930e+01; muH[78] =  2.338860e-01; 
enH[79] =  3.495570e+01; muH[79] =  2.268290e-01; 
enH[80] =  3.552110e+01; muH[80] =  2.199600e-01; 
enH[81] =  3.609560e+01; muH[81] =  2.132970e-01; 
enH[82] =  3.667940e+01; muH[82] =  2.068360e-01; 
enH[83] =  3.727270e+01; muH[83] =  2.005670e-01; 
enH[84] =  3.787550e+01; muH[84] =  1.944850e-01; 
enH[85] =  3.848820e+01; muH[85] =  1.885850e-01; 
enH[86] =  3.911070e+01; muH[86] =  1.828640e-01; 
enH[87] =  3.974320e+01; muH[87] =  1.773190e-01; 
enH[88] =  4.038610e+01; muH[88] =  1.719440e-01; 
enH[89] =  4.103930e+01; muH[89] =  1.667250e-01; 
enH[90] =  4.170310e+01; muH[90] =  1.615760e-01; 
enH[91] =  4.237760e+01; muH[91] =  1.565860e-01; 
enH[92] =  4.306300e+01; muH[92] =  1.517500e-01; 
enH[93] =  4.375950e+01; muH[93] =  1.470510e-01; 
enH[94] =  4.446730e+01; muH[94] =  1.424950e-01; 
enH[95] =  4.518650e+01; muH[95] =  1.380780e-01; 
enH[96] =  4.591740e+01; muH[96] =  1.337980e-01; 
enH[97] =  4.666000e+01; muH[97] =  1.296500e-01; 
enH[98] =  4.741470e+01; muH[98] =  1.256340e-01; 
enH[99] =  4.818160e+01; muH[99] =  1.217410e-01; 
enH[100] =  4.896090e+01; muH[100] =  1.179600e-01; 
enH[101] =  4.975280e+01; muH[101] =  1.142970e-01; 
enH[102] =  5.055760e+01; muH[102] =  1.107290e-01; 
enH[103] =  5.137530e+01; muH[103] =  1.072650e-01; 
enH[104] =  5.220620e+01; muH[104] =  1.039100e-01; 
enH[105] =  5.305060e+01; muH[105] =  1.006590e-01; 
enH[106] =  5.390870e+01; muH[106] =  9.751060e-02; 
enH[107] =  5.478060e+01; muH[107] =  9.446020e-02; 
enH[108] =  5.566670e+01; muH[108] =  9.150510e-02; 
enH[109] =  5.656700e+01; muH[109] =  8.864270e-02; 
enH[110] =  5.748200e+01; muH[110] =  8.586970e-02; 
enH[111] =  5.841170e+01; muH[111] =  8.318390e-02; 
enH[112] =  5.935640e+01; muH[112] =  8.058220e-02; 
enH[113] =  6.031650e+01; muH[113] =  7.806140e-02; 
enH[114] =  6.129210e+01; muH[114] =  7.560820e-02; 
enH[115] =  6.228340e+01; muH[115] =  7.320620e-02; 
enH[116] =  6.329080e+01; muH[116] =  7.087020e-02; 
enH[117] =  6.431450e+01; muH[117] =  6.858360e-02; 
enH[118] =  6.535470e+01; muH[118] =  6.637070e-02; 
enH[119] =  6.641180e+01; muH[119] =  6.422890e-02; 
enH[120] =  6.748590e+01; muH[120] =  6.215640e-02; 
enH[121] =  6.857750e+01; muH[121] =  6.015130e-02; 
enH[122] =  6.968670e+01; muH[122] =  5.821040e-02; 
enH[123] =  7.081380e+01; muH[123] =  5.633230e-02; 
enH[124] =  7.195910e+01; muH[124] =  5.451500e-02; 
enH[125] =  7.312300e+01; muH[125] =  5.275660e-02; 
enH[126] =  7.430570e+01; muH[126] =  5.105460e-02; 
enH[127] =  7.550760e+01; muH[127] =  4.940710e-02; 
enH[128] =  7.672890e+01; muH[128] =  4.781290e-02; 
enH[129] =  7.796990e+01; muH[129] =  4.627030e-02; 
enH[130] =  7.923100e+01; muH[130] =  4.477710e-02; 
enH[131] =  8.051250e+01; muH[131] =  4.333200e-02; 
enH[132] =  8.181470e+01; muH[132] =  4.193090e-02; 
enH[133] =  8.313800e+01; muH[133] =  4.057090e-02; 
enH[134] =  8.448270e+01; muH[134] =  3.923480e-02; 
enH[135] =  8.584910e+01; muH[135] =  3.794260e-02; 
enH[136] =  8.723770e+01; muH[136] =  3.669200e-02; 
enH[137] =  8.864870e+01; muH[137] =  3.548280e-02; 
enH[138] =  9.008250e+01; muH[138] =  3.431370e-02; 
enH[139] =  9.153950e+01; muH[139] =  3.318310e-02; 
enH[140] =  9.302010e+01; muH[140] =  3.208970e-02; 
enH[141] =  9.452460e+01; muH[141] =  3.103240e-02; 
enH[142] =  9.605350e+01; muH[142] =  3.001010e-02; 
enH[143] =  9.760710e+01; muH[143] =  2.902140e-02; 
enH[144] =  9.918580e+01; muH[144] =  2.806520e-02; 
enH[145] =  1.007900e+02; muH[145] =  2.713680e-02; 
enH[146] =  1.024200e+02; muH[146] =  2.623740e-02; 
enH[147] =  1.040770e+02; muH[147] =  2.537370e-02; 
enH[148] =  1.057600e+02; muH[148] =  2.453760e-02; 
enH[149] =  1.074710e+02; muH[149] =  2.372900e-02; 
enH[150] =  1.092090e+02; muH[150] =  2.294720e-02; 
enH[151] =  1.109750e+02; muH[151] =  2.219110e-02; 
enH[152] =  1.127700e+02; muH[152] =  2.145980e-02; 
enH[153] =  1.145940e+02; muH[153] =  2.075280e-02; 
enH[154] =  1.164480e+02; muH[154] =  2.006900e-02; 
enH[155] =  1.183310e+02; muH[155] =  1.940770e-02; 
enH[156] =  1.202450e+02; muH[156] =  1.876810e-02; 
enH[157] =  1.221900e+02; muH[157] =  1.814960e-02; 
enH[158] =  1.241660e+02; muH[158] =  1.755030e-02; 
enH[159] =  1.261750e+02; muH[159] =  1.695800e-02; 
enH[160] =  1.282150e+02; muH[160] =  1.638570e-02; 
enH[161] =  1.302890e+02; muH[161] =  1.583280e-02; 
enH[162] =  1.323970e+02; muH[162] =  1.529830e-02; 
enH[163] =  1.345380e+02; muH[163] =  1.478200e-02; 
enH[164] =  1.367140e+02; muH[164] =  1.428320e-02; 
enH[165] =  1.389250e+02; muH[165] =  1.380120e-02; 
enH[166] =  1.411720e+02; muH[166] =  1.333530e-02; 
enH[167] =  1.434560e+02; muH[167] =  1.288510e-02; 
enH[168] =  1.457760e+02; muH[168] =  1.245030e-02; 
enH[169] =  1.481340e+02; muH[169] =  1.203010e-02; 
enH[170] =  1.505300e+02; muH[170] =  1.162400e-02; 
enH[171] =  1.529640e+02; muH[171] =  1.123180e-02; 
enH[172] =  1.554390e+02; muH[172] =  1.085260e-02; 
enH[173] =  1.579530e+02; muH[173] =  1.048630e-02; 
enH[174] =  1.605070e+02; muH[174] =  1.013250e-02; 
enH[175] =  1.631030e+02; muH[175] =  9.789890e-03; 
enH[176] =  1.657420e+02; muH[176] =  9.457980e-03; 
enH[177] =  1.684220e+02; muH[177] =  9.131210e-03; 
enH[178] =  1.711460e+02; muH[178] =  8.815060e-03; 
enH[179] =  1.739150e+02; muH[179] =  8.509790e-03; 
enH[180] =  1.767270e+02; muH[180] =  8.215160e-03; 
enH[181] =  1.795860e+02; muH[181] =  7.930800e-03; 
enH[182] =  1.824910e+02; muH[182] =  7.656300e-03; 
enH[183] =  1.854420e+02; muH[183] =  7.391190e-03; 
enH[184] =  1.884420e+02; muH[184] =  7.135190e-03; 
enH[185] =  1.914890e+02; muH[185] =  6.888050e-03; 
enH[186] =  1.945870e+02; muH[186] =  6.649620e-03; 
enH[187] =  1.977340e+02; muH[187] =  6.419480e-03; 
enH[188] =  2.009320e+02; muH[188] =  6.197310e-03; 
enH[189] =  2.041820e+02; muH[189] =  5.982570e-03; 
enH[190] =  2.074850e+02; muH[190] =  5.775390e-03; 
enH[191] =  2.108400e+02; muH[191] =  5.575530e-03; 
enH[192] =  2.142510e+02; muH[192] =  5.382250e-03; 
enH[193] =  2.177160e+02; muH[193] =  5.195840e-03; 
enH[194] =  2.212370e+02; muH[194] =  5.016080e-03; 
enH[195] =  2.248160e+02; muH[195] =  4.842410e-03; 
enH[196] =  2.284520e+02; muH[196] =  4.674750e-03; 
enH[197] =  2.321470e+02; muH[197] =  4.512870e-03; 
enH[198] =  2.359020e+02; muH[198] =  4.356610e-03; 
enH[199] =  2.397170e+02; muH[199] =  4.205770e-03; 
enH[200] =  2.435950e+02; muH[200] =  4.060140e-03; 
enH[201] =  2.475350e+02; muH[201] =  3.919550e-03; 
enH[202] =  2.515380e+02; muH[202] =  3.783820e-03; 
enH[203] =  2.556070e+02; muH[203] =  3.652770e-03; 
enH[204] =  2.597410e+02; muH[204] =  3.526250e-03; 
enH[205] =  2.639420e+02; muH[205] =  3.404130e-03; 
enH[206] =  2.682110e+02; muH[206] =  3.286240e-03; 
enH[207] =  2.725490e+02; muH[207] =  3.172400e-03; 
enH[208] =  2.769570e+02; muH[208] =  3.062520e-03; 
enH[209] =  2.814370e+02; muH[209] =  2.956450e-03; 
enH[210] =  2.859890e+02; muH[210] =  2.854070e-03; 
enH[211] =  2.906150e+02; muH[211] =  2.755230e-03; 
enH[212] =  2.953150e+02; muH[212] =  2.659840e-03; 
enH[213] =  3.000920e+02; muH[213] =  2.567730e-03; 
enH[214] =  3.049450e+02; muH[214] =  2.478640e-03; 
enH[215] =  3.098780e+02; muH[215] =  2.392410e-03; 
enH[216] =  3.148900e+02; muH[216] =  2.307170e-03; 
enH[217] =  3.199830e+02; muH[217] =  2.224400e-03; 
enH[218] =  3.251580e+02; muH[218] =  2.144600e-03; 
enH[219] =  3.304180e+02; muH[219] =  2.067670e-03; 
enH[220] =  3.357620e+02; muH[220] =  1.993510e-03; 
enH[221] =  3.411920e+02; muH[221] =  1.922060e-03; 
enH[222] =  3.467110e+02; muH[222] =  1.853120e-03; 
enH[223] =  3.523190e+02; muH[223] =  1.786640e-03; 
enH[224] =  3.580170e+02; muH[224] =  1.722530e-03; 
enH[225] =  3.638080e+02; muH[225] =  1.660750e-03; 
enH[226] =  3.696920e+02; muH[226] =  1.601190e-03; 
enH[227] =  3.756720e+02; muH[227] =  1.543770e-03; 
enH[228] =  3.817480e+02; muH[228] =  1.488370e-03; 
enH[229] =  3.879220e+02; muH[229] =  1.434980e-03; 
enH[230] =  3.941970e+02; muH[230] =  1.383560e-03; 
enH[231] =  4.005730e+02; muH[231] =  1.333930e-03; 
enH[232] =  4.070520e+02; muH[232] =  1.286070e-03; 
enH[233] =  4.136350e+02; muH[233] =  1.239940e-03; 
enH[234] =  4.203260e+02; muH[234] =  1.195470e-03; 
enH[235] =  4.271240e+02; muH[235] =  1.152590e-03; 
enH[236] =  4.340320e+02; muH[236] =  1.111240e-03; 
enH[237] =  4.410520e+02; muH[237] =  1.071340e-03; 
enH[238] =  4.481860e+02; muH[238] =  1.032870e-03; 
enH[239] =  4.554350e+02; muH[239] =  9.958120e-04; 
enH[240] =  4.628020e+02; muH[240] =  9.600790e-04; 
enH[241] =  4.702870e+02; muH[241] =  9.256270e-04; 
enH[242] =  4.778940e+02; muH[242] =  8.924160e-04; 
enH[243] =  4.856230e+02; muH[243] =  8.603930e-04; 
enH[244] =  4.934780e+02; muH[244] =  8.295180e-04; 
enH[245] =  5.014590e+02; muH[245] =  7.997510e-04; 
enH[246] =  5.095700e+02; muH[246] =  7.710520e-04; 
enH[247] =  5.178120e+02; muH[247] =  7.433830e-04; 
enH[248] =  5.261870e+02; muH[248] =  7.167140e-04; 
enH[249] =  5.346980e+02; muH[249] =  6.909990e-04; 
enH[250] =  5.433460e+02; muH[250] =  6.662030e-04; 
enH[251] =  5.521340e+02; muH[251] =  6.422990e-04; 
enH[252] =  5.610650e+02; muH[252] =  6.192490e-04; 
enH[253] =  5.701390e+02; muH[253] =  5.970260e-04; 
enH[254] =  5.793610e+02; muH[254] =  5.755970e-04; 
enH[255] =  5.887320e+02; muH[255] =  5.549450e-04; 
enH[256] =  5.982540e+02; muH[256] =  5.350350e-04; 
enH[257] =  6.079300e+02; muH[257] =  5.158320e-04; 
enH[258] =  6.177630e+02; muH[258] =  4.972190e-04; 
enH[259] =  6.277550e+02; muH[259] =  4.791950e-04; 
enH[260] =  6.379080e+02; muH[260] =  4.616570e-04; 
enH[261] =  6.482260e+02; muH[261] =  4.447580e-04; 
enH[262] =  6.587110e+02; muH[262] =  4.284770e-04; 
enH[263] =  6.693650e+02; muH[263] =  4.127920e-04; 
enH[264] =  6.801910e+02; muH[264] =  3.976800e-04; 
enH[265] =  6.911930e+02; muH[265] =  3.831200e-04; 
enH[266] =  7.023720e+02; muH[266] =  3.690940e-04; 
enH[267] =  7.137330e+02; muH[267] =  3.555820e-04; 
enH[268] =  7.252770e+02; muH[268] =  3.425650e-04; 
enH[269] =  7.370080e+02; muH[269] =  3.300260e-04; 
enH[270] =  7.489280e+02; muH[270] =  3.179450e-04; 
enH[271] =  7.610420e+02; muH[271] =  3.063050e-04; 
enH[272] =  7.733510e+02; muH[272] =  2.950820e-04; 
enH[273] =  7.858590e+02; muH[273] =  2.842750e-04; 
enH[274] =  7.985700e+02; muH[274] =  2.738690e-04; 
enH[275] =  8.114860e+02; muH[275] =  2.638520e-04; 
enH[276] =  8.246110e+02; muH[276] =  2.541970e-04; 
enH[277] =  8.379490e+02; muH[277] =  2.448900e-04; 
enH[278] =  8.515020e+02; muH[278] =  2.359240e-04; 
enH[279] =  8.652740e+02; muH[279] =  2.272880e-04; 
enH[280] =  8.792690e+02; muH[280] =  2.189690e-04; 
enH[281] =  8.934910e+02; muH[281] =  2.109480e-04; 
enH[282] =  9.079430e+02; muH[282] =  2.032230e-04; 
enH[283] =  9.226280e+02; muH[283] =  1.957820e-04; 
enH[284] =  9.375510e+02; muH[284] =  1.886140e-04; 
enH[285] =  9.527150e+02; muH[285] =  1.817090e-04; 
enH[286] =  9.681240e+02; muH[286] =  1.750570e-04; 
enH[287] =  9.837830e+02; muH[287] =  1.686490e-04; 
enH[288] =  9.996950e+02; muH[288] =  1.624750e-04; 
enH[289] =  1.015860e+03; muH[289] =  1.565270e-04; 
enH[290] =  1.032290e+03; muH[290] =  1.507960e-04; 
enH[291] =  1.048990e+03; muH[291] =  1.452750e-04; 
enH[292] =  1.065960e+03; muH[292] =  1.399560e-04; 
enH[293] =  1.083200e+03; muH[293] =  1.348330e-04; 
enH[294] =  1.100720e+03; muH[294] =  1.298970e-04; 
enH[295] =  1.118520e+03; muH[295] =  1.251400e-04; 
enH[296] =  1.136610e+03; muH[296] =  1.205590e-04; 
enH[297] =  1.155000e+03; muH[297] =  1.161450e-04; 
enH[298] =  1.173680e+03; muH[298] =  1.118940e-04; 
enH[299] =  1.192660e+03; muH[299] =  1.077980e-04; 
enH[300] =  1.211950e+03; muH[300] =  1.038510e-04; 
enH[301] =  1.231550e+03; muH[301] =  1.000480e-04; 
enH[302] =  1.251470e+03; muH[302] =  9.634310e-05; 
enH[303] =  1.271720e+03; muH[303] =  9.276580e-05; 
enH[304] =  1.292290e+03; muH[304] =  8.931360e-05; 
enH[305] =  1.313190e+03; muH[305] =  8.598970e-05; 
enH[306] =  1.334430e+03; muH[306] =  8.278900e-05; 
enH[307] =  1.356010e+03; muH[307] =  7.970650e-05; 
enH[308] =  1.377940e+03; muH[308] =  7.673990e-05; 
enH[309] =  1.400230e+03; muH[309] =  7.388450e-05; 
enH[310] =  1.422880e+03; muH[310] =  7.113670e-05; 
enH[311] =  1.445890e+03; muH[311] =  6.848900e-05; 
enH[312] =  1.469280e+03; muH[312] =  6.593940e-05; 
enH[313] =  1.493040e+03; muH[313] =  6.348460e-05; 
enH[314] =  1.517190e+03; muH[314] =  6.112220e-05; 
enH[315] =  1.541730e+03; muH[315] =  5.884970e-05; 
enH[316] =  1.566670e+03; muH[316] =  5.666480e-05; 
enH[317] =  1.592010e+03; muH[317] =  5.456500e-05; 
enH[318] =  1.617760e+03; muH[318] =  5.254390e-05; 
enH[319] =  1.643920e+03; muH[319] =  5.059840e-05; 
enH[320] =  1.670510e+03; muH[320] =  4.872290e-05; 
enH[321] =  1.697530e+03; muH[321] =  4.691710e-05; 
enH[322] =  1.724990e+03; muH[322] =  4.517910e-05; 
enH[323] =  1.752890e+03; muH[323] =  4.349880e-05; 
enH[324] =  1.781240e+03; muH[324] =  4.187880e-05; 
enH[325] =  1.810050e+03; muH[325] =  4.031750e-05; 
enH[326] =  1.839320e+03; muH[326] =  3.881440e-05; 
enH[327] =  1.869070e+03; muH[327] =  3.736720e-05; 
enH[328] =  1.899300e+03; muH[328] =  3.597380e-05; 
enH[329] =  1.930020e+03; muH[329] =  3.463250e-05; 
enH[330] =  1.961240e+03; muH[330] =  3.334130e-05; 
enH[331] =  1.992960e+03; muH[331] =  3.209830e-05; 
enH[332] =  2.025200e+03; muH[332] =  3.091170e-05; 
enH[333] =  2.057950e+03; muH[333] =  2.975480e-05; 
enH[334] =  2.091240e+03; muH[334] =  2.863990e-05; 
enH[335] =  2.125060e+03; muH[335] =  2.756560e-05; 
enH[336] =  2.159430e+03; muH[336] =  2.653110e-05; 
enH[337] =  2.194360e+03; muH[337] =  2.553400e-05; 
enH[338] =  2.229850e+03; muH[338] =  2.457370e-05; 
enH[339] =  2.265920e+03; muH[339] =  2.364880e-05; 
enH[340] =  2.302570e+03; muH[340] =  2.275780e-05; 
enH[341] =  2.339810e+03; muH[341] =  2.189960e-05; 
enH[342] =  2.377660e+03; muH[342] =  2.107280e-05; 
enH[343] =  2.416110e+03; muH[343] =  2.027630e-05; 
enH[344] =  2.455190e+03; muH[344] =  1.950990e-05; 
enH[345] =  2.494900e+03; muH[345] =  1.877110e-05; 
enH[346] =  2.535260e+03; muH[346] =  1.806010e-05; 
enH[347] =  2.576260e+03; muH[347] =  1.737520e-05; 
enH[348] =  2.617930e+03; muH[348] =  1.671570e-05; 
enH[349] =  2.660270e+03; muH[349] =  1.608080e-05; 
enH[350] =  2.703300e+03; muH[350] =  1.546910e-05; 
enH[351] =  2.747030e+03; muH[351] =  1.488060e-05; 
enH[352] =  2.791460e+03; muH[352] =  1.431370e-05; 
enH[353] =  2.836610e+03; muH[353] =  1.376850e-05; 
enH[354] =  2.882490e+03; muH[354] =  1.324350e-05; 
enH[355] =  2.929110e+03; muH[355] =  1.273770e-05; 
enH[356] =  2.976480e+03; muH[356] =  1.225100e-05; 
enH[357] =  3.024630e+03; muH[357] =  1.178310e-05; 
enH[358] =  3.073550e+03; muH[358] =  1.133260e-05; 
enH[359] =  3.123260e+03; muH[359] =  1.089870e-05; 
enH[360] =  3.173780e+03; muH[360] =  1.048150e-05; 
enH[361] =  3.225110e+03; muH[361] =  1.008030e-05; 
enH[362] =  3.277270e+03; muH[362] =  9.694070e-06; 
enH[363] =  3.330280e+03; muH[363] =  9.322800e-06; 
enH[364] =  3.384150e+03; muH[364] =  8.965700e-06; 
enH[365] =  3.438880e+03; muH[365] =  8.621590e-06; 
enH[366] =  3.494500e+03; muH[366] =  8.291170e-06; 
enH[367] =  3.551020e+03; muH[367] =  7.973480e-06; 
enH[368] =  3.608460e+03; muH[368] =  7.667920e-06; 
enH[369] =  3.666820e+03; muH[369] =  7.374170e-06; 
enH[370] =  3.726130e+03; muH[370] =  7.091770e-06; 
enH[371] =  3.786400e+03; muH[371] =  6.820400e-06; 
enH[372] =  3.847640e+03; muH[372] =  6.559550e-06; 
enH[373] =  3.909870e+03; muH[373] =  6.308870e-06; 
enH[374] =  3.973110e+03; muH[374] =  6.068020e-06; 
enH[375] =  4.037380e+03; muH[375] =  5.834170e-06; 
enH[376] =  4.102680e+03; muH[376] =  5.611820e-06; 
enH[377] =  4.169030e+03; muH[377] =  5.398040e-06; 
enH[378] =  4.236460e+03; muH[378] =  5.192620e-06; 
enH[379] =  4.304980e+03; muH[379] =  4.995050e-06; 
enH[380] =  4.374620e+03; muH[380] =  4.805120e-06; 
enH[381] =  4.445370e+03; muH[381] =  4.622520e-06; 
enH[382] =  4.517270e+03; muH[382] =  4.446930e-06; 
enH[383] =  4.590330e+03; muH[383] =  4.278240e-06; 
enH[384] =  4.664580e+03; muH[384] =  4.115920e-06; 
enH[385] =  4.740030e+03; muH[385] =  3.959800e-06; 
enH[386] =  4.816690e+03; muH[386] =  3.809710e-06; 
enH[387] =  4.894600e+03; muH[387] =  3.665340e-06; 
enH[388] =  4.973770e+03; muH[388] =  3.526560e-06; 
enH[389] =  5.054210e+03; muH[389] =  3.393080e-06; 
enH[390] =  5.135960e+03; muH[390] =  3.264670e-06; 
enH[391] =  5.219030e+03; muH[391] =  3.141220e-06; 
enH[392] =  5.303440e+03; muH[392] =  3.022400e-06; 
enH[393] =  5.389220e+03; muH[393] =  2.908110e-06; 
enH[394] =  5.476390e+03; muH[394] =  2.798200e-06; 
enH[395] =  5.564970e+03; muH[395] =  2.692520e-06; 
enH[396] =  5.654980e+03; muH[396] =  2.590800e-06; 
enH[397] =  5.746440e+03; muH[397] =  2.492920e-06; 
enH[398] =  5.839390e+03; muH[398] =  2.398780e-06; 
enH[399] =  5.933830e+03; muH[399] =  2.308250e-06; 
enH[400] =  6.029810e+03; muH[400] =  2.221110e-06; 
enH[401] =  6.127330e+03; muH[401] =  2.137250e-06; 
enH[402] =  6.226440e+03; muH[402] =  2.056620e-06; 
enH[403] =  6.327150e+03; muH[403] =  1.978950e-06; 
enH[404] =  6.429480e+03; muH[404] =  1.904270e-06; 
enH[405] =  6.533480e+03; muH[405] =  1.832480e-06; 
enH[406] =  6.639150e+03; muH[406] =  1.763290e-06; 
enH[407] =  6.746540e+03; muH[407] =  1.696700e-06; 
enH[408] =  6.855650e+03; muH[408] =  1.632680e-06; 
enH[409] =  6.966540e+03; muH[409] =  1.571050e-06; 
enH[410] =  7.079220e+03; muH[410] =  1.511730e-06; 
enH[411] =  7.193720e+03; muH[411] =  1.454640e-06; 
enH[412] =  7.310070e+03; muH[412] =  1.399700e-06; 
enH[413] =  7.428310e+03; muH[413] =  1.346830e-06; 
enH[414] =  7.548450e+03; muH[414] =  1.295940e-06; 
enH[415] =  7.670540e+03; muH[415] =  1.246960e-06; 
enH[416] =  7.794610e+03; muH[416] =  1.199810e-06; 
enH[417] =  7.920680e+03; muH[417] =  1.154430e-06; 
enH[418] =  8.048790e+03; muH[418] =  1.110760e-06; 
enH[419] =  8.178980e+03; muH[419] =  1.068720e-06; 
enH[420] =  8.311260e+03; muH[420] =  1.028260e-06; 
enH[421] =  8.445690e+03; muH[421] =  9.893030e-07; 
enH[422] =  8.582290e+03; muH[422] =  9.518140e-07; 
enH[423] =  8.721110e+03; muH[423] =  9.157340e-07; 
enH[424] =  8.862160e+03; muH[424] =  8.809920e-07; 
enH[425] =  9.005500e+03; muH[425] =  8.475630e-07; 
enH[426] =  9.151160e+03; muH[426] =  8.153590e-07; 
enH[427] =  9.299170e+03; muH[427] =  7.843880e-07; 
enH[428] =  9.449580e+03; muH[428] =  7.545530e-07; 
enH[429] =  9.602420e+03; muH[429] =  7.258470e-07; 
enH[430] =  9.757730e+03; muH[430] =  6.982230e-07; 
enH[431] =  9.915550e+03; muH[431] =  6.716260e-07; 
enH[432] =  1.007590e+04; muH[432] =  6.460750e-07; 
enH[433] =  1.023890e+04; muH[433] =  6.214810e-07; 
enH[434] =  1.040450e+04; muH[434] =  5.978060e-07; 
enH[435] =  1.057280e+04; muH[435] =  5.750160e-07; 
enH[436] =  1.074380e+04; muH[436] =  5.530810e-07; 
enH[437] =  1.091760e+04; muH[437] =  5.319650e-07; 
enH[438] =  1.109420e+04; muH[438] =  5.116410e-07; 
enH[439] =  1.127360e+04; muH[439] =  4.920770e-07; 
enH[440] =  1.145590e+04; muH[440] =  4.732480e-07; 
enH[441] =  1.164120e+04; muH[441] =  4.551230e-07; 
enH[442] =  1.182950e+04; muH[442] =  4.376780e-07; 
enH[443] =  1.202080e+04; muH[443] =  4.208890e-07; 
enH[444] =  1.221530e+04; muH[444] =  4.047280e-07; 
enH[445] =  1.241280e+04; muH[445] =  3.891760e-07; 
enH[446] =  1.261360e+04; muH[446] =  3.742070e-07; 
enH[447] =  1.281760e+04; muH[447] =  3.598020e-07; 
enH[448] =  1.302500e+04; muH[448] =  3.459390e-07; 
enH[449] =  1.323560e+04; muH[449] =  3.325980e-07; 
enH[450] =  1.344970e+04; muH[450] =  3.197600e-07; 
enH[451] =  1.366720e+04; muH[451] =  3.074060e-07; 
enH[452] =  1.388830e+04; muH[452] =  2.955190e-07; 
enH[453] =  1.411290e+04; muH[453] =  2.840810e-07; 
enH[454] =  1.434120e+04; muH[454] =  2.730760e-07; 
enH[455] =  1.457310e+04; muH[455] =  2.624880e-07; 
enH[456] =  1.480890e+04; muH[456] =  2.523020e-07; 
enH[457] =  1.504840e+04; muH[457] =  2.425020e-07; 
enH[458] =  1.529180e+04; muH[458] =  2.330750e-07; 
enH[459] =  1.553910e+04; muH[459] =  2.240070e-07; 
enH[460] =  1.579040e+04; muH[460] =  2.152840e-07; 
enH[461] =  1.604580e+04; muH[461] =  2.068940e-07; 
enH[462] =  1.630540e+04; muH[462] =  1.988250e-07; 
enH[463] =  1.656910e+04; muH[463] =  1.910650e-07; 
enH[464] =  1.683710e+04; muH[464] =  1.836030e-07; 
enH[465] =  1.710940e+04; muH[465] =  1.764280e-07; 
enH[466] =  1.738610e+04; muH[466] =  1.695280e-07; 
enH[467] =  1.766740e+04; muH[467] =  1.628950e-07; 
enH[468] =  1.795310e+04; muH[468] =  1.565180e-07; 
enH[469] =  1.824350e+04; muH[469] =  1.503870e-07; 
enH[470] =  1.853860e+04; muH[470] =  1.444950e-07; 
enH[471] =  1.883840e+04; muH[471] =  1.388310e-07; 
enH[472] =  1.914310e+04; muH[472] =  1.333890e-07; 
enH[473] =  1.945270e+04; muH[473] =  1.281590e-07; 
enH[474] =  1.976740e+04; muH[474] =  1.231330e-07; 
enH[475] =  2.008710e+04; muH[475] =  1.183810e-07; 
enH[476] =  2.041200e+04; muH[476] =  1.137440e-07; 
enH[477] =  2.074210e+04; muH[477] =  1.092920e-07; 
enH[478] =  2.107760e+04; muH[478] =  1.050190e-07; 
enH[479] =  2.141850e+04; muH[479] =  1.009170e-07; 
enH[480] =  2.176500e+04; muH[480] =  9.697830e-08; 
enH[481] =  2.211700e+04; muH[481] =  9.319720e-08; 
enH[482] =  2.247470e+04; muH[482] =  8.956690e-08; 
enH[483] =  2.283820e+04; muH[483] =  8.608100e-08; 
enH[484] =  2.320760e+04; muH[484] =  8.273380e-08; 
enH[485] =  2.358300e+04; muH[485] =  7.951950e-08; 
enH[486] =  2.396440e+04; muH[486] =  7.643280e-08; 
enH[487] =  2.435200e+04; muH[487] =  7.346830e-08; 
enH[488] =  2.474590e+04; muH[488] =  7.062130e-08; 
enH[489] =  2.514620e+04; muH[489] =  6.788670e-08; 
enH[490] =  2.555290e+04; muH[490] =  6.526020e-08; 
enH[491] =  2.596620e+04; muH[491] =  6.273720e-08; 
enH[492] =  2.638610e+04; muH[492] =  6.031360e-08; 
enH[493] =  2.681290e+04; muH[493] =  5.798540e-08; 
enH[494] =  2.724660e+04; muH[494] =  5.574870e-08; 
enH[495] =  2.768730e+04; muH[495] =  5.359990e-08; 
enH[496] =  2.813510e+04; muH[496] =  5.153520e-08; 
enH[497] =  2.859020e+04; muH[497] =  4.955160e-08; 
enH[498] =  2.905260e+04; muH[498] =  4.764550e-08; 
enH[499] =  2.952250e+04; muH[499] =  4.581390e-08; 
enH[500] =  3.000000e+04; muH[500] =  4.405390e-08; 

% and now we convert them to en/mu columns
% (we want the energy grid to be in keV for acisabs
%  to avoid some multiplication)
muH *= 4.208e7 / ( enH * awH );
enH /= 1.0e3;

% species = N
enN = Float_Type[503];
muN = Float_Type[503];

% The columns here are the values from n.nff
enN[0] =  1.000000e+01; muN[0] =  6.778470e-01; 
enN[1] =  1.016170e+01; muN[1] =  6.896310e-01; 
enN[2] =  1.032610e+01; muN[2] =  7.016200e-01; 
enN[3] =  1.049310e+01; muN[3] =  7.138170e-01; 
enN[4] =  1.066280e+01; muN[4] =  7.262270e-01; 
enN[5] =  1.083530e+01; muN[5] =  7.388520e-01; 
enN[6] =  1.101060e+01; muN[6] =  7.516970e-01; 
enN[7] =  1.118860e+01; muN[7] =  7.658390e-01; 
enN[8] =  1.136960e+01; muN[8] =  7.860260e-01; 
enN[9] =  1.155350e+01; muN[9] =  8.067430e-01; 
enN[10] =  1.174040e+01; muN[10] =  8.280080e-01; 
enN[11] =  1.193030e+01; muN[11] =  8.498330e-01; 
enN[12] =  1.212320e+01; muN[12] =  8.722330e-01; 
enN[13] =  1.231930e+01; muN[13] =  8.952230e-01; 
enN[14] =  1.251860e+01; muN[14] =  9.192800e-01; 
enN[15] =  1.272100e+01; muN[15] =  9.546220e-01; 
enN[16] =  1.292680e+01; muN[16] =  9.913230e-01; 
enN[17] =  1.313590e+01; muN[17] =  1.029430e+00; 
enN[18] =  1.334830e+01; muN[18] =  1.069010e+00; 
enN[19] =  1.356420e+01; muN[19] =  1.110110e+00; 
enN[20] =  1.378360e+01; muN[20] =  1.152790e+00; 
enN[21] =  1.400660e+01; muN[21] =  1.224320e+00; 
enN[22] =  1.423310e+01; muN[22] =  1.304240e+00; 
enN[23] =  1.446330e+01; muN[23] =  1.389380e+00; 
enN[24] =  1.469730e+01; muN[24] =  1.480080e+00; 
enN[25] =  1.493500e+01; muN[25] =  1.576700e+00; 
enN[26] =  1.517650e+01; muN[26] =  1.686550e+00; 
enN[27] =  1.542200e+01; muN[27] =  1.806250e+00; 
enN[28] =  1.567140e+01; muN[28] =  1.934450e+00; 
enN[29] =  1.592490e+01; muN[29] =  2.071750e+00; 
enN[30] =  1.618250e+01; muN[30] =  2.218790e+00; 
enN[31] =  1.644420e+01; muN[31] =  2.376270e+00; 
enN[32] =  1.671020e+01; muN[32] =  2.544920e+00; 
enN[33] =  1.698050e+01; muN[33] =  2.680910e+00; 
enN[34] =  1.725510e+01; muN[34] =  2.820440e+00; 
enN[35] =  1.753420e+01; muN[35] =  2.967240e+00; 
enN[36] =  1.781780e+01; muN[36] =  3.121690e+00; 
enN[37] =  1.810600e+01; muN[37] =  3.284170e+00; 
enN[38] =  1.839890e+01; muN[38] =  3.455110e+00; 
enN[39] =  1.869640e+01; muN[39] =  3.634940e+00; 
enN[40] =  1.899880e+01; muN[40] =  3.791960e+00; 
enN[41] =  1.930610e+01; muN[41] =  3.940690e+00; 
enN[42] =  1.961840e+01; muN[42] =  4.095260e+00; 
enN[43] =  1.993570e+01; muN[43] =  4.255880e+00; 
enN[44] =  2.025820e+01; muN[44] =  4.422810e+00; 
enN[45] =  2.058580e+01; muN[45] =  4.541920e+00; 
enN[46] =  2.091880e+01; muN[46] =  4.621720e+00; 
enN[47] =  2.125710e+01; muN[47] =  4.702920e+00; 
enN[48] =  2.160090e+01; muN[48] =  4.785540e+00; 
enN[49] =  2.195030e+01; muN[49] =  4.869630e+00; 
enN[50] =  2.230530e+01; muN[50] =  4.923150e+00; 
enN[51] =  2.266610e+01; muN[51] =  4.947570e+00; 
enN[52] =  2.303270e+01; muN[52] =  4.972110e+00; 
enN[53] =  2.340530e+01; muN[53] =  4.996780e+00; 
enN[54] =  2.378380e+01; muN[54] =  5.021560e+00; 
enN[55] =  2.416850e+01; muN[55] =  5.021380e+00; 
enN[56] =  2.455940e+01; muN[56] =  4.992150e+00; 
enN[57] =  2.495660e+01; muN[57] =  4.963090e+00; 
enN[58] =  2.536030e+01; muN[58] =  4.934200e+00; 
enN[59] =  2.577050e+01; muN[59] =  4.905470e+00; 
enN[60] =  2.618730e+01; muN[60] =  4.876920e+00; 
enN[61] =  2.661090e+01; muN[61] =  4.848530e+00; 
enN[62] =  2.704130e+01; muN[62] =  4.820300e+00; 
enN[63] =  2.747860e+01; muN[63] =  4.772520e+00; 
enN[64] =  2.792310e+01; muN[64] =  4.696870e+00; 
enN[65] =  2.837470e+01; muN[65] =  4.622410e+00; 
enN[66] =  2.883370e+01; muN[66] =  4.549130e+00; 
enN[67] =  2.930000e+01; muN[67] =  4.477010e+00; 
enN[68] =  2.977390e+01; muN[68] =  4.406040e+00; 
enN[69] =  3.025550e+01; muN[69] =  4.336190e+00; 
enN[70] =  3.074490e+01; muN[70] =  4.267450e+00; 
enN[71] =  3.124210e+01; muN[71] =  4.199800e+00; 
enN[72] =  3.174750e+01; muN[72] =  4.133220e+00; 
enN[73] =  3.226090e+01; muN[73] =  4.067700e+00; 
enN[74] =  3.278270e+01; muN[74] =  4.007110e+00; 
enN[75] =  3.331300e+01; muN[75] =  3.952230e+00; 
enN[76] =  3.385180e+01; muN[76] =  3.898100e+00; 
enN[77] =  3.439930e+01; muN[77] =  3.844710e+00; 
enN[78] =  3.495570e+01; muN[78] =  3.792050e+00; 
enN[79] =  3.552110e+01; muN[79] =  3.740120e+00; 
enN[80] =  3.609560e+01; muN[80] =  3.688890e+00; 
enN[81] =  3.667940e+01; muN[81] =  3.638370e+00; 
enN[82] =  3.720000e+01; muN[82] =  3.594570e+00; 
enN[83] =  3.740000e+01; muN[83] =  3.641040e+00; 
enN[84] =  3.787550e+01; muN[84] =  3.601500e+00; 
enN[85] =  3.848820e+01; muN[85] =  3.551910e+00; 
enN[86] =  3.911070e+01; muN[86] =  3.503000e+00; 
enN[87] =  3.974320e+01; muN[87] =  3.454760e+00; 
enN[88] =  4.038610e+01; muN[88] =  3.407190e+00; 
enN[89] =  4.103930e+01; muN[89] =  3.360270e+00; 
enN[90] =  4.170310e+01; muN[90] =  3.314000e+00; 
enN[91] =  4.237760e+01; muN[91] =  3.268360e+00; 
enN[92] =  4.306300e+01; muN[92] =  3.223360e+00; 
enN[93] =  4.375950e+01; muN[93] =  3.178970e+00; 
enN[94] =  4.446730e+01; muN[94] =  3.135200e+00; 
enN[95] =  4.518650e+01; muN[95] =  3.092030e+00; 
enN[96] =  4.591740e+01; muN[96] =  3.049450e+00; 
enN[97] =  4.666000e+01; muN[97] =  3.002830e+00; 
enN[98] =  4.741470e+01; muN[98] =  2.954420e+00; 
enN[99] =  4.818160e+01; muN[99] =  2.906780e+00; 
enN[100] =  4.896090e+01; muN[100] =  2.859920e+00; 
enN[101] =  4.975280e+01; muN[101] =  2.813810e+00; 
enN[102] =  5.055760e+01; muN[102] =  2.768440e+00; 
enN[103] =  5.137530e+01; muN[103] =  2.723810e+00; 
enN[104] =  5.220620e+01; muN[104] =  2.679890e+00; 
enN[105] =  5.305060e+01; muN[105] =  2.636680e+00; 
enN[106] =  5.390870e+01; muN[106] =  2.658410e+00; 
enN[107] =  5.478060e+01; muN[107] =  2.631010e+00; 
enN[108] =  5.566670e+01; muN[108] =  2.598030e+00; 
enN[109] =  5.656700e+01; muN[109] =  2.565460e+00; 
enN[110] =  5.748200e+01; muN[110] =  2.533280e+00; 
enN[111] =  5.841170e+01; muN[111] =  2.501500e+00; 
enN[112] =  5.935640e+01; muN[112] =  2.469130e+00; 
enN[113] =  6.031650e+01; muN[113] =  2.435570e+00; 
enN[114] =  6.129210e+01; muN[114] =  2.401640e+00; 
enN[115] =  6.228340e+01; muN[115] =  2.366300e+00; 
enN[116] =  6.329080e+01; muN[116] =  2.331460e+00; 
enN[117] =  6.431450e+01; muN[117] =  2.297100e+00; 
enN[118] =  6.535470e+01; muN[118] =  2.263270e+00; 
enN[119] =  6.641180e+01; muN[119] =  2.229970e+00; 
enN[120] =  6.748590e+01; muN[120] =  2.197160e+00; 
enN[121] =  6.857750e+01; muN[121] =  2.164820e+00; 
enN[122] =  6.968670e+01; muN[122] =  2.131960e+00; 
enN[123] =  7.081380e+01; muN[123] =  2.098990e+00; 
enN[124] =  7.195910e+01; muN[124] =  2.065210e+00; 
enN[125] =  7.312300e+01; muN[125] =  2.031720e+00; 
enN[126] =  7.430570e+01; muN[126] =  1.998780e+00; 
enN[127] =  7.550760e+01; muN[127] =  1.966350e+00; 
enN[128] =  7.672890e+01; muN[128] =  1.934460e+00; 
enN[129] =  7.796990e+01; muN[129] =  1.903090e+00; 
enN[130] =  7.923100e+01; muN[130] =  1.871560e+00; 
enN[131] =  8.051250e+01; muN[131] =  1.840530e+00; 
enN[132] =  8.181470e+01; muN[132] =  1.809190e+00; 
enN[133] =  8.313800e+01; muN[133] =  1.777880e+00; 
enN[134] =  8.448270e+01; muN[134] =  1.747110e+00; 
enN[135] =  8.584910e+01; muN[135] =  1.716870e+00; 
enN[136] =  8.723770e+01; muN[136] =  1.687150e+00; 
enN[137] =  8.864870e+01; muN[137] =  1.657840e+00; 
enN[138] =  9.008250e+01; muN[138] =  1.628910e+00; 
enN[139] =  9.153950e+01; muN[139] =  1.600310e+00; 
enN[140] =  9.302010e+01; muN[140] =  1.571900e+00; 
enN[141] =  9.452460e+01; muN[141] =  1.544000e+00; 
enN[142] =  9.605350e+01; muN[142] =  1.516610e+00; 
enN[143] =  9.760710e+01; muN[143] =  1.489690e+00; 
enN[144] =  9.918580e+01; muN[144] =  1.463260e+00; 
enN[145] =  1.007900e+02; muN[145] =  1.436280e+00; 
enN[146] =  1.024200e+02; muN[146] =  1.408870e+00; 
enN[147] =  1.040770e+02; muN[147] =  1.382220e+00; 
enN[148] =  1.057600e+02; muN[148] =  1.356060e+00; 
enN[149] =  1.074710e+02; muN[149] =  1.330390e+00; 
enN[150] =  1.092090e+02; muN[150] =  1.305190e+00; 
enN[151] =  1.109750e+02; muN[151] =  1.280480e+00; 
enN[152] =  1.127700e+02; muN[152] =  1.256240e+00; 
enN[153] =  1.145940e+02; muN[153] =  1.232450e+00; 
enN[154] =  1.164480e+02; muN[154] =  1.209140e+00; 
enN[155] =  1.183310e+02; muN[155] =  1.186270e+00; 
enN[156] =  1.202450e+02; muN[156] =  1.163810e+00; 
enN[157] =  1.221900e+02; muN[157] =  1.141770e+00; 
enN[158] =  1.241660e+02; muN[158] =  1.120140e+00; 
enN[159] =  1.261750e+02; muN[159] =  1.098940e+00; 
enN[160] =  1.282150e+02; muN[160] =  1.078150e+00; 
enN[161] =  1.302890e+02; muN[161] =  1.057600e+00; 
enN[162] =  1.323970e+02; muN[162] =  1.036510e+00; 
enN[163] =  1.345380e+02; muN[163] =  1.015820e+00; 
enN[164] =  1.367140e+02; muN[164] =  9.955490e-01; 
enN[165] =  1.389250e+02; muN[165] =  9.756930e-01; 
enN[166] =  1.411720e+02; muN[166] =  9.562410e-01; 
enN[167] =  1.434560e+02; muN[167] =  9.371740e-01; 
enN[168] =  1.457760e+02; muN[168] =  9.184710e-01; 
enN[169] =  1.481340e+02; muN[169] =  9.001550e-01; 
enN[170] =  1.505300e+02; muN[170] =  8.822120e-01; 
enN[171] =  1.529640e+02; muN[171] =  8.646100e-01; 
enN[172] =  1.554390e+02; muN[172] =  8.473620e-01; 
enN[173] =  1.579530e+02; muN[173] =  8.304580e-01; 
enN[174] =  1.605070e+02; muN[174] =  8.135670e-01; 
enN[175] =  1.631030e+02; muN[175] =  7.955480e-01; 
enN[176] =  1.657420e+02; muN[176] =  7.762600e-01; 
enN[177] =  1.684220e+02; muN[177] =  7.574380e-01; 
enN[178] =  1.711460e+02; muN[178] =  7.390740e-01; 
enN[179] =  1.739150e+02; muN[179] =  7.211540e-01; 
enN[180] =  1.767270e+02; muN[180] =  7.036700e-01; 
enN[181] =  1.795860e+02; muN[181] =  6.866080e-01; 
enN[182] =  1.824910e+02; muN[182] =  6.699610e-01; 
enN[183] =  1.854420e+02; muN[183] =  6.537180e-01; 
enN[184] =  1.884420e+02; muN[184] =  6.378680e-01; 
enN[185] =  1.914890e+02; muN[185] =  6.224020e-01; 
enN[186] =  1.945870e+02; muN[186] =  6.073120e-01; 
enN[187] =  1.977340e+02; muN[187] =  5.925870e-01; 
enN[188] =  2.009320e+02; muN[188] =  5.782190e-01; 
enN[189] =  2.041820e+02; muN[189] =  5.642000e-01; 
enN[190] =  2.074850e+02; muN[190] =  5.505210e-01; 
enN[191] =  2.108400e+02; muN[191] =  5.371730e-01; 
enN[192] =  2.142510e+02; muN[192] =  5.241490e-01; 
enN[193] =  2.177160e+02; muN[193] =  5.114400e-01; 
enN[194] =  2.212370e+02; muN[194] =  4.989970e-01; 
enN[195] =  2.248160e+02; muN[195] =  4.862680e-01; 
enN[196] =  2.284520e+02; muN[196] =  4.738650e-01; 
enN[197] =  2.321470e+02; muN[197] =  4.617780e-01; 
enN[198] =  2.359020e+02; muN[198] =  4.499990e-01; 
enN[199] =  2.397170e+02; muN[199] =  4.385210e-01; 
enN[200] =  2.435950e+02; muN[200] =  4.273350e-01; 
enN[201] =  2.475350e+02; muN[201] =  4.164350e-01; 
enN[202] =  2.515380e+02; muN[202] =  4.058120e-01; 
enN[203] =  2.556070e+02; muN[203] =  3.954610e-01; 
enN[204] =  2.597410e+02; muN[204] =  3.853740e-01; 
enN[205] =  2.639420e+02; muN[205] =  3.755440e-01; 
enN[206] =  2.682110e+02; muN[206] =  3.659650e-01; 
enN[207] =  2.725490e+02; muN[207] =  3.566300e-01; 
enN[208] =  2.769570e+02; muN[208] =  3.475330e-01; 
enN[209] =  2.814370e+02; muN[209] =  3.386680e-01; 
enN[210] =  2.859890e+02; muN[210] =  3.300300e-01; 
enN[211] =  2.906150e+02; muN[211] =  3.216110e-01; 
enN[212] =  2.953150e+02; muN[212] =  3.134080e-01; 
enN[213] =  3.000920e+02; muN[213] =  3.054130e-01; 
enN[214] =  3.049450e+02; muN[214] =  2.976230e-01; 
enN[215] =  3.098780e+02; muN[215] =  2.900320e-01; 
enN[216] =  3.148900e+02; muN[216] =  2.809690e-01; 
enN[217] =  3.199830e+02; muN[217] =  2.710970e-01; 
enN[218] =  3.251580e+02; muN[218] =  2.615710e-01; 
enN[219] =  3.304180e+02; muN[219] =  2.523800e-01; 
enN[220] =  3.357620e+02; muN[220] =  2.435120e-01; 
enN[221] =  3.411920e+02; muN[221] =  2.349560e-01; 
enN[222] =  3.467110e+02; muN[222] =  2.267000e-01; 
enN[223] =  3.523190e+02; muN[223] =  2.187340e-01; 
enN[224] =  3.580170e+02; muN[224] =  2.110490e-01; 
enN[225] =  3.638080e+02; muN[225] =  2.036440e-01; 
enN[226] =  3.696920e+02; muN[226] =  1.966140e-01; 
enN[227] =  3.756720e+02; muN[227] =  1.898270e-01; 
enN[228] =  3.817480e+02; muN[228] =  1.832750e-01; 
enN[229] =  3.879220e+02; muN[229] =  1.769480e-01; 
enN[230] =  3.941970e+02; muN[230] =  1.708400e-01; 
enN[231] =  4.005730e+02; muN[231] =  1.649430e-01; 
enN[232] =  4.070520e+02; muN[232] =  1.592490e-01; 
enN[233] =  4.098000e+02; muN[233] =  1.569190e-01; 
enN[234] =  4.100000e+02; muN[234] =  4.191450e+00; 
enN[235] =  4.136350e+02; muN[235] =  4.143080e+00; 
enN[236] =  4.203260e+02; muN[236] =  4.056570e+00; 
enN[237] =  4.271240e+02; muN[237] =  3.971870e+00; 
enN[238] =  4.340320e+02; muN[238] =  3.888940e+00; 
enN[239] =  4.410520e+02; muN[239] =  3.807740e+00; 
enN[240] =  4.481860e+02; muN[240] =  3.728240e+00; 
enN[241] =  4.554350e+02; muN[241] =  3.650390e+00; 
enN[242] =  4.628020e+02; muN[242] =  3.574170e+00; 
enN[243] =  4.702870e+02; muN[243] =  3.499540e+00; 
enN[244] =  4.778940e+02; muN[244] =  3.426480e+00; 
enN[245] =  4.856230e+02; muN[245] =  3.354930e+00; 
enN[246] =  4.934780e+02; muN[246] =  3.284880e+00; 
enN[247] =  5.014590e+02; muN[247] =  3.216290e+00; 
enN[248] =  5.095700e+02; muN[248] =  3.149140e+00; 
enN[249] =  5.178120e+02; muN[249] =  3.083390e+00; 
enN[250] =  5.261870e+02; muN[250] =  3.019010e+00; 
enN[251] =  5.346980e+02; muN[251] =  2.955970e+00; 
enN[252] =  5.433460e+02; muN[252] =  2.894250e+00; 
enN[253] =  5.521340e+02; muN[253] =  2.833820e+00; 
enN[254] =  5.610650e+02; muN[254] =  2.774650e+00; 
enN[255] =  5.701390e+02; muN[255] =  2.710650e+00; 
enN[256] =  5.793610e+02; muN[256] =  2.643960e+00; 
enN[257] =  5.887320e+02; muN[257] =  2.578900e+00; 
enN[258] =  5.982540e+02; muN[258] =  2.515450e+00; 
enN[259] =  6.079300e+02; muN[259] =  2.453550e+00; 
enN[260] =  6.177630e+02; muN[260] =  2.393190e+00; 
enN[261] =  6.277550e+02; muN[261] =  2.334300e+00; 
enN[262] =  6.379080e+02; muN[262] =  2.276870e+00; 
enN[263] =  6.482260e+02; muN[263] =  2.220840e+00; 
enN[264] =  6.587110e+02; muN[264] =  2.166200e+00; 
enN[265] =  6.693650e+02; muN[265] =  2.112900e+00; 
enN[266] =  6.801910e+02; muN[266] =  2.060910e+00; 
enN[267] =  6.911930e+02; muN[267] =  2.010210e+00; 
enN[268] =  7.023720e+02; muN[268] =  1.960740e+00; 
enN[269] =  7.137330e+02; muN[269] =  1.912500e+00; 
enN[270] =  7.252770e+02; muN[270] =  1.865440e+00; 
enN[271] =  7.370080e+02; muN[271] =  1.819540e+00; 
enN[272] =  7.489280e+02; muN[272] =  1.774780e+00; 
enN[273] =  7.610420e+02; muN[273] =  1.731110e+00; 
enN[274] =  7.733510e+02; muN[274] =  1.687550e+00; 
enN[275] =  7.858590e+02; muN[275] =  1.643480e+00; 
enN[276] =  7.985700e+02; muN[276] =  1.599800e+00; 
enN[277] =  8.114860e+02; muN[277] =  1.557270e+00; 
enN[278] =  8.246110e+02; muN[278] =  1.515880e+00; 
enN[279] =  8.379490e+02; muN[279] =  1.475590e+00; 
enN[280] =  8.515020e+02; muN[280] =  1.436720e+00; 
enN[281] =  8.652740e+02; muN[281] =  1.399240e+00; 
enN[282] =  8.792690e+02; muN[282] =  1.363050e+00; 
enN[283] =  8.934910e+02; muN[283] =  1.327780e+00; 
enN[284] =  9.079430e+02; muN[284] =  1.293440e+00; 
enN[285] =  9.226280e+02; muN[285] =  1.260000e+00; 
enN[286] =  9.375510e+02; muN[286] =  1.226690e+00; 
enN[287] =  9.527150e+02; muN[287] =  1.194100e+00; 
enN[288] =  9.681240e+02; muN[288] =  1.162230e+00; 
enN[289] =  9.837830e+02; muN[289] =  1.131210e+00; 
enN[290] =  9.996950e+02; muN[290] =  1.101040e+00; 
enN[291] =  1.015860e+03; muN[291] =  1.071690e+00; 
enN[292] =  1.032290e+03; muN[292] =  1.043400e+00; 
enN[293] =  1.048990e+03; muN[293] =  1.015390e+00; 
enN[294] =  1.065960e+03; muN[294] =  9.876350e-01; 
enN[295] =  1.083200e+03; muN[295] =  9.606440e-01; 
enN[296] =  1.100720e+03; muN[296] =  9.343820e-01; 
enN[297] =  1.118520e+03; muN[297] =  9.088280e-01; 
enN[298] =  1.136610e+03; muN[298] =  8.839670e-01; 
enN[299] =  1.155000e+03; muN[299] =  8.597920e-01; 
enN[300] =  1.173680e+03; muN[300] =  8.362890e-01; 
enN[301] =  1.192660e+03; muN[301] =  8.134260e-01; 
enN[302] =  1.211950e+03; muN[302] =  7.911890e-01; 
enN[303] =  1.231550e+03; muN[303] =  7.695610e-01; 
enN[304] =  1.251470e+03; muN[304] =  7.482200e-01; 
enN[305] =  1.271720e+03; muN[305] =  7.271540e-01; 
enN[306] =  1.292290e+03; muN[306] =  7.062990e-01; 
enN[307] =  1.313190e+03; muN[307] =  6.860310e-01; 
enN[308] =  1.334430e+03; muN[308] =  6.663480e-01; 
enN[309] =  1.356010e+03; muN[309] =  6.472370e-01; 
enN[310] =  1.377940e+03; muN[310] =  6.286400e-01; 
enN[311] =  1.400230e+03; muN[311] =  6.105910e-01; 
enN[312] =  1.422880e+03; muN[312] =  5.930890e-01; 
enN[313] =  1.445890e+03; muN[313] =  5.760610e-01; 
enN[314] =  1.469280e+03; muN[314] =  5.595050e-01; 
enN[315] =  1.493040e+03; muN[315] =  5.434070e-01; 
enN[316] =  1.517190e+03; muN[316] =  5.275860e-01; 
enN[317] =  1.541730e+03; muN[317] =  5.122130e-01; 
enN[318] =  1.566670e+03; muN[318] =  4.973100e-01; 
enN[319] =  1.592010e+03; muN[319] =  4.828310e-01; 
enN[320] =  1.617760e+03; muN[320] =  4.687720e-01; 
enN[321] =  1.643920e+03; muN[321] =  4.551200e-01; 
enN[322] =  1.670510e+03; muN[322] =  4.418670e-01; 
enN[323] =  1.697530e+03; muN[323] =  4.290000e-01; 
enN[324] =  1.724990e+03; muN[324] =  4.165090e-01; 
enN[325] =  1.752890e+03; muN[325] =  4.042010e-01; 
enN[326] =  1.781240e+03; muN[326] =  3.921980e-01; 
enN[327] =  1.810050e+03; muN[327] =  3.805080e-01; 
enN[328] =  1.839320e+03; muN[328] =  3.691690e-01; 
enN[329] =  1.869070e+03; muN[329] =  3.581670e-01; 
enN[330] =  1.899300e+03; muN[330] =  3.474920e-01; 
enN[331] =  1.930020e+03; muN[331] =  3.371340e-01; 
enN[332] =  1.961240e+03; muN[332] =  3.270860e-01; 
enN[333] =  1.992960e+03; muN[333] =  3.173400e-01; 
enN[334] =  2.025200e+03; muN[334] =  3.079970e-01; 
enN[335] =  2.057950e+03; muN[335] =  2.987330e-01; 
enN[336] =  2.091240e+03; muN[336] =  2.897250e-01; 
enN[337] =  2.125060e+03; muN[337] =  2.809750e-01; 
enN[338] =  2.159430e+03; muN[338] =  2.724690e-01; 
enN[339] =  2.194360e+03; muN[339] =  2.642100e-01; 
enN[340] =  2.229850e+03; muN[340] =  2.561850e-01; 
enN[341] =  2.265920e+03; muN[341] =  2.483940e-01; 
enN[342] =  2.302570e+03; muN[342] =  2.408220e-01; 
enN[343] =  2.339810e+03; muN[343] =  2.334710e-01; 
enN[344] =  2.377660e+03; muN[344] =  2.263260e-01; 
enN[345] =  2.416110e+03; muN[345] =  2.193900e-01; 
enN[346] =  2.455190e+03; muN[346] =  2.126560e-01; 
enN[347] =  2.494900e+03; muN[347] =  2.061200e-01; 
enN[348] =  2.535260e+03; muN[348] =  1.997720e-01; 
enN[349] =  2.576260e+03; muN[349] =  1.936080e-01; 
enN[350] =  2.617930e+03; muN[350] =  1.876300e-01; 
enN[351] =  2.660270e+03; muN[351] =  1.818160e-01; 
enN[352] =  2.703300e+03; muN[352] =  1.761820e-01; 
enN[353] =  2.747030e+03; muN[353] =  1.707110e-01; 
enN[354] =  2.791460e+03; muN[354] =  1.653990e-01; 
enN[355] =  2.836610e+03; muN[355] =  1.602500e-01; 
enN[356] =  2.882490e+03; muN[356] =  1.552480e-01; 
enN[357] =  2.929110e+03; muN[357] =  1.503890e-01; 
enN[358] =  2.976480e+03; muN[358] =  1.456800e-01; 
enN[359] =  3.024630e+03; muN[359] =  1.411100e-01; 
enN[360] =  3.073550e+03; muN[360] =  1.366730e-01; 
enN[361] =  3.123260e+03; muN[361] =  1.323700e-01; 
enN[362] =  3.173780e+03; muN[362] =  1.281950e-01; 
enN[363] =  3.225110e+03; muN[363] =  1.241530e-01; 
enN[364] =  3.277270e+03; muN[364] =  1.202190e-01; 
enN[365] =  3.330280e+03; muN[365] =  1.164110e-01; 
enN[366] =  3.384150e+03; muN[366] =  1.127170e-01; 
enN[367] =  3.438880e+03; muN[367] =  1.091340e-01; 
enN[368] =  3.494500e+03; muN[368] =  1.056580e-01; 
enN[369] =  3.551020e+03; muN[369] =  1.022860e-01; 
enN[370] =  3.608460e+03; muN[370] =  9.901610e-02; 
enN[371] =  3.666820e+03; muN[371] =  9.584530e-02; 
enN[372] =  3.726130e+03; muN[372] =  9.277100e-02; 
enN[373] =  3.786400e+03; muN[373] =  8.978900e-02; 
enN[374] =  3.847640e+03; muN[374] =  8.689810e-02; 
enN[375] =  3.909870e+03; muN[375] =  8.409510e-02; 
enN[376] =  3.973110e+03; muN[376] =  8.137770e-02; 
enN[377] =  4.037380e+03; muN[377] =  7.873120e-02; 
enN[378] =  4.102680e+03; muN[378] =  7.617940e-02; 
enN[379] =  4.169030e+03; muN[379] =  7.370620e-02; 
enN[380] =  4.236460e+03; muN[380] =  7.131200e-02; 
enN[381] =  4.304980e+03; muN[381] =  6.899150e-02; 
enN[382] =  4.374620e+03; muN[382] =  6.674400e-02; 
enN[383] =  4.445370e+03; muN[383] =  6.456730e-02; 
enN[384] =  4.517270e+03; muN[384] =  6.245880e-02; 
enN[385] =  4.590330e+03; muN[385] =  6.041670e-02; 
enN[386] =  4.664580e+03; muN[386] =  5.843900e-02; 
enN[387] =  4.740030e+03; muN[387] =  5.652300e-02; 
enN[388] =  4.816690e+03; muN[388] =  5.466880e-02; 
enN[389] =  4.894600e+03; muN[389] =  5.287190e-02; 
enN[390] =  4.973770e+03; muN[390] =  5.113270e-02; 
enN[391] =  5.054210e+03; muN[391] =  4.944840e-02; 
enN[392] =  5.135960e+03; muN[392] =  4.781780e-02; 
enN[393] =  5.219030e+03; muN[393] =  4.623920e-02; 
enN[394] =  5.303440e+03; muN[394] =  4.470960e-02; 
enN[395] =  5.389220e+03; muN[395] =  4.322990e-02; 
enN[396] =  5.476390e+03; muN[396] =  4.179650e-02; 
enN[397] =  5.564970e+03; muN[397] =  4.040850e-02; 
enN[398] =  5.654980e+03; muN[398] =  3.906650e-02; 
enN[399] =  5.746440e+03; muN[399] =  3.776630e-02; 
enN[400] =  5.839390e+03; muN[400] =  3.650790e-02; 
enN[401] =  5.933830e+03; muN[401] =  3.528900e-02; 
enN[402] =  6.029810e+03; muN[402] =  3.411030e-02; 
enN[403] =  6.127330e+03; muN[403] =  3.296850e-02; 
enN[404] =  6.226440e+03; muN[404] =  3.186400e-02; 
enN[405] =  6.327150e+03; muN[405] =  3.079560e-02; 
enN[406] =  6.429480e+03; muN[406] =  2.976180e-02; 
enN[407] =  6.533480e+03; muN[407] =  2.876090e-02; 
enN[408] =  6.639150e+03; muN[408] =  2.779160e-02; 
enN[409] =  6.746540e+03; muN[409] =  2.685360e-02; 
enN[410] =  6.855650e+03; muN[410] =  2.594750e-02; 
enN[411] =  6.966540e+03; muN[411] =  2.507000e-02; 
enN[412] =  7.079220e+03; muN[412] =  2.422180e-02; 
enN[413] =  7.193720e+03; muN[413] =  2.340030e-02; 
enN[414] =  7.310070e+03; muN[414] =  2.260540e-02; 
enN[415] =  7.428310e+03; muN[415] =  2.183670e-02; 
enN[416] =  7.548450e+03; muN[416] =  2.109320e-02; 
enN[417] =  7.670540e+03; muN[417] =  2.037390e-02; 
enN[418] =  7.794610e+03; muN[418] =  1.967820e-02; 
enN[419] =  7.920680e+03; muN[419] =  1.900530e-02; 
enN[420] =  8.048790e+03; muN[420] =  1.835460e-02; 
enN[421] =  8.178980e+03; muN[421] =  1.772510e-02; 
enN[422] =  8.311260e+03; muN[422] =  1.711630e-02; 
enN[423] =  8.445690e+03; muN[423] =  1.652790e-02; 
enN[424] =  8.582290e+03; muN[424] =  1.595850e-02; 
enN[425] =  8.721110e+03; muN[425] =  1.540810e-02; 
enN[426] =  8.862160e+03; muN[426] =  1.487600e-02; 
enN[427] =  9.005500e+03; muN[427] =  1.436140e-02; 
enN[428] =  9.151160e+03; muN[428] =  1.386400e-02; 
enN[429] =  9.299170e+03; muN[429] =  1.338310e-02; 
enN[430] =  9.449580e+03; muN[430] =  1.291810e-02; 
enN[431] =  9.602420e+03; muN[431] =  1.246890e-02; 
enN[432] =  9.757730e+03; muN[432] =  1.203440e-02; 
enN[433] =  9.915550e+03; muN[433] =  1.161460e-02; 
enN[434] =  1.007590e+04; muN[434] =  1.120880e-02; 
enN[435] =  1.023890e+04; muN[435] =  1.081660e-02; 
enN[436] =  1.040450e+04; muN[436] =  1.043760e-02; 
enN[437] =  1.057280e+04; muN[437] =  1.007140e-02; 
enN[438] =  1.074380e+04; muN[438] =  9.717530e-03; 
enN[439] =  1.091760e+04; muN[439] =  9.375600e-03; 
enN[440] =  1.109420e+04; muN[440] =  9.045240e-03; 
enN[441] =  1.127360e+04; muN[441] =  8.726070e-03; 
enN[442] =  1.145590e+04; muN[442] =  8.417760e-03; 
enN[443] =  1.164120e+04; muN[443] =  8.119930e-03; 
enN[444] =  1.182950e+04; muN[444] =  7.832230e-03; 
enN[445] =  1.202080e+04; muN[445] =  7.554380e-03; 
enN[446] =  1.221530e+04; muN[446] =  7.286010e-03; 
enN[447] =  1.241280e+04; muN[447] =  7.026850e-03; 
enN[448] =  1.261360e+04; muN[448] =  6.776590e-03; 
enN[449] =  1.281760e+04; muN[449] =  6.534940e-03; 
enN[450] =  1.302500e+04; muN[450] =  6.301620e-03; 
enN[451] =  1.323560e+04; muN[451] =  6.076360e-03; 
enN[452] =  1.344970e+04; muN[452] =  5.858900e-03; 
enN[453] =  1.366720e+04; muN[453] =  5.648980e-03; 
enN[454] =  1.388830e+04; muN[454] =  5.446360e-03; 
enN[455] =  1.411290e+04; muN[455] =  5.250800e-03; 
enN[456] =  1.434120e+04; muN[456] =  5.062070e-03; 
enN[457] =  1.457310e+04; muN[457] =  4.879940e-03; 
enN[458] =  1.480890e+04; muN[458] =  4.704210e-03; 
enN[459] =  1.504840e+04; muN[459] =  4.534650e-03; 
enN[460] =  1.529180e+04; muN[460] =  4.371070e-03; 
enN[461] =  1.553910e+04; muN[461] =  4.213280e-03; 
enN[462] =  1.579040e+04; muN[462] =  4.061060e-03; 
enN[463] =  1.604580e+04; muN[463] =  3.914260e-03; 
enN[464] =  1.630540e+04; muN[464] =  3.772680e-03; 
enN[465] =  1.656910e+04; muN[465] =  3.636160e-03; 
enN[466] =  1.683710e+04; muN[466] =  3.504530e-03; 
enN[467] =  1.710940e+04; muN[467] =  3.377620e-03; 
enN[468] =  1.738610e+04; muN[468] =  3.255280e-03; 
enN[469] =  1.766740e+04; muN[469] =  3.137360e-03; 
enN[470] =  1.795310e+04; muN[470] =  3.023700e-03; 
enN[471] =  1.824350e+04; muN[471] =  2.914180e-03; 
enN[472] =  1.853860e+04; muN[472] =  2.808640e-03; 
enN[473] =  1.883840e+04; muN[473] =  2.706970e-03; 
enN[474] =  1.914310e+04; muN[474] =  2.609020e-03; 
enN[475] =  1.945270e+04; muN[475] =  2.514670e-03; 
enN[476] =  1.976740e+04; muN[476] =  2.423800e-03; 
enN[477] =  2.008710e+04; muN[477] =  2.337060e-03; 
enN[478] =  2.041200e+04; muN[478] =  2.252770e-03; 
enN[479] =  2.074210e+04; muN[479] =  2.171580e-03; 
enN[480] =  2.107760e+04; muN[480] =  2.093370e-03; 
enN[481] =  2.141850e+04; muN[481] =  2.018030e-03; 
enN[482] =  2.176500e+04; muN[482] =  1.945450e-03; 
enN[483] =  2.211700e+04; muN[483] =  1.875520e-03; 
enN[484] =  2.247470e+04; muN[484] =  1.808150e-03; 
enN[485] =  2.283820e+04; muN[485] =  1.743240e-03; 
enN[486] =  2.320760e+04; muN[486] =  1.680690e-03; 
enN[487] =  2.358300e+04; muN[487] =  1.620430e-03; 
enN[488] =  2.396440e+04; muN[488] =  1.562350e-03; 
enN[489] =  2.435200e+04; muN[489] =  1.506390e-03; 
enN[490] =  2.474590e+04; muN[490] =  1.452460e-03; 
enN[491] =  2.514620e+04; muN[491] =  1.400480e-03; 
enN[492] =  2.555290e+04; muN[492] =  1.350390e-03; 
enN[493] =  2.596620e+04; muN[493] =  1.302120e-03; 
enN[494] =  2.638610e+04; muN[494] =  1.255590e-03; 
enN[495] =  2.681290e+04; muN[495] =  1.210740e-03; 
enN[496] =  2.724660e+04; muN[496] =  1.167510e-03; 
enN[497] =  2.768730e+04; muN[497] =  1.125840e-03; 
enN[498] =  2.813510e+04; muN[498] =  1.085670e-03; 
enN[499] =  2.859020e+04; muN[499] =  1.046950e-03; 
enN[500] =  2.905260e+04; muN[500] =  1.009620e-03; 
enN[501] =  2.952250e+04; muN[501] =  9.736290e-04; 
enN[502] =  3.000000e+04; muN[502] =  9.389340e-04; 

% and now we convert them to en/mu columns
% (we want the energy grid to be in keV for acisabs
%  to avoid some multiplication)
muN *= 4.208e7 / ( enN * awN );
enN /= 1.0e3;

% species = C
enC = Float_Type[502];
muC = Float_Type[502];

% The columns here are the values from c.nff
enC[0] =  1.000000e+01; muC[0] =  8.068850e-01; 
enC[1] =  1.016170e+01; muC[1] =  8.515220e-01; 
enC[2] =  1.032610e+01; muC[2] =  8.986280e-01; 
enC[3] =  1.049310e+01; muC[3] =  9.483410e-01; 
enC[4] =  1.066280e+01; muC[4] =  1.000800e+00; 
enC[5] =  1.083530e+01; muC[5] =  1.057550e+00; 
enC[6] =  1.101060e+01; muC[6] =  1.121670e+00; 
enC[7] =  1.118860e+01; muC[7] =  1.189680e+00; 
enC[8] =  1.136960e+01; muC[8] =  1.261810e+00; 
enC[9] =  1.155350e+01; muC[9] =  1.338320e+00; 
enC[10] =  1.174040e+01; muC[10] =  1.419460e+00; 
enC[11] =  1.193030e+01; muC[11] =  1.505530e+00; 
enC[12] =  1.212320e+01; muC[12] =  1.589710e+00; 
enC[13] =  1.231930e+01; muC[13] =  1.678200e+00; 
enC[14] =  1.251860e+01; muC[14] =  1.771610e+00; 
enC[15] =  1.272100e+01; muC[15] =  1.870230e+00; 
enC[16] =  1.292680e+01; muC[16] =  1.974330e+00; 
enC[17] =  1.313590e+01; muC[17] =  2.084230e+00; 
enC[18] =  1.334830e+01; muC[18] =  2.200250e+00; 
enC[19] =  1.356420e+01; muC[19] =  2.292350e+00; 
enC[20] =  1.378360e+01; muC[20] =  2.382870e+00; 
enC[21] =  1.400660e+01; muC[21] =  2.476960e+00; 
enC[22] =  1.423310e+01; muC[22] =  2.574770e+00; 
enC[23] =  1.446330e+01; muC[23] =  2.676450e+00; 
enC[24] =  1.469730e+01; muC[24] =  2.771140e+00; 
enC[25] =  1.493500e+01; muC[25] =  2.846050e+00; 
enC[26] =  1.517650e+01; muC[26] =  2.922980e+00; 
enC[27] =  1.542200e+01; muC[27] =  3.001990e+00; 
enC[28] =  1.567140e+01; muC[28] =  3.083130e+00; 
enC[29] =  1.592490e+01; muC[29] =  3.166470e+00; 
enC[30] =  1.618250e+01; muC[30] =  3.242130e+00; 
enC[31] =  1.644420e+01; muC[31] =  3.287780e+00; 
enC[32] =  1.671020e+01; muC[32] =  3.334080e+00; 
enC[33] =  1.698050e+01; muC[33] =  3.381030e+00; 
enC[34] =  1.725510e+01; muC[34] =  3.428630e+00; 
enC[35] =  1.753420e+01; muC[35] =  3.476910e+00; 
enC[36] =  1.781780e+01; muC[36] =  3.520410e+00; 
enC[37] =  1.810600e+01; muC[37] =  3.542770e+00; 
enC[38] =  1.839890e+01; muC[38] =  3.565280e+00; 
enC[39] =  1.869640e+01; muC[39] =  3.587930e+00; 
enC[40] =  1.899880e+01; muC[40] =  3.610720e+00; 
enC[41] =  1.930610e+01; muC[41] =  3.633660e+00; 
enC[42] =  1.961840e+01; muC[42] =  3.656740e+00; 
enC[43] =  1.993570e+01; muC[43] =  3.679970e+00; 
enC[44] =  2.025820e+01; muC[44] =  3.703350e+00; 
enC[45] =  2.058580e+01; muC[45] =  3.694890e+00; 
enC[46] =  2.091880e+01; muC[46] =  3.672040e+00; 
enC[47] =  2.125710e+01; muC[47] =  3.649340e+00; 
enC[48] =  2.160090e+01; muC[48] =  3.626770e+00; 
enC[49] =  2.195030e+01; muC[49] =  3.604340e+00; 
enC[50] =  2.230530e+01; muC[50] =  3.582050e+00; 
enC[51] =  2.266610e+01; muC[51] =  3.559900e+00; 
enC[52] =  2.303270e+01; muC[52] =  3.533200e+00; 
enC[53] =  2.340530e+01; muC[53] =  3.478920e+00; 
enC[54] =  2.378380e+01; muC[54] =  3.425470e+00; 
enC[55] =  2.416850e+01; muC[55] =  3.372850e+00; 
enC[56] =  2.455940e+01; muC[56] =  3.321030e+00; 
enC[57] =  2.495660e+01; muC[57] =  3.270000e+00; 
enC[58] =  2.536030e+01; muC[58] =  3.219770e+00; 
enC[59] =  2.577050e+01; muC[59] =  3.170300e+00; 
enC[60] =  2.618730e+01; muC[60] =  3.120960e+00; 
enC[61] =  2.661090e+01; muC[61] =  3.063500e+00; 
enC[62] =  2.704130e+01; muC[62] =  3.007080e+00; 
enC[63] =  2.747860e+01; muC[63] =  2.951710e+00; 
enC[64] =  2.792310e+01; muC[64] =  2.897360e+00; 
enC[65] =  2.837470e+01; muC[65] =  2.844010e+00; 
enC[66] =  2.883370e+01; muC[66] =  2.791640e+00; 
enC[67] =  2.930000e+01; muC[67] =  2.740240e+00; 
enC[68] =  2.977390e+01; muC[68] =  2.689780e+00; 
enC[69] =  3.025550e+01; muC[69] =  2.636180e+00; 
enC[70] =  3.074490e+01; muC[70] =  2.583600e+00; 
enC[71] =  3.124210e+01; muC[71] =  2.532060e+00; 
enC[72] =  3.174750e+01; muC[72] =  2.481550e+00; 
enC[73] =  3.226090e+01; muC[73] =  2.432050e+00; 
enC[74] =  3.278270e+01; muC[74] =  2.383540e+00; 
enC[75] =  3.331300e+01; muC[75] =  2.335990e+00; 
enC[76] =  3.385180e+01; muC[76] =  2.289400e+00; 
enC[77] =  3.439930e+01; muC[77] =  2.243730e+00; 
enC[78] =  3.495570e+01; muC[78] =  2.198980e+00; 
enC[79] =  3.552110e+01; muC[79] =  2.157990e+00; 
enC[80] =  3.609560e+01; muC[80] =  2.118270e+00; 
enC[81] =  3.667940e+01; muC[81] =  2.079280e+00; 
enC[82] =  3.727270e+01; muC[82] =  2.041010e+00; 
enC[83] =  3.787550e+01; muC[83] =  2.003440e+00; 
enC[84] =  3.848820e+01; muC[84] =  1.966570e+00; 
enC[85] =  3.911070e+01; muC[85] =  1.930370e+00; 
enC[86] =  3.974320e+01; muC[86] =  1.890110e+00; 
enC[87] =  4.038610e+01; muC[87] =  1.846500e+00; 
enC[88] =  4.103930e+01; muC[88] =  1.803900e+00; 
enC[89] =  4.170310e+01; muC[89] =  1.762280e+00; 
enC[90] =  4.237760e+01; muC[90] =  1.721630e+00; 
enC[91] =  4.306300e+01; muC[91] =  1.681910e+00; 
enC[92] =  4.375950e+01; muC[92] =  1.643100e+00; 
enC[93] =  4.446730e+01; muC[93] =  1.605200e+00; 
enC[94] =  4.518650e+01; muC[94] =  1.569170e+00; 
enC[95] =  4.591740e+01; muC[95] =  1.543760e+00; 
enC[96] =  4.666000e+01; muC[96] =  1.518760e+00; 
enC[97] =  4.741470e+01; muC[97] =  1.494170e+00; 
enC[98] =  4.818160e+01; muC[98] =  1.469980e+00; 
enC[99] =  4.896090e+01; muC[99] =  1.446180e+00; 
enC[100] =  4.975280e+01; muC[100] =  1.422760e+00; 
enC[101] =  5.055760e+01; muC[101] =  1.399720e+00; 
enC[102] =  5.137530e+01; muC[102] =  1.377050e+00; 
enC[103] =  5.220620e+01; muC[103] =  1.354750e+00; 
enC[104] =  5.305060e+01; muC[104] =  1.332820e+00; 
enC[105] =  5.390870e+01; muC[105] =  1.315460e+00; 
enC[106] =  5.478060e+01; muC[106] =  1.301680e+00; 
enC[107] =  5.566670e+01; muC[107] =  1.288040e+00; 
enC[108] =  5.656700e+01; muC[108] =  1.274540e+00; 
enC[109] =  5.748200e+01; muC[109] =  1.261180e+00; 
enC[110] =  5.841170e+01; muC[110] =  1.247970e+00; 
enC[111] =  5.935640e+01; muC[111] =  1.234890e+00; 
enC[112] =  6.031650e+01; muC[112] =  1.221950e+00; 
enC[113] =  6.129210e+01; muC[113] =  1.206540e+00; 
enC[114] =  6.228340e+01; muC[114] =  1.184980e+00; 
enC[115] =  6.329080e+01; muC[115] =  1.163810e+00; 
enC[116] =  6.431450e+01; muC[116] =  1.143010e+00; 
enC[117] =  6.535470e+01; muC[117] =  1.122580e+00; 
enC[118] =  6.641180e+01; muC[118] =  1.102520e+00; 
enC[119] =  6.748590e+01; muC[119] =  1.082820e+00; 
enC[120] =  6.857750e+01; muC[120] =  1.063470e+00; 
enC[121] =  6.968670e+01; muC[121] =  1.044470e+00; 
enC[122] =  7.081380e+01; muC[122] =  1.025800e+00; 
enC[123] =  7.195910e+01; muC[123] =  1.007470e+00; 
enC[124] =  7.312300e+01; muC[124] =  9.894700e-01; 
enC[125] =  7.430570e+01; muC[125] =  9.717880e-01; 
enC[126] =  7.550760e+01; muC[126] =  9.544230e-01; 
enC[127] =  7.672890e+01; muC[127] =  9.373690e-01; 
enC[128] =  7.796990e+01; muC[128] =  9.206170e-01; 
enC[129] =  7.923100e+01; muC[129] =  9.041670e-01; 
enC[130] =  8.051250e+01; muC[130] =  8.880090e-01; 
enC[131] =  8.181470e+01; muC[131] =  8.721410e-01; 
enC[132] =  8.313800e+01; muC[132] =  8.565560e-01; 
enC[133] =  8.448270e+01; muC[133] =  8.412500e-01; 
enC[134] =  8.584910e+01; muC[134] =  8.262170e-01; 
enC[135] =  8.723770e+01; muC[135] =  8.114530e-01; 
enC[136] =  8.864870e+01; muC[136] =  7.969520e-01; 
enC[137] =  9.008250e+01; muC[137] =  7.827110e-01; 
enC[138] =  9.153950e+01; muC[138] =  7.687240e-01; 
enC[139] =  9.302010e+01; muC[139] =  7.549870e-01; 
enC[140] =  9.452460e+01; muC[140] =  7.414960e-01; 
enC[141] =  9.605350e+01; muC[141] =  7.282460e-01; 
enC[142] =  9.760710e+01; muC[142] =  7.152320e-01; 
enC[143] =  9.918580e+01; muC[143] =  7.024510e-01; 
enC[144] =  1.007900e+02; muC[144] =  6.898980e-01; 
enC[145] =  1.024200e+02; muC[145] =  6.775700e-01; 
enC[146] =  1.040770e+02; muC[146] =  6.654620e-01; 
enC[147] =  1.057600e+02; muC[147] =  6.535700e-01; 
enC[148] =  1.074710e+02; muC[148] =  6.418910e-01; 
enC[149] =  1.092090e+02; muC[149] =  6.304210e-01; 
enC[150] =  1.109750e+02; muC[150] =  6.191560e-01; 
enC[151] =  1.127700e+02; muC[151] =  6.080910e-01; 
enC[152] =  1.145940e+02; muC[152] =  5.972240e-01; 
enC[153] =  1.164480e+02; muC[153] =  5.865520e-01; 
enC[154] =  1.183310e+02; muC[154] =  5.760710e-01; 
enC[155] =  1.202450e+02; muC[155] =  5.657770e-01; 
enC[156] =  1.221900e+02; muC[156] =  5.556660e-01; 
enC[157] =  1.241660e+02; muC[157] =  5.457370e-01; 
enC[158] =  1.261750e+02; muC[158] =  5.359850e-01; 
enC[159] =  1.282150e+02; muC[159] =  5.264060e-01; 
enC[160] =  1.302890e+02; muC[160] =  5.170000e-01; 
enC[161] =  1.323970e+02; muC[161] =  5.077610e-01; 
enC[162] =  1.345380e+02; muC[162] =  4.974570e-01; 
enC[163] =  1.367140e+02; muC[163] =  4.873610e-01; 
enC[164] =  1.389250e+02; muC[164] =  4.774690e-01; 
enC[165] =  1.411720e+02; muC[165] =  4.677770e-01; 
enC[166] =  1.434560e+02; muC[166] =  4.582830e-01; 
enC[167] =  1.457760e+02; muC[167] =  4.489810e-01; 
enC[168] =  1.481340e+02; muC[168] =  4.398680e-01; 
enC[169] =  1.505300e+02; muC[169] =  4.309410e-01; 
enC[170] =  1.529640e+02; muC[170] =  4.221930e-01; 
enC[171] =  1.554390e+02; muC[171] =  4.136240e-01; 
enC[172] =  1.579530e+02; muC[172] =  4.052290e-01; 
enC[173] =  1.605070e+02; muC[173] =  3.970040e-01; 
enC[174] =  1.631030e+02; muC[174] =  3.889460e-01; 
enC[175] =  1.657420e+02; muC[175] =  3.810520e-01; 
enC[176] =  1.684220e+02; muC[176] =  3.733180e-01; 
enC[177] =  1.711460e+02; muC[177] =  3.657410e-01; 
enC[178] =  1.739150e+02; muC[178] =  3.583170e-01; 
enC[179] =  1.767270e+02; muC[179] =  3.510450e-01; 
enC[180] =  1.795860e+02; muC[180] =  3.439190e-01; 
enC[181] =  1.824910e+02; muC[181] =  3.369390e-01; 
enC[182] =  1.854420e+02; muC[182] =  3.301000e-01; 
enC[183] =  1.884420e+02; muC[183] =  3.209950e-01; 
enC[184] =  1.914890e+02; muC[184] =  3.120960e-01; 
enC[185] =  1.945870e+02; muC[185] =  3.034430e-01; 
enC[186] =  1.977340e+02; muC[186] =  2.950310e-01; 
enC[187] =  2.009320e+02; muC[187] =  2.868510e-01; 
enC[188] =  2.041820e+02; muC[188] =  2.788980e-01; 
enC[189] =  2.074850e+02; muC[189] =  2.711660e-01; 
enC[190] =  2.108400e+02; muC[190] =  2.636480e-01; 
enC[191] =  2.142510e+02; muC[191] =  2.563380e-01; 
enC[192] =  2.177160e+02; muC[192] =  2.492310e-01; 
enC[193] =  2.212370e+02; muC[193] =  2.425160e-01; 
enC[194] =  2.248160e+02; muC[194] =  2.363810e-01; 
enC[195] =  2.284520e+02; muC[195] =  2.304020e-01; 
enC[196] =  2.321470e+02; muC[196] =  2.245730e-01; 
enC[197] =  2.359020e+02; muC[197] =  2.188920e-01; 
enC[198] =  2.397170e+02; muC[198] =  2.133550e-01; 
enC[199] =  2.435950e+02; muC[199] =  2.079580e-01; 
enC[200] =  2.475350e+02; muC[200] =  2.021790e-01; 
enC[201] =  2.515380e+02; muC[201] =  1.946830e-01; 
enC[202] =  2.556070e+02; muC[202] =  1.874650e-01; 
enC[203] =  2.597410e+02; muC[203] =  1.805140e-01; 
enC[204] =  2.639420e+02; muC[204] =  1.738210e-01; 
enC[205] =  2.682110e+02; muC[205] =  1.673770e-01; 
enC[206] =  2.725490e+02; muC[206] =  1.611710e-01; 
enC[207] =  2.769570e+02; muC[207] =  1.551950e-01; 
enC[208] =  2.814370e+02; muC[208] =  1.494410e-01; 
enC[209] =  2.841000e+02; muC[209] =  1.461630e-01; 
enC[210] =  2.843000e+02; muC[210] =  4.188720e+00; 
enC[211] =  2.859890e+02; muC[211] =  4.158570e+00; 
enC[212] =  2.906150e+02; muC[212] =  4.077980e+00; 
enC[213] =  2.953150e+02; muC[213] =  3.998960e+00; 
enC[214] =  3.000920e+02; muC[214] =  3.921460e+00; 
enC[215] =  3.049450e+02; muC[215] =  3.845460e+00; 
enC[216] =  3.098780e+02; muC[216] =  3.770950e+00; 
enC[217] =  3.148900e+02; muC[217] =  3.697870e+00; 
enC[218] =  3.199830e+02; muC[218] =  3.626200e+00; 
enC[219] =  3.251580e+02; muC[219] =  3.555930e+00; 
enC[220] =  3.304180e+02; muC[220] =  3.487030e+00; 
enC[221] =  3.357620e+02; muC[221] =  3.419450e+00; 
enC[222] =  3.411920e+02; muC[222] =  3.353180e+00; 
enC[223] =  3.467110e+02; muC[223] =  3.288200e+00; 
enC[224] =  3.523190e+02; muC[224] =  3.224480e+00; 
enC[225] =  3.580170e+02; muC[225] =  3.162000e+00; 
enC[226] =  3.638080e+02; muC[226] =  3.100720e+00; 
enC[227] =  3.696920e+02; muC[227] =  3.040630e+00; 
enC[228] =  3.756720e+02; muC[228] =  2.981700e+00; 
enC[229] =  3.817480e+02; muC[229] =  2.916490e+00; 
enC[230] =  3.879220e+02; muC[230] =  2.848170e+00; 
enC[231] =  3.941970e+02; muC[231] =  2.781440e+00; 
enC[232] =  4.005730e+02; muC[232] =  2.716280e+00; 
enC[233] =  4.070520e+02; muC[233] =  2.652640e+00; 
enC[234] =  4.136350e+02; muC[234] =  2.590500e+00; 
enC[235] =  4.203260e+02; muC[235] =  2.529810e+00; 
enC[236] =  4.271240e+02; muC[236] =  2.470540e+00; 
enC[237] =  4.340320e+02; muC[237] =  2.412660e+00; 
enC[238] =  4.410520e+02; muC[238] =  2.356140e+00; 
enC[239] =  4.481860e+02; muC[239] =  2.300940e+00; 
enC[240] =  4.554350e+02; muC[240] =  2.247030e+00; 
enC[241] =  4.628020e+02; muC[241] =  2.194390e+00; 
enC[242] =  4.702870e+02; muC[242] =  2.142980e+00; 
enC[243] =  4.778940e+02; muC[243] =  2.092780e+00; 
enC[244] =  4.856230e+02; muC[244] =  2.043670e+00; 
enC[245] =  4.934780e+02; muC[245] =  1.994180e+00; 
enC[246] =  5.014590e+02; muC[246] =  1.945890e+00; 
enC[247] =  5.095700e+02; muC[247] =  1.898760e+00; 
enC[248] =  5.178120e+02; muC[248] =  1.852780e+00; 
enC[249] =  5.261870e+02; muC[249] =  1.807910e+00; 
enC[250] =  5.346980e+02; muC[250] =  1.764120e+00; 
enC[251] =  5.433460e+02; muC[251] =  1.721400e+00; 
enC[252] =  5.521340e+02; muC[252] =  1.679710e+00; 
enC[253] =  5.610650e+02; muC[253] =  1.639030e+00; 
enC[254] =  5.701390e+02; muC[254] =  1.599330e+00; 
enC[255] =  5.793610e+02; muC[255] =  1.560600e+00; 
enC[256] =  5.887320e+02; muC[256] =  1.522810e+00; 
enC[257] =  5.982540e+02; muC[257] =  1.485930e+00; 
enC[258] =  6.079300e+02; muC[258] =  1.449940e+00; 
enC[259] =  6.177630e+02; muC[259] =  1.414830e+00; 
enC[260] =  6.277550e+02; muC[260] =  1.380560e+00; 
enC[261] =  6.379080e+02; muC[261] =  1.347130e+00; 
enC[262] =  6.482260e+02; muC[262] =  1.314500e+00; 
enC[263] =  6.587110e+02; muC[263] =  1.282670e+00; 
enC[264] =  6.693650e+02; muC[264] =  1.251600e+00; 
enC[265] =  6.801910e+02; muC[265] =  1.222490e+00; 
enC[266] =  6.911930e+02; muC[266] =  1.195230e+00; 
enC[267] =  7.023720e+02; muC[267] =  1.168970e+00; 
enC[268] =  7.137330e+02; muC[268] =  1.139580e+00; 
enC[269] =  7.252770e+02; muC[269] =  1.108440e+00; 
enC[270] =  7.370080e+02; muC[270] =  1.078160e+00; 
enC[271] =  7.489280e+02; muC[271] =  1.048710e+00; 
enC[272] =  7.610420e+02; muC[272] =  1.020060e+00; 
enC[273] =  7.733510e+02; muC[273] =  9.921820e-01; 
enC[274] =  7.858590e+02; muC[274] =  9.650460e-01; 
enC[275] =  7.985700e+02; muC[275] =  9.386350e-01; 
enC[276] =  8.114860e+02; muC[276] =  9.129390e-01; 
enC[277] =  8.246110e+02; muC[277] =  8.879540e-01; 
enC[278] =  8.379490e+02; muC[278] =  8.636610e-01; 
enC[279] =  8.515020e+02; muC[279] =  8.399230e-01; 
enC[280] =  8.652740e+02; muC[280] =  8.167140e-01; 
enC[281] =  8.792690e+02; muC[281] =  7.940410e-01; 
enC[282] =  8.934910e+02; muC[282] =  7.720060e-01; 
enC[283] =  9.079430e+02; muC[283] =  7.505840e-01; 
enC[284] =  9.226280e+02; muC[284] =  7.297610e-01; 
enC[285] =  9.375510e+02; muC[285] =  7.093580e-01; 
enC[286] =  9.527150e+02; muC[286] =  6.895000e-01; 
enC[287] =  9.681240e+02; muC[287] =  6.701770e-01; 
enC[288] =  9.837830e+02; muC[288] =  6.513780e-01; 
enC[289] =  9.996950e+02; muC[289] =  6.331160e-01; 
enC[290] =  1.015860e+03; muC[290] =  6.153770e-01; 
enC[291] =  1.032290e+03; muC[291] =  5.981940e-01; 
enC[292] =  1.048990e+03; muC[292] =  5.812260e-01; 
enC[293] =  1.065960e+03; muC[293] =  5.644590e-01; 
enC[294] =  1.083200e+03; muC[294] =  5.481980e-01; 
enC[295] =  1.100720e+03; muC[295] =  5.323910e-01; 
enC[296] =  1.118520e+03; muC[296] =  5.170200e-01; 
enC[297] =  1.136610e+03; muC[297] =  5.021160e-01; 
enC[298] =  1.155000e+03; muC[298] =  4.876400e-01; 
enC[299] =  1.173680e+03; muC[299] =  4.735790e-01; 
enC[300] =  1.192660e+03; muC[300] =  4.599330e-01; 
enC[301] =  1.211950e+03; muC[301] =  4.466750e-01; 
enC[302] =  1.231550e+03; muC[302] =  4.337940e-01; 
enC[303] =  1.251470e+03; muC[303] =  4.211750e-01; 
enC[304] =  1.271720e+03; muC[304] =  4.088040e-01; 
enC[305] =  1.292290e+03; muC[305] =  3.966520e-01; 
enC[306] =  1.313190e+03; muC[306] =  3.848720e-01; 
enC[307] =  1.334430e+03; muC[307] =  3.734390e-01; 
enC[308] =  1.356010e+03; muC[308] =  3.623420e-01; 
enC[309] =  1.377940e+03; muC[309] =  3.515780e-01; 
enC[310] =  1.400230e+03; muC[310] =  3.411320e-01; 
enC[311] =  1.422880e+03; muC[311] =  3.309940e-01; 
enC[312] =  1.445890e+03; muC[312] =  3.211590e-01; 
enC[313] =  1.469280e+03; muC[313] =  3.116050e-01; 
enC[314] =  1.493040e+03; muC[314] =  3.023160e-01; 
enC[315] =  1.517190e+03; muC[315] =  2.932300e-01; 
enC[316] =  1.541730e+03; muC[316] =  2.844070e-01; 
enC[317] =  1.566670e+03; muC[317] =  2.758470e-01; 
enC[318] =  1.592010e+03; muC[318] =  2.675450e-01; 
enC[319] =  1.617760e+03; muC[319] =  2.594930e-01; 
enC[320] =  1.643920e+03; muC[320] =  2.516840e-01; 
enC[321] =  1.670510e+03; muC[321] =  2.441100e-01; 
enC[322] =  1.697530e+03; muC[322] =  2.367640e-01; 
enC[323] =  1.724990e+03; muC[323] =  2.296420e-01; 
enC[324] =  1.752890e+03; muC[324] =  2.226650e-01; 
enC[325] =  1.781240e+03; muC[325] =  2.158770e-01; 
enC[326] =  1.810050e+03; muC[326] =  2.092790e-01; 
enC[327] =  1.839320e+03; muC[327] =  2.028840e-01; 
enC[328] =  1.869070e+03; muC[328] =  1.966850e-01; 
enC[329] =  1.899300e+03; muC[329] =  1.906760e-01; 
enC[330] =  1.930020e+03; muC[330] =  1.848500e-01; 
enC[331] =  1.961240e+03; muC[331] =  1.792020e-01; 
enC[332] =  1.992960e+03; muC[332] =  1.737260e-01; 
enC[333] =  2.025200e+03; muC[333] =  1.684740e-01; 
enC[334] =  2.057950e+03; muC[334] =  1.632920e-01; 
enC[335] =  2.091240e+03; muC[335] =  1.582630e-01; 
enC[336] =  2.125060e+03; muC[336] =  1.533840e-01; 
enC[337] =  2.159430e+03; muC[337] =  1.486480e-01; 
enC[338] =  2.194360e+03; muC[338] =  1.440530e-01; 
enC[339] =  2.229850e+03; muC[339] =  1.395930e-01; 
enC[340] =  2.265920e+03; muC[340] =  1.352670e-01; 
enC[341] =  2.302570e+03; muC[341] =  1.310690e-01; 
enC[342] =  2.339810e+03; muC[342] =  1.269980e-01; 
enC[343] =  2.377660e+03; muC[343] =  1.230420e-01; 
enC[344] =  2.416110e+03; muC[344] =  1.192100e-01; 
enC[345] =  2.455190e+03; muC[345] =  1.154920e-01; 
enC[346] =  2.494900e+03; muC[346] =  1.118820e-01; 
enC[347] =  2.535260e+03; muC[347] =  1.083830e-01; 
enC[348] =  2.576260e+03; muC[348] =  1.049900e-01; 
enC[349] =  2.617930e+03; muC[349] =  1.016940e-01; 
enC[350] =  2.660270e+03; muC[350] =  9.850420e-02; 
enC[351] =  2.703300e+03; muC[351] =  9.540280e-02; 
enC[352] =  2.747030e+03; muC[352] =  9.239710e-02; 
enC[353] =  2.791460e+03; muC[353] =  8.948320e-02; 
enC[354] =  2.836610e+03; muC[354] =  8.666170e-02; 
enC[355] =  2.882490e+03; muC[355] =  8.391480e-02; 
enC[356] =  2.929110e+03; muC[356] =  8.125990e-02; 
enC[357] =  2.976480e+03; muC[357] =  7.868210e-02; 
enC[358] =  3.024630e+03; muC[358] =  7.618250e-02; 
enC[359] =  3.073550e+03; muC[359] =  7.375840e-02; 
enC[360] =  3.123260e+03; muC[360] =  7.140800e-02; 
enC[361] =  3.173780e+03; muC[361] =  6.912880e-02; 
enC[362] =  3.225110e+03; muC[362] =  6.691940e-02; 
enC[363] =  3.277270e+03; muC[363] =  6.477670e-02; 
enC[364] =  3.330280e+03; muC[364] =  6.269910e-02; 
enC[365] =  3.384150e+03; muC[365] =  6.068530e-02; 
enC[366] =  3.438880e+03; muC[366] =  5.873270e-02; 
enC[367] =  3.494500e+03; muC[367] =  5.683960e-02; 
enC[368] =  3.551020e+03; muC[368] =  5.500420e-02; 
enC[369] =  3.608460e+03; muC[369] =  5.322440e-02; 
enC[370] =  3.666820e+03; muC[370] =  5.149970e-02; 
enC[371] =  3.726130e+03; muC[371] =  4.982780e-02; 
enC[372] =  3.786400e+03; muC[372] =  4.820710e-02; 
enC[373] =  3.847640e+03; muC[373] =  4.663520e-02; 
enC[374] =  3.909870e+03; muC[374] =  4.511270e-02; 
enC[375] =  3.973110e+03; muC[375] =  4.363670e-02; 
enC[376] =  4.037380e+03; muC[376] =  4.222120e-02; 
enC[377] =  4.102680e+03; muC[377] =  4.083510e-02; 
enC[378] =  4.169030e+03; muC[378] =  3.949190e-02; 
enC[379] =  4.236460e+03; muC[379] =  3.819210e-02; 
enC[380] =  4.304980e+03; muC[380] =  3.693240e-02; 
enC[381] =  4.374620e+03; muC[381] =  3.571300e-02; 
enC[382] =  4.445370e+03; muC[382] =  3.453290e-02; 
enC[383] =  4.517270e+03; muC[383] =  3.338960e-02; 
enC[384] =  4.590330e+03; muC[384] =  3.228250e-02; 
enC[385] =  4.664580e+03; muC[385] =  3.121000e-02; 
enC[386] =  4.740030e+03; muC[386] =  3.017230e-02; 
enC[387] =  4.816690e+03; muC[387] =  2.916780e-02; 
enC[388] =  4.894600e+03; muC[388] =  2.819530e-02; 
enC[389] =  4.973770e+03; muC[389] =  2.725390e-02; 
enC[390] =  5.054210e+03; muC[390] =  2.634280e-02; 
enC[391] =  5.135960e+03; muC[391] =  2.546060e-02; 
enC[392] =  5.219030e+03; muC[392] =  2.460740e-02; 
enC[393] =  5.303440e+03; muC[393] =  2.378100e-02; 
enC[394] =  5.389220e+03; muC[394] =  2.298090e-02; 
enC[395] =  5.476390e+03; muC[395] =  2.220630e-02; 
enC[396] =  5.564970e+03; muC[396] =  2.145690e-02; 
enC[397] =  5.654980e+03; muC[397] =  2.073290e-02; 
enC[398] =  5.746440e+03; muC[398] =  2.003140e-02; 
enC[399] =  5.839390e+03; muC[399] =  1.935220e-02; 
enC[400] =  5.933830e+03; muC[400] =  1.869480e-02; 
enC[401] =  6.029810e+03; muC[401] =  1.805950e-02; 
enC[402] =  6.127330e+03; muC[402] =  1.744530e-02; 
enC[403] =  6.226440e+03; muC[403] =  1.685010e-02; 
enC[404] =  6.327150e+03; muC[404] =  1.627460e-02; 
enC[405] =  6.429480e+03; muC[405] =  1.571800e-02; 
enC[406] =  6.533480e+03; muC[406] =  1.517950e-02; 
enC[407] =  6.639150e+03; muC[407] =  1.465870e-02; 
enC[408] =  6.746540e+03; muC[408] =  1.415490e-02; 
enC[409] =  6.855650e+03; muC[409] =  1.366780e-02; 
enC[410] =  6.966540e+03; muC[410] =  1.319660e-02; 
enC[411] =  7.079220e+03; muC[411] =  1.274070e-02; 
enC[412] =  7.193720e+03; muC[412] =  1.230000e-02; 
enC[413] =  7.310070e+03; muC[413] =  1.187380e-02; 
enC[414] =  7.428310e+03; muC[414] =  1.146190e-02; 
enC[415] =  7.548450e+03; muC[415] =  1.106340e-02; 
enC[416] =  7.670540e+03; muC[416] =  1.067820e-02; 
enC[417] =  7.794610e+03; muC[417] =  1.030590e-02; 
enC[418] =  7.920680e+03; muC[418] =  9.945830e-03; 
enC[419] =  8.048790e+03; muC[419] =  9.597750e-03; 
enC[420] =  8.178980e+03; muC[420] =  9.261340e-03; 
enC[421] =  8.311260e+03; muC[421] =  8.936240e-03; 
enC[422] =  8.445690e+03; muC[422] =  8.621900e-03; 
enC[423] =  8.582290e+03; muC[423] =  8.318200e-03; 
enC[424] =  8.721110e+03; muC[424] =  8.024750e-03; 
enC[425] =  8.862160e+03; muC[425] =  7.741040e-03; 
enC[426] =  9.005500e+03; muC[426] =  7.466860e-03; 
enC[427] =  9.151160e+03; muC[427] =  7.202040e-03; 
enC[428] =  9.299170e+03; muC[428] =  6.946140e-03; 
enC[429] =  9.449580e+03; muC[429] =  6.699020e-03; 
enC[430] =  9.602420e+03; muC[430] =  6.460260e-03; 
enC[431] =  9.757730e+03; muC[431] =  6.229380e-03; 
enC[432] =  9.915550e+03; muC[432] =  6.006590e-03; 
enC[433] =  1.007590e+04; muC[433] =  5.791360e-03; 
enC[434] =  1.023890e+04; muC[434] =  5.583500e-03; 
enC[435] =  1.040450e+04; muC[435] =  5.382760e-03; 
enC[436] =  1.057280e+04; muC[436] =  5.188910e-03; 
enC[437] =  1.074380e+04; muC[437] =  5.001730e-03; 
enC[438] =  1.091760e+04; muC[438] =  4.821010e-03; 
enC[439] =  1.109420e+04; muC[439] =  4.646540e-03; 
enC[440] =  1.127360e+04; muC[440] =  4.478100e-03; 
enC[441] =  1.145590e+04; muC[441] =  4.315520e-03; 
enC[442] =  1.164120e+04; muC[442] =  4.158590e-03; 
enC[443] =  1.182950e+04; muC[443] =  4.007120e-03; 
enC[444] =  1.202080e+04; muC[444] =  3.860960e-03; 
enC[445] =  1.221530e+04; muC[445] =  3.719910e-03; 
enC[446] =  1.241280e+04; muC[446] =  3.583820e-03; 
enC[447] =  1.261360e+04; muC[447] =  3.452510e-03; 
enC[448] =  1.281760e+04; muC[448] =  3.325840e-03; 
enC[449] =  1.302500e+04; muC[449] =  3.203650e-03; 
enC[450] =  1.323560e+04; muC[450] =  3.085800e-03; 
enC[451] =  1.344970e+04; muC[451] =  2.972130e-03; 
enC[452] =  1.366720e+04; muC[452] =  2.862520e-03; 
enC[453] =  1.388830e+04; muC[453] =  2.756830e-03; 
enC[454] =  1.411290e+04; muC[454] =  2.654920e-03; 
enC[455] =  1.434120e+04; muC[455] =  2.556680e-03; 
enC[456] =  1.457310e+04; muC[456] =  2.461980e-03; 
enC[457] =  1.480890e+04; muC[457] =  2.370710e-03; 
enC[458] =  1.504840e+04; muC[458] =  2.282750e-03; 
enC[459] =  1.529180e+04; muC[459] =  2.197990e-03; 
enC[460] =  1.553910e+04; muC[460] =  2.116320e-03; 
enC[461] =  1.579040e+04; muC[461] =  2.037640e-03; 
enC[462] =  1.604580e+04; muC[462] =  1.961860e-03; 
enC[463] =  1.630540e+04; muC[463] =  1.888870e-03; 
enC[464] =  1.656910e+04; muC[464] =  1.818580e-03; 
enC[465] =  1.683710e+04; muC[465] =  1.750890e-03; 
enC[466] =  1.710940e+04; muC[466] =  1.685740e-03; 
enC[467] =  1.738610e+04; muC[467] =  1.623010e-03; 
enC[468] =  1.766740e+04; muC[468] =  1.562650e-03; 
enC[469] =  1.795310e+04; muC[469] =  1.504550e-03; 
enC[470] =  1.824350e+04; muC[470] =  1.448660e-03; 
enC[471] =  1.853860e+04; muC[471] =  1.394890e-03; 
enC[472] =  1.883840e+04; muC[472] =  1.343170e-03; 
enC[473] =  1.914310e+04; muC[473] =  1.293430e-03; 
enC[474] =  1.945270e+04; muC[474] =  1.245610e-03; 
enC[475] =  1.976740e+04; muC[475] =  1.199640e-03; 
enC[476] =  2.008710e+04; muC[476] =  1.162950e-03; 
enC[477] =  2.041200e+04; muC[477] =  1.118180e-03; 
enC[478] =  2.074210e+04; muC[478] =  1.075190e-03; 
enC[479] =  2.107760e+04; muC[479] =  1.033910e-03; 
enC[480] =  2.141850e+04; muC[480] =  9.942680e-04; 
enC[481] =  2.176500e+04; muC[481] =  9.561950e-04; 
enC[482] =  2.211700e+04; muC[482] =  9.196290e-04; 
enC[483] =  2.247470e+04; muC[483] =  8.845080e-04; 
enC[484] =  2.283820e+04; muC[484] =  8.507710e-04; 
enC[485] =  2.320760e+04; muC[485] =  8.183640e-04; 
enC[486] =  2.358300e+04; muC[486] =  7.872300e-04; 
enC[487] =  2.396440e+04; muC[487] =  7.573200e-04; 
enC[488] =  2.435200e+04; muC[488] =  7.285820e-04; 
enC[489] =  2.474590e+04; muC[489] =  7.009710e-04; 
enC[490] =  2.514620e+04; muC[490] =  6.744380e-04; 
enC[491] =  2.555290e+04; muC[491] =  6.489420e-04; 
enC[492] =  2.596620e+04; muC[492] =  6.244390e-04; 
enC[493] =  2.638610e+04; muC[493] =  6.008900e-04; 
enC[494] =  2.681290e+04; muC[494] =  5.782570e-04; 
enC[495] =  2.724660e+04; muC[495] =  5.565010e-04; 
enC[496] =  2.768730e+04; muC[496] =  5.355900e-04; 
enC[497] =  2.813510e+04; muC[497] =  5.154870e-04; 
enC[498] =  2.859020e+04; muC[498] =  4.961610e-04; 
enC[499] =  2.905260e+04; muC[499] =  4.775810e-04; 
enC[500] =  2.952250e+04; muC[500] =  4.597160e-04; 
enC[501] =  3.000000e+04; muC[501] =  4.425400e-04; 

% and now we convert them to en/mu columns
% (we want the energy grid to be in keV for acisabs
%  to avoid some multiplication)
muC *= 4.208e7 / ( enC * awC );
enC /= 1.0e3;

%---------------------------------------

% a few things used by the model
% (try and evaluate them once only)
%
% Ecal1/2 are in keV (now than enX is in keV)
%
static variable Ecal1 = 0.67;
static variable Ecal2 = 5.895;

%static variable muO_at_Ecal1 = interpol(enO,muO,Ecal1);
%static variable muO_at_Ecal2 = interpol(enO,muO,Ecal2);
static variable delta_muO = interpol(enO,muO,Ecal2) - interpol(enO,muO,Ecal1);

%static variable muH_at_Ecal1 = interpol(enH,muH,Ecal1);
%static variable muH_at_Ecal2 = interpol(enH,muH,Ecal2);
static variable delta_muH = interpol(enH,muH,Ecal2) - interpol(enH,muH,Ecal1);

%static variable muN_at_Ecal1 = interpol(enN,muN,Ecal1);
%static variable muN_at_Ecal2 = interpol(enN,muN,Ecal2);
static variable delta_muN = interpol(enN,muN,Ecal2) - interpol(enN,muN,Ecal1);

%static variable muC_at_Ecal1 = interpol(enC,muC,Ecal1);
%static variable muC_at_Ecal2 = interpol(enC,muC,Ecal2);
static variable delta_muC = interpol(enC,muC,Ecal2) - interpol(enC,muC,Ecal1);

%---------------------------------------

% variables used to cache the results
%
% As of 11/08/02 there appears to be some bizarre behaviour with the
% cache (ie it doesn't always seem to cache, at least for repeated
% calls to 'lp source')
%

% could use associative array (and so allow a possible speedup with non-fixed
% parameters), but that would involve a lot more memory
%
static variable _acisabs_cache_tdays  = -1;
static variable _acisabs_cache_tauinf = -1;
static variable _acisabs_cache_tefold = -1;
static variable _acisabs_cache_nC     = -1;
static variable _acisabs_cache_nH     = -1;
static variable _acisabs_cache_nO     = -1;
static variable _acisabs_cache_nN     = -1;

% need to cache the energy grid over which we are evaluated
% - could cache first/last/nbins but let's store the grid itself
%
static variable _acisabs_cache_bintype = -1;
static variable _acisabs_cache_binlo   = NULL;
static variable _acisabs_cache_binhi   = NULL;

static variable _acisabs_cache_result;

%---------------------------------------

public define acisabs() { % Tdays, tauinf, tefold, nC, nH, nO, nN, bintype, cache, binlo, binhi

%  	FORTRAN version of xcont.pro, written by George Chartas
%	Conversion performed by Konstantin Getman 05/20/02
%       S-lang version of acisabs, converted by Randall Smith
%
% ***       Ver 1.1 - Modified the previous ver. of acisabs.f
% ***       to use exponential decay instead of linear
   
%      Conversion performed by Konstantin Getman 07/17/02
%      S-lang version converted by Randall Smith 07/19/02
%      11/06/02 Doug Burke (S-Lang version 2)
%        added rudimentary caching
%      11/08/02 Doug Burke (S-Lang version 3)
%        data files are now included in the script 
%        some speed-ups (e.g. explicitly don't use the norm parameter)
%        removed the norm parameter
%        added the bintype and cache parameters
%        The energies from the Henke data files are now converted to be
%        in keV rather than eV
%
%			THE CONTENT OF xcont.pro	
%	 (1) Use observed decay with time in the ratio of observed 
%	 X-ray transmissions at 0.67keV and 5.895keV.
%
%        Observed decay in S3 based on Catherine Grants' analysis:
%        Fit parameters provided by Allyn Tennant
%
%        R(t) = (norm*exp(-tauinf*(1.0-exp(-t/tefold))))       eq(1)
%          norm=0.00722+/-0.00007
%          tefold=620.+/-66.
%          tauinf=0.582+/-0.024
%        
%	 
%	 (2) Model decay
%
%	 Trans(E1,t) = TOBF(E1) * Tcont(E1,t)   eq(2)
%
%	 where  TOBF(E1) is the transmission of the OBF at energy E1
%        Tcont(E1,t) is the transmission of the contamination
%        layer at energy E1 and time t
%
%	 The modeled decay is:
%
%        R(t)=Trans(E1,t)/Trans(E2,t)=const*exp{-rho*d(t)*[mac(E1)-mac(E2)]} eq(3)
%
%        where : rho is the density of the contaminant
%        d(t) is the thickness of the contaminant
%        mac(E) is the mass absorption coefficient of the 
%        contaminant at energy E.
%        mac(E) calculated from atomic scattering factor files
%        provided at http://www-cxro.lbl.gov/optical_constants/asf.html
%
%	The plot of the thickness of the contaminant vs. time is 
%       derived form eq(1) and eq(3):
%
%       d(t)*rho = ln(R(t)/const) / [mac(E2) - mac(E1)] 
%  
%	(3) Suggested modification of ACIS arf files to account for 
%           contamination
%
%       Tcont(E,t) = exp(-mac(E)*rho*d(t))
%	
%			THE CONTENT OF acisabs xspec model
%	Number of model parameters: 8
%	1	Tdays	Days between launch and observation
%       2       norm    Normalization factor
%       3       tauinf  Optical depth at infinite time
%       4       tefold  e-folding time for build-up of contaminant 
%	5	nC	Number of carbon atoms in hydrocarbon molecule
%	6	nH	Number of hydrogen atoms in hydrocarbon molecule
%	7	nO	Number of oxygen atoms in hydrocarbon molecule
%	8	nN	Number of nitrogen atoms in hydrocarbon molecule
%	
%	THIS PROGRAM REQUIRES the following data files to be present in the 
%       $HENKEDIR directory :
%	c.nff, h.nff, o.nff, n.nff
%

   variable Tdays, tauinf, tefold, nC, nH, nO, nN, bintype_flag, cache_flag, binlo, binhi;
   switch( _NARGS )
   { case 10:
      (Tdays, tauinf, tefold, nC, nH, nO, nN, bintype_flag, cache_flag, binlo) = ();
      binhi = NULL;
   }
   { case 11:
      (Tdays, tauinf, tefold, nC, nH, nO, nN, bintype_flag, cache_flag, binlo, binhi) = ();
   }
   {
     error("Internal error: acisabs called with an incorrect number of parameters");
   }

   % convert "integer" parameters to Integer_Type
   bintype_flag = int( bintype_flag );
   cache_flag   = int( cache_flag );

% HACK - use the cached result - if available
   if ( andelse
	{ cache_flag }
	{ Tdays         == _acisabs_cache_tdays }
	{ tauinf        == _acisabs_cache_tauinf }
	{ tefold        == _acisabs_cache_tefold }
	{ nC            == _acisabs_cache_nC }
	{ nH            == _acisabs_cache_nH }
	{ nO            == _acisabs_cache_nO }
	{ nN            == _acisabs_cache_nN }
	{ bintype_flag  == _acisabs_cache_bintype }
        { length(binlo) == length(_acisabs_cache_binlo) }
        { length(binhi) == length(_acisabs_cache_binhi) }
        { 0             == sum(_acisabs_cache_binlo != binlo) }
        { 0             == sum(_acisabs_cache_binhi != binhi) }
      )
      return _acisabs_cache_result;

% Calculate the relative abundance of each element
   variable xO = nO * awO;
   variable xH = nH * awH;
   variable xN = nN * awN;
   variable xC = nC * awC;
   variable elnorm = xO + xH + xN + xC;

   xO = xO/elnorm;
   xH = xH/elnorm;
   xN = xN/elnorm;
   xC = xC/elnorm;

% Calculate difference in mass absorption coefficients at Ecal1 and Ecal2
   variable dMacE21 = 
     xO * delta_muO + xH * delta_muH + xN * delta_muN + xC * delta_muC;

% Estimate thickness of contaminate during observation
% NOTE: have removed use of norm parameter (DJB 11/08/02)
   variable thRho = log10( exp(-tauinf*(1.0 - exp(-Tdays/tefold))) )*2.30259/dMacE21;
   
% Calculate transmission of contamination layer for input energies
%
% if bintype_flag = 0 we assume the input bins are in keV
% if bintype_flag = 1 we assume the input bins are in Angstrom
%              
% and we want energies in keV for the code below
%
   variable eninpt;
   if ( binhi == NULL ) {
     if ( bintype_flag )
       eninpt = 1.2398427e1 / binlo;
     else
       eninpt = @binlo; % ensure it's a copy
   } else {
     if ( bintype_flag )
       eninpt = 2.4796854e1 / (binlo + binhi);
     else
       eninpt = 0.5 * (binlo + binhi);
   }

   variable muO_on_grid = interpol(enO,muO,eninpt);
   variable muH_on_grid = interpol(enH,muH,eninpt);
   variable muN_on_grid = interpol(enN,muN,eninpt);
   variable muC_on_grid = interpol(enC,muC,eninpt);

   % it's important that the output type is correct
   % (otherwise seg faults are a distinct possibility)
   %
   variable result =
	typecast( 
		exp( -thRho*( xO*muO_on_grid + xH*muH_on_grid + xN*muN_on_grid + xC*muC_on_grid) ),
		_typeof(binlo) );

   % set up the cache
   if ( cache_flag ) {
     _acisabs_cache_tdays   = Tdays;
     _acisabs_cache_tauinf  = tauinf;
     _acisabs_cache_tefold  = tefold;
     _acisabs_cache_nC      = nC;
     _acisabs_cache_nH      = nH;
     _acisabs_cache_nO      = nO;
     _acisabs_cache_nN      = nN;
     _acisabs_cache_bintype = bintype_flag;
     _acisabs_cache_binlo   = binlo;
     _acisabs_cache_binhi   = binhi;
     _acisabs_cache_result  = result;
   }

   return result;

} % acisabs()

%---------------------------------------

% if we can, register the model with Sherpa
%

static variable _maxt   = 10000.0;
static variable _maxn   = 50.0;

% must be visible via namespace for the eval() call below to work
%
variable _params = [ "Tdays","tauinf","tefold","nC","nH","nO","nN","bintype","cache" ];
variable _defs   = [     850,  0.582,    620.0,  10,  20,   2,   1,        0,      1 ];
variable _mins   = [       0,      0,        1,   0,   1,   0,   0,        0,      0 ];
variable _maxs   = [   _maxt,      1,    _maxt, _maxn, _maxn, _maxn, _maxn, 1, 1 ];
variable _state  = [       0,      0,        0,   0,   0,   0,   0,        0,      0 ];

% since we are using an eval() statement below to call the
% register function then we need to make all the variables in the
% argument list visible from the Global namespace, hence the
% sets of "acisabs->"'s
%
static variable _eval_str =
  "register_model(\"acisabs\",acisabs->_params,1,acisabs->_defs,acisabs->_mins,acisabs->_maxs,acisabs->_state);";

switch ( _is_reg_model )
{ case 1 : eval( "() = sherpa_" + _eval_str ); }
{ case 2 : eval( "() = " + _eval_str ); }

%---------------------------------------

% used by apply_acisabs() to ensure that FITS cards fit into 
% 80 charactes (and adds it as a comment)
%
private define _str2hist (in) {
  variable out = sprintf( "HISTORY   %-70s", in );
  if ( strlen(out) > 80 ) out = out[[0:79]];
  return out;
}

% this function taken from sherpa_utils.sl 
% (see http://cxc.harvard.edu/ciao/download/scripts/ )
% - comments/checking stripped out of this version
%
private define is_struct_field () {
  variable args = __pop_args( _NARGS );
  variable instruct = args[0].value;
  if ( 0 == is_struct_type(instruct) ) return 0;
  variable names = Assoc_Type [Integer_Type];
  variable fname;
  foreach( get_struct_field_names( instruct ) ) {
    fname = ();
    names[fname] = 1;
  }
  variable arg;
  foreach ( args[[1:-1]] ) {
    arg = ();
    fname = arg.value;
    if ( typeof(fname) != String_Type ) return 0;
    !if ( assoc_key_exists( names, fname ) ) return 0;
  }
  return 1;
} % is_struct_field()

% the "corrarf" function
%
% Usage:
%   flag = apply_acisabs( oldarfname, newarfname [, display_correction ] );
%
% Aim:
%   Apply the acisabs model to the ARF stored in
%   oldarfname and write out the corrected file
%   newarfname. If a third parameter is supplied and evaluates
%   to true (such as the integer 1), then the correction
%   factor will be displayed in ChIPS using a log log scale
%
%   We use the default parameter values used for the
%   model except for the Tdays parameter, for which 
%   we use the TSTART value from the header.
%
%   Returns a 0 on success, 1 on error
%
% Notes:
%
%   Can be used with grating ARFS, but you probably shouldn't
%
%   This routine loses the header information in the
%   primary block of the ARF.
%
%   Makes use of the acisabs->_defs array set up when
%   the model is registered (actually, this function can
%   be called from within ChIPS as it doesn't need the
%   acisabs model to be registered with Sherpa, but it
%   does need the _defs array, and have Tdays be the
%   first parameter)
%
%   We keep an array of filenames created by the routine
%   so that (in CIAO 2.3) we can warn the user if they try
%   to use the same name (since the output file won't be changed
%   due to the cacheing in Varmm in CIAO 2.3). This shouldn't
%   be necessary in CIAO 3.0. We should perhaps also
%   try inarf as well, although it's less obvious that it's
%   a problem
%

public define apply_acisabs () { % inarf, outarf [, display_flag ]

  variable usage_str = 
    "Usage: apply_acisabs(inarf,outarf[,display_flag]);\n\n" +
    "where inarf is the name of the ARF to be corrected and outarf is\n" +
    "the name of the corrected ARF created by this routine.\n" +
    "Supply a true value (such as 1) for display_flag to display the\n" +
    "correction factor in the current ChIPS display\n";

  variable inarf, outarf, display_flag;
  display_flag = 0;
  switch ( _NARGS )
  { case 2: ( inarf, outarf ) = (); }
  { case 3: ( inarf, outarf, display_flag ) = (); }
  { message( usage_str ); return 1; }

  % safety check
  if ( typeof(inarf) != String_Type or typeof(outarf) != String_Type ) {
    message( usage_str );
    return 1;
  }

  % check the output name (we update the array after the writefits() call)
  % - note: we do not create the output file
  if ( andelse
       { _does_varmm_cache }
       { assoc_key_exists( _apply_acisabs_outnames, outarf ) } ) { 
    message( "\nERROR: you have already used outarf=" + outarf + "\n" +
      "in this session.  Please use a different filename, since VARMM in CIAO 2.3\n" +
      "caches data by filename (so your output file would probably not contain the\n" +
      "data you expect it to).\n" );
    return 1;
  }

  % read in the ARF: ensure we are in upper case
  variable old_flag = varmm.caseinsen;
  varmm.caseinsen = 0;
  variable in = readarf( inarf );
  varmm.caseinsen = old_flag;
  if ( in == NULL ) {
    message( "\nERROR: the file " + inarf + " does not appear to be an ARF\n" );
    return 1;
  }

  % warning if a grating ARF since using the default parameters is probably not
  % sensible here
  %
  if ( is_struct_field( in, "BIN_LO" ) )
    message( "\nWARNING: it appears that you are trying to correct a grating ARF\n" +
      "The acisabs parameters used to create the model are not optimised for\n" +
      "gratings work - see http://cxc.harvard.edu/cal/Acis/Cal_prods/qeDeg/\n" );

  % get the parameter values (could get TSTART by parsing in._header but that's painful)
  %
  variable tdays = tstart2days( inarf );

  % create the ACISABS model for the ARF's energy grid
  %
  % NOTE:
  %   we assume an "integrate off" model so that we
  %   match the default settings in Sherpa (ie that we get the
  %   same result as we would if the model we're bing used as a
  %   'source' component)
  %
  variable transmission = acisabs(
    tdays, _defs[1], _defs[2], _defs[3], _defs[4], _defs[5], _defs[6], _defs[7], _defs[8], 
    in.ENERG_LO );
%%%    in.ENERG_LO, in.ENERG_HI );
  
  % correct the arf
  %
  variable out = @in;
  out.SPECRESP *= typecast( transmission, _typeof(out.SPECRESP) );

  % display the correction?
  % - we change the setting of redraw, which might annoy chips/sherpa
  %   (unfortunately in CIAO 2.3 we can't query the state of the redraw flag)
  % - we use a major-axis dashed grid
  % - we clear the display just to make sure that we know what
  %   the state of the plot is. It would be nice to not do this,
  %   or at least make it configurable, but that's more options
  %
  if ( display_flag ) {

    variable enaxis = in.ENERG_LO;
%%    variable enaxis = (in.ENERG_LO+in.ENERG_HI)/2;
 
    variable ochips = @chips;
    chips.curvestyle  = _chips->histo;
    chips.symbolstyle = _chips->none;

    () = chips_eval("redraw off");
    () = chips_eval("clear");
    () = chips_eval("split 2");
    () = chips_eval("d 1 location 0.15 0.9 0.6 0.9");
    () = chips_eval("d 2 location 0.15 0.9 0.1 0.6");
    () = chips_eval("d 1 tickvals x off");

    () = chips_eval("d 1");
    () = curve( enaxis, in.SPECRESP );
    () = curve( enaxis, out.SPECRESP );
    () = chips_eval( "dot" );
    () = chips_eval( "red" );
    () = chips_eval( "log" );
    () = chips_eval( "grids maj dash" );
    () = chips_eval( "ylabel 'ARF (cm^2)'" );
    () = chips_eval( "ylabel size 1.2" );

    () = chips_eval("d 2");
    () = curve( enaxis, transmission );
    () = chips_eval( "log x" );
    () = chips_eval( "grids maj dash" );
    () = chips_eval( "xlabel 'Energy (keV)'" );
    () = chips_eval( "ylabel 'Transmission'" );

    () = chips_eval( "title 'Correction for Tdays = " + string(tdays) + "'" );
    () = chips_eval( "xlabel size 1.2" );
    () = chips_eval( "ylabel size 1.2" );
    () = chips_eval( "title size 1.2" );

    () = chips_eval("redraw on");

    % restore the ChIPS settings
    set_state( "chips", ochips );
  }

  % add something to the header
  % (this isn't as easy as we'd like it in CIAO 2.3)
  %
  variable nlines = 13;
  variable len = length(in._header);
  variable nheader = String_Type [nlines+len];
  nheader[[0:len-1]] = in._header;

  nlines *= -1;
  nheader[nlines] = _str2hist( sprintf("%s %s at %s",
     _function_name,_acisabs_version_string,ctime(_time)) );
  nlines++;
  nheader[nlines] = _str2hist( sprintf("inarf=%s",inarf) );
  nlines++;
  nheader[nlines] = _str2hist( sprintf("outarf=%s",outarf) );
  nlines++;
  nheader[nlines] = _str2hist( sprintf("Tdays=%f",tdays) );

  variable ctr;
  _for ( 1, 6, 1 ) {
    ctr = ();
    nlines++;
    nheader[nlines] = _str2hist( sprintf("%s=%f",_params[ctr],_defs[ctr]) );
  }
  _for ( 7, 8, 1 ) {
    ctr = ();
    nlines++;
    nheader[nlines] = _str2hist( sprintf("%s=%d",_params[ctr],int(_defs[ctr])) );
  }

  % include a warning about the loss of header information
  nlines++;
  nheader[nlines] = _str2hist( "*** note: header information from PRIMARY block has been lost ***" );

  % Could use EXTNAME value from the header, but let's hard code it
  if ( -1 == writefits( outarf, out, "SPECRESP", nheader ) ) {
    message("\nERROR: unable to write ARF to " + outarf + "\n");
    return 1;
  }

  if ( _does_varmm_cache ) _apply_acisabs_outnames[outarf] = 1;
  message("Created: " + outarf );
  return 0;
 
} % apply_acisabs()

%---------------------------------------
